// BV_S: classes to manage solutions, moves and attributes of generic boolean vector problems

#if !defined(__HotFrame_BV_S__)
#define __HotFrame_BV_S__

#include "basic.h"
#include "generic_s.h"


//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
namespace HotFrame
{


// forward declarations
template <class C> class BV_S;
template <class C> class BV_S_N;
template <class C> class BV_S_A;
template <class C> class BV_S_I;


// ----- class BV_S -----------------------------------------------------------------
/*
  this class manages solutions (stores, evaluates, provides, ...)
  - needed for all heuristics
*/


template <class C>
class BV_S
: public Generic_S<C>
{
  // to efficiently implement interfaces to third-party-software software,
  // we may need to use int as bit representation where we actually should
  // use bool; so keep it separated by introducing the type Bit
  public:
    typedef bool Bit;

  protected:
    Range _n;  // client view of object numbering: 1,...,n
    vector<Bit> _b;  // current solution: _b[0],...,_b[_n-1]

  public:
    BV_S( Range n, Observer<C>* observer );

    virtual bool computeEvaluation( const Generic_S_N<C>& neighbor, T& evaluation, T& delta );

    virtual void doMove( Generic_S_N<C>& neighbor );

    virtual Generic_S<C>& operator=( const Generic_S<C>& rhs )
    {
      if ( this != &rhs )
      {
        assert( typeid(rhs) == typeid(*this) );
        const BV_S<C>& drhs = static_cast<const BV_S<C>&>( rhs );
        Generic_S<C>::operator=( rhs );
        _n = drhs._n;
        _b = drhs._b;
			}
      return *this;
    }

    BV_S<C>& operator=( const BV_S<C>& rhs )
    {
      if ( this != &rhs )
        operator=( static_cast<const Generic_S<C>&>( rhs ) );
      return *this;
    }

    Bit getValue( Range i ) const
    {
      assert( i>=1 ); assert( i<=_n );
      return _b[i-1];
    }

    void setValue( Range i, Bit value )
	  {
      assert( i>=1 ); assert( i<=_n );
		  _b[i-1] = value;
	  }

    Range n( ) const
      { return _n; }

    virtual void print( ostream& output ) const
    {
      for( Range i=0; i<_n; ++i )
        output << _b[i] << " ";

      output << "\n" << _evaluation << "\n";
    }

    // the neighborhood iterator class uses the following two methods to set the first and
    // increment to the next neighbor; usually you would not change this, but by changing
    // this you can achieve things such as fixing variables ...
    virtual Range firstVariable( ) const
      { return 1; }
    virtual void nextVariable( Range &i ) const
      { ++i; }
    virtual Range randomVariable( ) const
      { return ( rand( n() ) ) + 1; }  // i.e. [1..n] }
};

template <class C>
float distance< >( const BV_S<C>& s1, const BV_S<C>& s2 )
{
  assert( s1.size() == s2.size() );

  Range diffs = 0;
  for( Range i=0; i<s1.size(); ++i )
      if ( _b[i] != rhs._b[i] )
        ++diffs;

  return ( (float(diffs)) / s1.size() );
}

// ----- class BV_S_N ---------------------------------------------------------------
/*
  this class manages neighbors (stores, evaluates, provides, ...)
  - needed for all neighborhood based heuristics
*/

template <class C>
class BV_S_N
: public Generic_S_N<C>
{
  public:
    enum NeighborhoodType { AllFlips, Constructive, Destructive };
    static NeighborhoodType _s_neighborhoodType;

  public:  //  ((protected would be the right thing here ...))
    BV_S<C> *_currentSolution;
    vector<Range> _flips;  // starting from _flips.begin() (i.e. _flips[0])

    virtual void evaluate( );

  public:
    // constructor
    BV_S_N( BV_S<C> *currentSolution,
			      NeighborhoodPosition itpos = FirstNeighbor,
            short depth = 1 );

    // standard constructor, needed by Generic_S_NCLA
    BV_S_N( ) : _currentSolution( 0 )
      { ; }

    BV_S_N( const BV_S_N& original )
      : Generic_S_N<C>( original ),
        _currentSolution( original._currentSolution ), _flips( original._flips )
    { ; }

    virtual void operator++( );

    virtual bool isValid( ) const
      { return ( _flips[0] <= _currentSolution->n() ); }

    BV_S<C> * currentSolution( ) const
	  	{ return _currentSolution; }

    Range flip( Range i ) const
      { return _flips[i]; }

    Range actualDepth( ) const
      { return _flips.size(); }

    virtual Generic_S_N<C>& operator=( const Generic_S_N<C>& rhs )
    {
      if ( this != &rhs )
      {
        // assert( typeid(rhs) == typeid(*this) );  // does not work in connection with candidate lists
        const BV_S_N<C>& drhs = static_cast<const BV_S_N<C>&>( rhs );
        Generic_S_N<C>::operator=( rhs );
        _flips = drhs._flips;
        _currentSolution = drhs._currentSolution;
      }
      return *this;
    }

    BV_S_N<C>& operator=( const BV_S_N<C>& rhs )
    {
      if ( this != &rhs )
        operator=( static_cast<const Generic_S_N<C>&>( rhs ) );
      return *this;
    }

    virtual unsigned long size( ) const
      { return _currentSolution->n(); }

    static void combineNeighbors( BV_S_N<C>& to, const BV_S_N<C>& from1, const BV_S_N<C>& from2 )
    // so far: restrict to combining two neighbors of depth 1 to one neighbor of depth 2
    {
      assert( from1.getLambda() == 1 );
      assert( from2.getLambda() == 1 );

      to._flips.clear();
      to._flips.push_back( from1.flip( 0 ) );
      to._flips.push_back( from2.flip( 0 ) );
      to._depth = 2;
    }

    Observer<C>* getObserver( ) const
      { return _currentSolution->observer(); }

    virtual void print( ostream& output ) const;
};

template <class C>
typename BV_S_N<C>::NeighborhoodType BV_S_N<C>::_s_neighborhoodType = BV_S_N<C>::AllFlips;


// the following template function defines a transformation from Perm_S_N_Shift<C>
// moves to numbers, which is used by the REMTabuCriterion; so one must specialize
// this template function to use the REMTabuCriterion

template <class C>
unsigned long moveNumber( const BV_S_N<C>& neighbor )
{
  return neighbor._flips[0]; // of course, this works so far only for depth=1
}



// ----- class BV_S_A ---------------------------------------------------------------
/*
  this class manages attributes about solutions (e.g., extracts add/drop attributes from moves)
  - needed for all tabu search heuristics that use an attribute based tabu list
*/

template <class C>
class BV_S_A
{
  public:
    typedef typename C::T  T;
    typedef typename C::Range Range;

  protected:
    Range _flip;
    // _flip > 0 ( < 0 ) represents the variable flip to be set to 1 (to 0)

  public:
    BV_S_A( const BV_S_N<C>& neighbor, unsigned int attributeIndex = 1 );
    BV_S_A( const BV_S<C>& s, unsigned int attributeIndex );
    BV_S_A( );  // default constructor has to be provided (constructs dummy object)
    virtual bool operator==( const BV_S_N<C>& neighbor ) const;

    static Range getNumberOfPositiveAttributes( const BV_S_N<C>& neighbor );
    static Range getNumberOfNegativeAttributes( const BV_S_N<C>& neighbor )
      { return getNumberOfPositiveAttributes( neighbor ); }

    virtual void print( ostream& output ) const;
};


// ----- class BV_S_I ---------------------------------------------------------------
/*
  this class manages information about solutions
  - possibly needed for all tabu search heuristics that use a trajectory based memory
  - here: done approximately using hash codes
*/

template <class C>
class BV_S_I
{
  public:
    typedef typename C::T  T;
    typedef typename C::Range Range;

  protected:
    unsigned long _hashCode;
    static HashCoefficients _s_hashCoefficients;

  public:
    BV_S_I( const BV_S<C>& solution );
    BV_S_I( const BV_S_N<C>& neighbor );

    virtual bool operator==( const BV_S_I<C>& rhs ) const
      { return ( rhs._hashCode == _hashCode ); }
    virtual bool operator<( const BV_S_I<C>& rhs ) const
      { return ( rhs._hashCode < _hashCode ); }

    Range getHashCode() const
      { return _hashCode; }
    virtual void print( ostream& output ) const;
};


// definition of static class elements -----------------------------------------

template<class C>
HashCoefficients BV_S_I<C>::_s_hashCoefficients;


// ----- implementation of class BV_S ------------------------------------------

template<class C>
BV_S<C>::BV_S( Range n, Observer<C>* observer )
: Generic_S<C>( observer ),
  _n( n ),
  _b( _n, false )
{
}


template<class C>
bool BV_S<C>::computeEvaluation( const Generic_S_N<C>& neighbor,
                                 T& evaluation, T& delta )
{
  // this is an inefficient default evaluation, which only works for BV_S_N neighbors;
  // this should be redefined in an inheriting class to efficiently evaluate a neighhbor adaptively

  // dynamic downcast
  const BV_S_N<C>& ne = dynamic_cast<const BV_S_N<C>&>( neighbor );

  T oldEvaluation = _evaluation;

  Range i;
  for( i=0; i<ne.actualDepth(); ++i )
    _b[ne.flip(i)-1] = invert( _b[ne.flip(i)-1] );

  evaluate();

  delta = _evaluation - oldEvaluation;
  evaluation = -delta;

  // restore current solution (*this)
  for( i=0; i<ne.actualDepth(); ++i )
    _b[ne.flip(i)-1] = invert( _b[ne.flip(i)-1] );

  _evaluation = oldEvaluation;

  return true;  // i.e., delta computed
}


template<class C>
void BV_S<C>::doMove( Generic_S_N<C>& neighbor )
{
  // dynamic downcast
  BV_S_N<C>& ne = dynamic_cast<BV_S_N<C>&>( neighbor );

  if ( _observer->active() )
  {
    _observer->move();
    _observer->getStream() << ne;
    _observer->flush();
    _observer->moveEvaluation( *ne );
    if ( ne.deltaInformation() )
      _observer->deltaEvaluation( ne.delta( ) );
  }

  for( Range i=0; i<ne.actualDepth(); ++i )
    _b[ne.flip(i)-1] = invert( _b[ne.flip(i)-1] );

  if ( ne.deltaInformation( ) )
    _evaluation += ne.delta();
  else
    evaluate();

  if ( _evaluation < _s_upperBound )
  {
    _s_upperBound = _evaluation;
    _observer->newUpperBound( _evaluation );
    _observer->timeWhenNewUpperBoundWasFound( round( (double(clock()))/CLOCKS_PER_SEC, 3 )  );
  }
}


// ----- implementation of class BV_S_N ---------------------------------------

template <class C>
BV_S_N<C>
  ::BV_S_N( BV_S<C> *currentSolution,
	          NeighborhoodPosition itpos, short depth ) // =FirstNeighbor, =1
  : Generic_S_N<C>( depth ), _currentSolution( currentSolution ), _flips( 1, 0 )
{
  if ( itpos == InvalidNeighbor )
  {
    _flips[0] = currentSolution->n() + 1;
    return;
  }

  // else
  if ( itpos == SomeRandomNeighbor )
  {
    // so far, depth is not considered here (but depth = 1 is assumed)

    if ( _s_neighborhoodType == AllFlips )
      _flips[0] = _currentSolution->randomVariable( );
    else
    {
      vector<Range> candidates;
      Range c = _currentSolution->firstVariable( );
      while ( c <= _currentSolution->n() )
      {
        if (    ( _s_neighborhoodType == Constructive )
             && ( _currentSolution->getValue(c) == 0 ) )
          candidates.push_back( c );
        if (    ( _s_neighborhoodType == Destructive )
             && ( _currentSolution->getValue(c) == 1 ) )
          candidates.push_back( c );
        _currentSolution->nextVariable( c );
      }

      if ( candidates.size() != 0 )
        _flips[0] = candidates[ rand( candidates.size() ) ];
      else
        _flips[0] = currentSolution->n() + 1;
    }

    if ( _flips[_flips.size()-1] <= _currentSolution->n() )
      evaluate();
    return;
  }

  // else construct first neighbor
  _flips[0] = _currentSolution->firstVariable( );
  while (    ( _flips[_flips.size()-1] <= _currentSolution->n() )
          && (    (    ( _s_neighborhoodType == Constructive )
                    && ( _currentSolution->getValue(_flips[_flips.size()-1]) == 1 ) )
               ||
                  (    ( _s_neighborhoodType == Destructive )
                    && ( _currentSolution->getValue(_flips[_flips.size()-1]) == 0 ) )
             )
        )
    _currentSolution->nextVariable( _flips[0] );

  if ( _flips[_flips.size()-1] <= _currentSolution->n() )
    evaluate();
}


template<class C>
void BV_S_N<C>::evaluate()
{
  _deltaInformation =
    _currentSolution->computeEvaluation( *this, _evaluation, _delta );
}


template<class C>
void BV_S_N<C>::operator++( )
{
  // the following code requires a little bit of thinking
  // (due to potentially larger neighborhood depths and non-allflips-neigborhoodtype)

  // ToDo: verify combination of _depth >= 2 with non-allflips-neigborhoodtype

  if ( _depth >= 2 )
  {
    if ( _s_neighborhoodType == AllFlips )
    {
      if ( _flips.size() < _depth )
        _flips.push_back( _currentSolution->firstVariable( ) );
      else
      {
        _currentSolution->nextVariable( _flips[_flips.size()-1] );

        if( _flips[_flips.size()-1] > _currentSolution->n() )
        // backtracking
        {
          do
          {
            _flips.pop_back( );
            _currentSolution->nextVariable( _flips[_flips.size()-1] );
          }
          while( ( _flips.size() > 1 ) && ( _flips[_flips.size()-1] > _currentSolution->n() ) );

          if ( _flips[_flips.size()-1] > _currentSolution->n() )
            return; // i.e., _flips.size()==1, last neighbor reached
        }
      }
    }
    else // ToDo: chech this, there is still some error ...
    {
      Range newFlip = _currentSolution->firstVariable( );
      if ( _flips.size() < _depth )
      {
        while (    ( newFlip <= _currentSolution->n() )
                && (    (    ( _s_neighborhoodType == Constructive )
                          && ( _currentSolution->getValue(newFlip) == 1 ) )
                     ||
                        (    ( _s_neighborhoodType == Destructive )
                          && ( _currentSolution->getValue(newFlip) == 0 ) )
                   )
              )
          _currentSolution->nextVariable( newFlip );
        if ( newFlip <= _currentSolution->n() )
          _flips.push_back( newFlip );
      }

      if ( ( _flips.size() >= _depth ) || ( newFlip > _currentSolution->n() ) )
      {
        do
          _currentSolution->nextVariable( _flips[_flips.size()-1] );
        while (    ( _flips[_flips.size()-1] <= _currentSolution->n() )
                && (    (    ( _s_neighborhoodType == Constructive )
                          && ( currentSolution()->getValue(_flips[_flips.size()-1]) == 1 ) )
                     ||
                        (    ( _s_neighborhoodType == Destructive )
                          && ( currentSolution()->getValue(_flips[_flips.size()-1]) == 0 ) )
                   )
              );

        if( _flips[_flips.size()-1] > _currentSolution->n() )
        // backtracking
        {
          do
          {
            _flips.pop_back( );
            _currentSolution->nextVariable( _flips[_flips.size()-1] );
          }
          while (    ( ( _flips.size() > 1 ) && ( _flips[_flips.size()-1] > _currentSolution->n() ) )
                  && (    (    ( _s_neighborhoodType == Constructive )
                            && ( currentSolution()->getValue(_flips[_flips.size()-1]) == 1 ) )
                       ||
                          (    ( _s_neighborhoodType == Destructive )
                            && ( currentSolution()->getValue(_flips[_flips.size()-1]) == 0 ) )
                     )
                );

          if ( _flips[_flips.size()-1] > _currentSolution->n() )
            return; // i.e., _flips.size()==1, last neighbor reached
        }
      }
    }
  }
  else
  {
    do
      _currentSolution->nextVariable( _flips[0] );
    while (    ( _flips[_flips.size()-1] <= _currentSolution->n() )
            && (    (    ( _s_neighborhoodType == Constructive )
                      && ( currentSolution()->getValue(_flips[0]) == 1 ) )
                 ||
                    (    ( _s_neighborhoodType == Destructive )
                      && ( currentSolution()->getValue(_flips[0]) == 0 ) )
               )
          );
  }

  if ( _flips[_flips.size()-1] <= _currentSolution->n() )
    evaluate();
}

template<class C>
void BV_S_N<C>::print( ostream& output ) const
{
  for ( Range i=0; i<_flips.size(); ++i )
    if ( _currentSolution->getValue( _flips[i] ) )
      output << -_flips[i] << " ";
    else
      output << _flips[i] << " ";

  output << "\n";
}


// ----- implementation of class BV_S_A ---------------------------------------

template<class C>
BV_S_A<C>::BV_S_A( const BV_S_N<C>& neighbor,
                   unsigned int attributeIndex ) // =1
{
  _flip = neighbor.flip( attributeIndex - 1 );

  if ( neighbor.currentSolution()->getValue( _flip ) == 1 )
    // variable flip would be set 0, so we have the (positive) attribute -_flip
    _flip = -_flip;

  if ( attributeIndex < 0 )
    _flip = -_flip;
}


template<class C>
BV_S_A<C>::BV_S_A( const BV_S<C>& s, unsigned int attributeIndex )
{
  if ( s.getValue( attributeIndex ) == 1 )
    _flip = attributeIndex;
  else
    _flip = -attributeIndex;
}


template<class C>
BV_S_A<C>::BV_S_A( )
// default constructor has to be provided (constructs dummy object/attribute)
: _flip( 0 )
{
  ;
}


template<class C>
bool BV_S_A<C>::operator==( const BV_S_N<C>& neighbor ) const
{
  for( Range i=0; i<neighbor.actualDepth(); ++i )
  {
    Range destroyedAttribute = - neighbor.flip( i );

    if ( ( _flip > 0 ) && ( neighbor.currentSolution()->getValue( _flip ) == 1 ) )
      // variable flip would be set 0, so we have the (positive) attribute -_flip
      destroyedAttribute = -destroyedAttribute;

    if ( _flip == destroyedAttribute )
      return true;
  }

  return false;
}


template<class C>
typename BV_S_A<C>::Range BV_S_A<C>::getNumberOfPositiveAttributes( const BV_S_N<C>& neighbor )
{
  return neighbor.actualDepth();
}


template<class C>
void BV_S_A<C>::print( ostream& output ) const
{
  output << _flip << "\n";
}


// ----- implementation of class BV_S_I ---------------------------------------

template <class C>
BV_S_I<C>::BV_S_I( const BV_S<C>& solution )
{
  if ( _s_hashCoefficients.size() == 0 )
    _s_hashCoefficients.initialize( solution.n(), solution.n()*solution.n() );

  _hashCode = 0;
  for( Range i=1; i<=solution.n(); ++i )
    if ( solution.getValue( i ) )
      _hashCode += _s_hashCoefficients.get_coefficient( i );
}


template <class C>
BV_S_I<C>::BV_S_I( const BV_S_N<C>& neighbor )
{
  if ( _s_hashCoefficients.size() == 0 )
    _s_hashCoefficients.initialize( neighbor.currentSolution()->n(),
                                    neighbor.currentSolution()->n()
                                     *neighbor.currentSolution()->n() );

  _hashCode = 0;
  for( Range i=1; i<=neighbor.currentSolution()->n(); ++i )
  {
    BV_S<C>::Bit value = neighbor.currentSolution()->getValue( i );

    for( Range j=0; j<neighbor.actualDepth(); ++j )
     if ( neighbor.flip(j) == i )
       value = invert( value );

    if ( value )
      _hashCode += _s_hashCoefficients.get_coefficient( i );
  }
}


template<class C>
void BV_S_I<C>::print( ostream& output ) const
{
  output << _hashCode << "\n";
}


// stream output functions -----------------------------------------------------
// (all of these just call the respective (virtual) member function)

template <class C>
ostream& operator<<( ostream& output, const BV_S<C>& solution )
{
  solution.print( output );
  return output;
}

template <class C>
ostream& operator<<( ostream& output, const BV_S_N<C>& neighbor )
{
  neighbor.print( output );
  return output;
}

template <class C>
ostream& operator<<( ostream& output, const BV_S_A<C>& attribute )
{
  attribute.print( output );
  return output;
}

template <class C>
ostream& operator<<( ostream& output, const BV_S_I<C>& solutionInfo )
{
  solutionInfo.print( output );
  return output;
}


}
//--- end of namespace HotFrame -----------------------------------------------
//------------------------------------------------------------------------------


#endif
