// Generic_S: (base) classes to manage solutions, moves and attributes of generic problems

#if !defined(__HotFrame_Generic_S__)
#define __HotFrame_Generic_S__

#include "basic.h"


//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
namespace HotFrame
{


// forward declarations

template <class C> class Generic_S;
template <class C> class Generic_S_N;


// ----- class Generic_S -----------------------------------------------------------------
/*
  this class manages solutions (stores, evaluates, provides, ...)
  - needed for all heuristics
*/


template <class C>
class Generic_S
{
  public:
    typedef typename C::T T;
    typedef typename C::Range Range;

  protected:
    T _evaluation;
    Observer<C>* _observer;

  public:
    // assumption: we only have one problem instance to be solved simultaneously
    static T _s_lowerBound;  // _s_lowerBound <= evaluation
    static T _s_upperBound;  // evaluation of best solution found so far

    Generic_S( Observer<C>* observer )
    : _observer( observer )
      { ; }

    virtual ~Generic_S<C>( )
      { ; }

    virtual void evaluate( ) = 0;

    virtual void checkEvaluation( )
    {
      T currentEvaluation = _evaluation;
      evaluate();
      assert( _evaluation == currentEvaluation );
    }

    virtual T f( ) const
      { return _evaluation; }

    static T lowerBound( )
      { return Generic_S<C>::_s_lowerBound; }

    static T upperBound( )
      { return Generic_S<C>::_s_upperBound; }

    Observer<C>* observer( ) const
      { return _observer; }

    virtual void print( ostream& output ) const = 0;

    virtual Generic_S<C>* clone() const = 0; 
    // in a derived, non abstract class D, one usually defines
    // virtual Generic_S<C>* clone() const { return new D<C>(*this); }

    virtual void doMove( Generic_S_N<C>& move ) = 0;

    // actually, this method should be defined const
    // however, for reasons of efficiency we may need to modify *this temporarily
    virtual bool computeEvaluation( const Generic_S_N<C>& move,
                                    T& evaluation, T& delta ) = 0;


    // assignment operators that work for polymorphic classes, a non-trivial story ...
    // program invariant assumption: "assignment operators shall always be called with
    //                                like classes on either side of the equal sign"
    // we follow the following policy: define a virtual assignment operator in the base class
    //   for every derived class: override this virtual assignment operator in every derived class
    //   additionally define a non-virtual assignment operator that just calls the virtual one 
    // see the example given below, we need RTTI (tun-time type information)
    // see R. Gillam: The Assignment Operator Revisited, C++ Report Vol. 10, No. 5 (May 1998), pp. 28
    virtual Generic_S<C>& operator=( const Generic_S<C>& rhs )
    {
      if ( this != &rhs )
      {
        _evaluation = rhs._evaluation;
        _observer = rhs._observer;
        if ( _evaluation < Generic_S<C>::_s_upperBound )
          Generic_S<C>::_s_upperBound = _evaluation;
      }  
      return *this;
    }
    // for every derived class D, define both of the following operators:
    // virtual Generic_S<C>& operator=( const Generic_S<C>& rhs )
    // {
    //   if ( this != &rhs )
    //   {
    //     assert( typeid(rhs) == typeid(*this) );
    //     D& drhs = static_cast<D&>( rhs );
    //     SuperClassOfD<C>::operator=( rhs ); 
    //     ... copy D's member variables using d ...
    //   }  
    //   return *this;
    // }
    // D<C>& operator=( const D<C>& rhs )
    // {
    //   if ( this != &rhs )
    //     operator=( static_cast<const Generic_S<C>&>( rhs ) ); 
    //   return *this;
    // }

    // ... the same concerns should be valid for the equality operator ...
};


template <class C>
typename Generic_S<C>::T Generic_S<C>::_s_upperBound = numeric_limits<T>::max();

template <class C>
typename Generic_S<C>::T Generic_S<C>::_s_lowerBound = -numeric_limits<T>::max();


// the following template function defines a general distance template,
// one must specialize this template function to use respective functionality

template <class S>
float distance( const S& s1, const S& s2 )
{
  return 0;
}


// ----- class Generic_S_N ---------------------------------------------------------------
/*
  this class manages moves (stores, evaluates, provides, ...)
  - needed for all neighborhood based heuristics
*/

template <class C>
class Generic_S_N
{
  public:
    typedef typename C::T  T;
    typedef typename C::Range Range;

  // protected: 
  // temporarily replaced by public, to get a compile of BV_S_NCL (bug ?)
  public:
    T _evaluation;  // evaluation of this move (higher value means better move)
    bool _deltaInformation;
    T _delta;       // improvement of objective function (often: _evaluation == -_delta)
		short _depth;

    virtual void evaluate( ) = 0;

  public:
    Generic_S_N( short depth = 1 )
      : _depth( depth ) { ; }

    virtual ~Generic_S_N<C>( )
      { ; }

    virtual bool isValid( ) const = 0;

    virtual void operator++( ) = 0;

    virtual T operator*( ) const
      { return _evaluation; }

    virtual bool deltaInformation( ) const
      { return _deltaInformation; }

    virtual T delta( ) const
      { 
        if ( deltaInformation( ) )
          return _delta;
        else
          return _evaluation;
      }

    virtual short depth( ) const
      { return _depth; }

    virtual unsigned long size( ) const = 0;  // returns approximate upper bound
                                              // for current neighborhood size

    // may be overwritten by candidate list neighborhoods
    // (some methods may want to know about the activeness of the candidate list)
    //    static bool candidateListActive( )
    //      { return false; }

    virtual void print( ostream& output ) const = 0;
    
    // consider the discussion above for virtual assignment operators ...
    virtual Generic_S_N<C>& operator=( const Generic_S_N<C>& rhs )
    { 
      if ( this != &rhs )
      {
        _evaluation = rhs._evaluation;
        _delta = rhs._delta;
        _deltaInformation = rhs._deltaInformation;
        _depth = rhs._depth;
      }
      return *this;
    }
};


// stream output functions -----------------------------------------------------
// (all of these just call the respective (virtual) member function print)
// (in an ideal world, there should be one template for this, however ...)

template <class C>
ostream& operator<<( ostream& output, const Generic_S<C>& solution )
{
  solution.print( output );
  return output;
}

template <class C>
ostream& operator<<( ostream& output, const Generic_S_N<C>& neighbor )
{
  neighbor.print( output );
  return output;
}



}
//--- end of namespace HotFrame -----------------------------------------------
//------------------------------------------------------------------------------


#endif
