// matrix.h: provides two simple matrix classes Matrix and SymMatrix


#if !defined(__HotFrame_MATRIX__)
#define __HotFrame_MATRIX__


//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
namespace HotFrame
{


// ----------------------------------------------------------------------------
// -- class Matrix ------------------------------------------------------------
// ----------------------------------------------------------------------------

template <class T, class Range>
class Matrix
{
  protected:
    Range _dimension;
    T *_data;  // quadratic matrix, stored row by row

  public:
    Matrix( Range dimension );

    Matrix( const Matrix<T, Range>& original );

    Matrix<T, Range>& operator=( const Matrix<T, Range>& rhs );

    void assign( Range dimension );

    ~Matrix( );

    T operator()( Range i, Range j ) const; // read

    T& operator()( Range i, Range j ); // write

    Range get_dimension( ) const
      { return _dimension; }

    void multiply( T factor );
};

template <class T, class Range>
Matrix<T, Range>::Matrix( Range dimension )
: _dimension( dimension )
{
  if ( dimension == 0 )
  {
    _data = 0;
    _dimension = 0;  
  }
  else
  {
    _data = new T[_dimension * _dimension];
    if ( !_data )
      cerr << "Not enough memory!" << endl, exit( 1 );
  }
}

template <class T, class Range>
Matrix<T, Range>::Matrix( const Matrix<T, Range>& original )
: _dimension( original._dimension )
{
  _data = new T[_dimension*_dimension];
  if ( !_data )
    cerr << "Not enough memory!" << endl, exit( 1 );

  for( Range i=0; i<_dimension*_dimension; ++i )
    _data[i] = original._data[i];
}

template <class T, class Range>
Matrix<T, Range>& Matrix<T, Range>::operator=( const Matrix<T, Range>& rhs )
{
  if ( this != &rhs )
  {
    if ( _dimension == rhs._dimension )
      memcpy( _data, rhs._data, _dimension*_dimension );
    else
    {
      _dimension = rhs._dimension;
      delete[] _data;
      _data = new T[_dimension*_dimension];
      memcpy( _data, rhs._data, _dimension*_dimension );
    }
  }  
  return *this;
}

template <class T, class Range>
void Matrix<T, Range>::assign( Range dimension )
{
  if ( _data != 0 )
    delete[] _data;
  _dimension = dimension;
  _data = new T[_dimension * _dimension];
  if ( !_data )
    cerr << "Not enough memory!" << endl, exit( 1 );
}

template <class T, class Range>
Matrix<T, Range>::~Matrix( )
{
  delete[] _data;
}

template <class T, class Range>
inline
T Matrix<T, Range>::operator()( Range i, Range j ) const
{
  assert( i >= 1 ); assert( i <= _dimension );
  assert( j >= 1 ); assert( j <= _dimension );

  return _data[(i-1)*_dimension+j-1];
}

template <class T, class Range>
inline
T& Matrix<T, Range>::operator()( Range i, Range j )
{
  assert( i >= 1 ); assert( i <= _dimension );
  assert( j >= 1 ); assert( j <= _dimension );

  return _data[(i-1)*_dimension+j-1];
}

template <class T, class Range>
void Matrix<T, Range>::multiply( T factor )
{
  for( Range i=1; i<=get_dimension(); ++i )
    for( Range j=1; j<=get_dimension(); ++j )
      _data[(i-1)*_dimension+j-1] = factor * _data[(i-1)*_dimension+j-1];
}

template <class T, class Range>
ostream& operator<<( ostream& output, const Matrix<T, Range>& matrix )
{
  // format: dimension
  //         a_1,1 a_1,2 ...
  //         a_2,1 a_2,2 ...
  //         ..
  //         .. .. a_dimension,dimension

  output << matrix.get_dimension() << "\n";

  for( Range i=1; i<=matrix.get_dimension(); ++i )
  {
    for( Range j=1; j<=matrix.get_dimension(); ++j )
      output << matrix( i, j ) << " ";

    output << "\n";
  }

  return output;
}

template <class T, class Range>
istream& operator>>( istream& input, Matrix<T, Range>& matrix )
{
  // format: dimension
  //         a_1,1 a_1,2 ...
  //         a_2,1 a_2,2 ...
  //         ..
  //         .. .. a_dimension,dimension

  Range dim; 
  input >> dim;
  assert( dim == matrix.get_dimension() );

  for( Range i=1; i<=matrix.get_dimension(); ++i )
    for( Range j=1; j<=matrix.get_dimension(); ++j )
    {
      T data;
      input >> data;
      matrix( i, j ) = data;
    }

  return input;
}


// ----------------------------------------------------------------------------
// -- class SymMatrix ---------------------------------------------------------
// ----------------------------------------------------------------------------

template <class T, class Range>
class SymMatrix
{
  protected:
    Range _dimension;
    T *_data;  // lower left triangular matrix (i.e. i>=j)

  public:
    SymMatrix( Range dimension );

    SymMatrix( const SymMatrix<T, Range>& original );

    SymMatrix<T, Range>& operator=( const SymMatrix<T, Range>& rhs );

    void assign( Range dimension );

    ~SymMatrix( );

    T operator()( Range i, Range j ) const; // read

    T& operator()( Range i, Range j ); // write

    Range get_dimension( )  const
      { return _dimension; }

    void multiply( T factor );
};

template <class T, class Range>
SymMatrix<T, Range>::SymMatrix( Range dimension )
: _dimension( dimension )
{
  if ( dimension == 0 )
    _data = 0;
  else
  {
    _data = new T[(_dimension*(_dimension+1))/2];
    if ( !_data )
      cerr << "Not enough memory!" << endl, exit( 1 );
  }
}

template <class T, class Range>
SymMatrix<T, Range>::SymMatrix( const SymMatrix<T, Range>& original )
: _dimension( original._dimension )
{
  _data = new T[(_dimension*(_dimension+1))/2];
  if ( !_data )
    cerr << "Not enough memory!" << endl, exit( 1 );

  for( Range i=0; i<(_dimension*(_dimension+1))/2; ++i )
    _data[i] = original._data[i];
}

template <class T, class Range>
SymMatrix<T, Range>& SymMatrix<T, Range>::operator=( const SymMatrix<T, Range>& rhs )
{
  if ( this != &rhs )
  {
    if ( _dimension == rhs._dimension )
      memcpy( _data, rhs._data, (_dimension*(_dimension+1))/2 );
    else
    {
      _dimension = rhs._dimension;
      delete[] _data;
      _data = new T[(_dimension*(_dimension+1))/2];
      memcpy( _data, rhs._data, (_dimension*(_dimension+1))/2 );
    }
  }  
  return *this;
}

template <class T, class Range>
void SymMatrix<T, Range>::assign( Range dimension )
{
  if ( _data != 0 )
    delete[] _data;

  _dimension = dimension;
  
  _data = new T[(_dimension*(_dimension+1))/2];
  if ( !_data )
    cerr << "Not enough memory!" << endl, exit( 1 );
}

template <class T, class Range>
SymMatrix<T, Range>::~SymMatrix( )
{
  delete[] _data;
}

template <class T, class Range>
inline
T SymMatrix<T, Range>::operator()( Range i, Range j ) const
{
  assert( i >= 1 ); assert( i <= _dimension );
  assert( j >= 1 ); assert( j <= _dimension );

  if ( i < j )
    swap( i, j );
  return _data[i*(i-1)/2+j-1];
}

template <class T, class Range>
inline
T& SymMatrix<T, Range>::operator()( Range i, Range j )
{
  assert( i >= 1 ); assert( i <= _dimension );
  assert( j >= 1 ); assert( j <= _dimension );

  if ( i < j )
    swap( i, j );
  return _data[i*(i-1)/2+j-1];
}

template <class T, class Range>
void SymMatrix<T, Range>::multiply( T factor )
{
  for( Range i=1; i<=get_dimension(); ++i )
    for( Range j=1; j<=i; ++j )
      _data[i*(i-1)/2+j-1] = factor * _data[i*(i-1)/2+j-1];
}

template <class T, class Range>
ostream& operator<<( ostream& output, const SymMatrix<T, Range>& matrix )
{
  // format: dimension
  //         a_1,1
  //         a_2,1 a_2,2
  //         a_2,1 a_2,2 a_2,3
  //         ..
  //         .. .. a_dimension,dimension

  output << matrix.get_dimension() << "\n";

  for( Range i=1; i<=matrix.get_dimension(); ++i )
  {
    for( Range j=1; j<=i; ++j )
      output << matrix( i, j ) << " ";

    output << "\n";
  }


  return output;
}

template <class T, class Range>
istream& operator>>( istream& input, SymMatrix<T, Range>& matrix )
{
  // format: dimension
  //         a_1,1
  //         a_2,1 a_2,2
  //         a_2,1 a_2,2 a_2,3
  //         ..
  //         .. .. a_dimension,dimension

  Range dim; 
  input >> dim;
  assert( dim == matrix.get_dimension() );

  for( Range i=1; i<=matrix.get_dimension(); ++i )
    for( Range j=1; j<=i; ++j )
    {
      T data;
      input >> data;
      matrix( i, j ) = data;
    }

  return input;
}



}
//--- end of namespace HotFrame -----------------------------------------------
//------------------------------------------------------------------------------


#endif
