// observer classes

#if !defined(__HotFrame_OBSERVER__)
#define __HotFrame_OBSERVER__

#include "basic.h"


//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
namespace HotFrame
{

// ----------------------------------------------------------------------------
// -- base class Observer -----------------------------------------------------
// ----------------------------------------------------------------------------

// do nothing

template <class C>
class Observer
{
  public:
    typedef typename C::T T;
    typedef typename C::Range Range;

  protected:
    bool _active;
    ostringstream _ostringstream;

    double _timeWhenNewUpperBoundWasFound;

  public:
    Observer( ) : _active( false ), _timeWhenNewUpperBoundWasFound( 0 )  { ; }
    virtual ~Observer( )  { ; }

    bool active( ) const { return _active; }
    void activate( ) { _active=true; }
    void deactivate( ) { _active=false; }

    virtual ostream& getStream( )  { return _ostringstream; }
    virtual void flush( )  { _ostringstream.seekp( 0 ); }

    virtual void distance( float distance ) { ; }
    virtual void cpuTimeOfProcess( double cpuTime )  { ; }  
    virtual void numberOfIterations( unsigned long iterations ) { ; }

    virtual void timeWhenNewUpperBoundWasFound( double cpuTime )
      { _timeWhenNewUpperBoundWasFound = cpuTime; }  

    virtual void about( )  { ; }
    virtual void start( const string& info )  { ; }
    virtual void comment( const string& info )  { ; }

    virtual void lowerBound( T eval )  { ; }

    virtual void currentProblem( )  { ; }
    virtual void problemEPS( string name )  { ; }

    virtual void end( const string& info )  { ; }
    virtual void searchSuspended( const string& info )  { ; }
    virtual void iteration( unsigned long it )  { ; }
    virtual void move( )  { ; }
    virtual void moveEvaluation( T eval )  { ; }
    virtual void deltaEvaluation( T eval )  { ; }
    virtual void newSolutionWithEvaluation( T eval )  { ; }
    virtual void newUpperBound( T eval )  { ; }
    virtual void bestEvaluation( T eval )  { ; }
    virtual void currentSolution( )  { ; }
    virtual void currentNeighbor( )  { ; }
    virtual void averageMoveEvaluation( )  { ; }

    virtual void improvingSAMove( T eval )  { ; }
    virtual void deterioratingSAMove( T eval )  { ; }
    virtual void newTemperature( double temperature )  { ; }
    virtual void newMaxRepetitions( double maxRepetitions )  { ; }

    virtual void improvingTAMove( T eval )  { ; }
    virtual void deterioratingTAMove( T eval )  { ; }
    virtual void newThreshold( double threshold )  { ; }
    
    virtual void tabuCriterionInitialized( const string& info ) { ; }
    virtual void aspirationCriterionSatisfied( )  { ; }

    virtual void attributeAddedToHistory( )  { ; }
    virtual void solutionInfoAddedToHistory( )  { ; }
    virtual void currentTabuList( )  { ; }
    virtual void currentTrajectory( )  { ; }
    virtual void noAdmissibleNeighbor( const string& info )  { ; }
    virtual void escape( const string& info )  { ; }
    virtual void tabuListLengthExtended( Range newLength )  { ; }
    virtual void tabuListLengthShortened( Range newLength )  { ; }
    virtual void newMovingAverage( unsigned long movingAverage ) { ; }

    virtual void newNeighborhoodDepth( short lambda )  { ; }
};


// ----------------------------------------------------------------------------
// -- class StreamingObserver -------------------------------------------------
// ----------------------------------------------------------------------------

// just pipe all data to an output stream

template <class C>
class StreamingObserver
: public Observer<C>
{
  public:
    typedef typename C::T T;
    typedef typename C::Range Range;

  protected:
    ostream& _output;

  public:
    StreamingObserver( ostream& output )
      : _output( output )  { _active = true; }

    virtual ostream& getStream( )  { return _output; }

    virtual void flush( )
      { 
         _output << _ostringstream.str();
         _ostringstream.str( "" );
      }

    virtual void distance( float distance )
      { _output << "distance: " << distance << "\n"; };

    virtual void cpuTimeOfProcess( double cpuTime )
      { _output << "cpu time: " << cpuTime
                << "  best time: " << _timeWhenNewUpperBoundWasFound << "\n"; };

    virtual void about( )
      { _output << "StreamingObserver starts streaming ..." << "\n"; };

    virtual void start( const string& info )
      { _output << "start of: " << info << "\n"; };

    virtual void comment( const string& info )
      { _output << info << "\n"; };

    virtual void lowerBound( T eval )
      { _output << "lower bound: " << eval << "\n"; };

    virtual void currentProblem( )
      { _output << "current problem: " << "\n"; };

    virtual void end( const string& info )
      { _output << "end: " << info << "\n"; };

    virtual void searchSuspended( const string& info )
      { _output << "search suspended: " << info << "\n"; };

    virtual void iteration( unsigned long it )
      { _output << "iteration: " << it << "\n"; };

    virtual void move( )
      { _output << "move: " << "\n"; };

    virtual void moveEvaluation( T eval )
      { _output << "move evaluation: " << eval << "\n"; };

    virtual void deltaEvaluation( T eval )
      { _output << "delta of solution evaluation: " << eval << "\n"; };

    virtual void newSolutionWithEvaluation( T eval )
      { _output << "new solution with evaluation: " << eval << "\n"; };

    virtual void newUpperBound( T eval )
      { _output << "new upper bound: " << eval << "\n"; };

    virtual void bestEvaluation( T eval )
      { _output << "best evaluation: " << eval << "\n"; };

    virtual void currentSolution( )
      { _output << "current solution: " << "\n"; };

    virtual void currentNeighbor( )
      { _output << "current neighbor: " << "\n"; };

    virtual void improvingSAMove( T eval )
      { _output << "improving SA move: " << eval << "\n"; };

    virtual void deterioratingSAMove( T eval )
      { _output << "deteriorating SA move: " << eval << "\n"; };

    virtual void newTemperature( double temperature )
      { _output << "new temperature: " << temperature << "\n"; };

    virtual void newMaxRepetitions( double maxRepetitions )
      { _output << "new maxRepetitions: " << maxRepetitions << "\n"; };


    virtual void improvingTAMove( T eval )
      { _output << "improving TA move: " << eval << "\n"; };

    virtual void deterioratingTAMove( T eval )
      { _output << "deteriorating TA move: " << eval << "\n"; };

    virtual void newThreshold( double threshold )
      { _output << "new threshold: " << threshold << "\n"; };


    virtual void tabuCriterionInitialized( const string& info )
      { _output << "tabu criterion initialized: " << info << "\n"; };

    virtual void aspirationCriterionSatisfied(  )
      { _output << "aspiration criterion satisfied" << "\n"; };

    virtual void attributeAddedToHistory( )
      { _output << "attribute added to history: " << "\n"; };

    virtual void solutionInfoAddedToHistory( )
      { _output << "solution info added to history: " << "\n"; };

    virtual void currentTabuList( ) 
      { _output << "current tabu list: " << "\n"; };

    virtual void currentTrajectory( ) 
      { _output << "current trajectory: " << "\n"; };

    virtual void noAdmissibleNeighbor( const string& info )
      { _output << "no admissible neighbor: " << info << "\n"; };

    virtual void escape( const string& info ) 
      { _output << "escape (" << info << ")" << "\n"; };

    virtual void tabuListLengthExtended( Range newLength )
      { _output << "tabu list length extended to: " << newLength << "\n"; };

    virtual void tabuListLengthShortened( Range newLength )
      { _output << "tabu list length shortend to: " << newLength << "\n"; };

    virtual void newMovingAverage( unsigned long movingAverage )
      { _output << "new moving average: " << movingAverage << "\n"; }

    virtual void newNeighborhoodDepth( short lambda )
      { _output << "new neighborhood depth: " << lambda << "\n"; }
};



// ----------------------------------------------------------------------------
// -- class JustTheBestEvaluationToStream -------------------------------------
// ----------------------------------------------------------------------------

// just pipe the best evaluation to an output stream

template <class C>
class JustTheBestEvaluationToStream
: public Observer<C>
{
  public:
    typedef typename C::T T;
    typedef typename C::Range Range;

  protected:
    ostream& _output;

  public:
    JustTheBestEvaluationToStream( ostream& output )
      : _output( output )  { };

  virtual void bestEvaluation( T eval )
    { _output << eval << "\n"; };
};


// ----------------------------------------------------------------------------
// -- class JustTheBestEvaluationAndTheCPUTimeToStream ------------------------
// ----------------------------------------------------------------------------

// just pipe the best evaluation <tab> and cpu time <"\n"> to an output stream

template <class C>
class JustTheBestEvaluationAndTheCPUTimeToStream
: public Observer<C>
{
  public:
    typedef typename C::T T;
    typedef typename C::Range Range;

  protected:
    ostream& _output;

  public:
    JustTheBestEvaluationAndTheCPUTimeToStream( ostream& output )
      : _output( output )  { _active = true ; };

  virtual void bestEvaluation( T eval )
    { _output << eval << "\t"; };

  virtual void cpuTimeOfProcess( double cpuTime )
//    { _output << cpuTime << "\n"; };
    { _output << "cpu time: " << cpuTime
              << "  best time: " << _timeWhenNewUpperBoundWasFound << "\n"; };
};


// ----------------------------------------------------------------------------
// -- class JustTheBestEvaluationAndTheNumberOfIterationsToStream -------------
// ----------------------------------------------------------------------------

// just pipe the best evaluation <tab> and the number of iterations <"\n">
// to an output stream

template <class C>
class JustTheBestEvaluationAndTheNumberOfIterationsToStream
: public Observer<C>
{
  public:
    typedef typename C::T T;
    typedef typename C::Range Range;

  protected:
    ostream& _output;
    unsigned long _iterations;

  public:
    JustTheBestEvaluationAndTheNumberOfIterationsToStream( ostream& output )
      : _output( output )  { _active = true ; };

  virtual void numberOfIterations( unsigned long iterations )
    { _iterations = iterations; }

  virtual void bestEvaluation( T eval )
    { _output << eval << "\t" << _iterations << "\n"; };
};


// ----------------------------------------------------------------------------
// -- class JustTheBestEvaluationAndTheNumberOfIterationsAndTheCPUTimeToStream 
// ----------------------------------------------------------------------------

template <class C>
class JustTheBestEvaluationAndTheNumberOfIterationsAndTheCPUTimeToStream
: public Observer<C>
{
  public:
    typedef typename C::T T;
    typedef typename C::Range Range;

  protected:
    ostream& _output;

  public:
    JustTheBestEvaluationAndTheNumberOfIterationsAndTheCPUTimeToStream( ostream& output )
      : _output( output )  { _active = true; }

  virtual void numberOfIterations( unsigned long iterations )
    { _output << iterations << "\t"; };

  virtual void bestEvaluation( T eval )
    { _output << eval << "\t"; };

  virtual void cpuTimeOfProcess( double cpuTime )
    { _output << "cpu time: " << cpuTime
              << "  best time: " << _timeWhenNewUpperBoundWasFound << "\n"; };
};


// ----------------------------------------------------------------------------
// -- class JustLowerBoundToStream --------------------------------------------
// ----------------------------------------------------------------------------

// just pipe the best evaluation to an output stream

template <class C>
class JustLowerBoundToStream
: public Observer<C>
{
  public:
    typedef typename C::T T;
    typedef typename C::Range Range;

  protected:
    ostream& _output;

  public:
    JustLowerBoundToStream( ostream& output )
      : Observer<C>(), _output( output ) { ; }

  virtual void lowerBound( T eval )
    { _output << eval << "\n"; };
};


// ----------------------------------------------------------------------------
// -- class MoveRangeObserver -------------------------------------------------
// ----------------------------------------------------------------------------

// just pipe the best evaluation to an output stream

template <class C>
class MoveRangeObserver
: public Observer<C>
{
  public:
    typedef typename C::T T;
    typedef typename C::Range Range;

  protected:
    ostream& _output;
    T accumulatedMoveEvaluations;
    Range numberOfMoveEvaluations;

  public:
    MoveRangeObserver( ostream& output )
      : Observer<C>(), _output( output ),
        accumulatedMoveEvaluations( 0 ), numberOfMoveEvaluations( 0 )
      { ; }

  void moveEvaluation( T eval )
    { 
      accumulatedMoveEvaluations += abs( eval );
      ++numberOfMoveEvaluations;
    };

  void averageMoveEvaluation( )
    { 
      if ( numberOfMoveEvaluations == 0 )
        _output << 0 << "\n";
      else
        _output << accumulatedMoveEvaluations / double( numberOfMoveEvaluations ) << "\n";
    };
};




// ----------------------------------------------------------------------------
// -- Statistics Observer -----------------------------------------------------
// ----------------------------------------------------------------------------

template <class T>
class StatisticsObserver
{
  protected:
    int _p;  // dimension of a data point
    int _n;  // number of data points
    vector< vector<T> > _data;

  public:
    StatisticsObserver( int p )
      : _p( p ), _n( 0 )  { ; }
    virtual ~StatisticsObserver( )  { ; }

    virtual void newDataPoint( )
      { _data.push_back( vector<T>( _p, 0 ) ); ++_n; }
    
    virtual void setDataPoint( int i, T z )
      { assert( ( i >= 1 ) && ( i <= _p ) );
        _data[_n-1][i-1] = z;
      }
    
    virtual void end( )  { ; }
};


// ----------------------------------------------------------------------------
// -- StreamingStatisticsObserver ---------------------------------------------
// ----------------------------------------------------------------------------

// pipe all data to an output stream at the end

template <class T>
class StreamingStatisticsObserver
: public StatisticsObserver<T>
{
  protected:
    ostream& _output;

  public:
    StreamingStatisticsObserver( int p, ostream& output )
      : StatisticsObserver<T>( p ), _output( output )  { ; }
    virtual ~StreamingStatisticsObserver( )  { ; }

    virtual void end( );
};

template <class T>
void StreamingStatisticsObserver<T>::end( )
{
  for( int i=1; i<=_n; ++i )
  {
    for( int j=1; j<=_p; ++j )
      _output << "\t" << _data[i-1][j-1];
    _output << "\n";
  }
}



}
//--- end of namespace HotFrame -----------------------------------------------
//------------------------------------------------------------------------------


#endif
