// Perm_PS: class to store partial solutions with respect to the permutation solution space

#if !defined(__HotFrame_Perm_PS__)
#define __HotFrame_Perm_PS__

#include "basic.h"
#include "generic_s.h"
#include "observer.h"


//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
namespace HotFrame
{


// declaration of class Perm_PS ------------------------------------------------


template <class C>
class Perm_PS
: public Generic_S<C>
{
  public:
    typedef typename C::T T;
    typedef typename C::Range Range;

  protected:
    vector<Range> _perm; // _perm[1],...,_perm[_n]
    vector<Range> _remaining;

  public:
    enum FirstSolution { Empty=0, OneRandomJob, StartWithGivenSolution };

    Perm_PS( );

    Perm_PS( Observer<C>* observer,
             Range nFinal = 0,
             FirstSolution firstSolution = Empty,
             vector<Range> startPermutation = vector<Range>() );

    virtual void print( ostream& output ) const;

    virtual void doMove( Generic_S_N<C>& move );

    virtual bool computeEvaluation( const Generic_S_N<C>& move,
                                    T& evaluation, T& delta );

    Range getObjectAtPosition( Range i ) const
      { assert( (i>=0) && (i<=_perm.size()-1) );
        if (i==0) return 0;
        else return _perm[i];
      }

    Range getRemainingObjectAtPosition( Range i ) const
      { assert( (i>=0) && (i<=_remaining.size()-1) );
        if (i==0) return 0;
        else return _remaining[i];
      }

    Range n( ) const
      { return _perm.size()-1; }

    const vector<Range>& permutation( ) const
      { return _perm; }

    vector<Range>& permutation( )
      { return _perm; }

    Range nRemaining( ) const
      { return _remaining.size()-1; }

    const vector<Range>& remaining( ) const
      { return _remaining; }

    vector<Range>& remaining( )
      { return _remaining; }

    virtual Generic_S<C>& operator=( const Generic_S<C>& rhs )
    {
      if ( this != &rhs )
      {
        assert( typeid(rhs) == typeid(*this) );
        const Perm_PS<C>& drhs = static_cast<const Perm_PS<C>&>( rhs );
        Generic_S<C>::operator=( rhs ); 
        _perm = drhs._perm;
        _remaining = drhs._remaining;
			}  
      return *this;
    }
    
    Perm_PS<C>& operator=( const Perm_PS<C>& rhs )
    {
      if ( this != &rhs )
        operator=( static_cast<const Generic_S<C>&>( rhs ) ); 
      return *this;
    }
};



// ----- class Perm_PS_N -------------------------------------------------------
/*
  abstract base class for neighbors of Perm_PS solutions
*/

template <class C>
class Perm_PS_N
: public Generic_S_N<C>
{
  public:
    typedef typename C::T T;
    typedef typename C::Range Range;

  protected:
    Perm_PS<C> *_s;

  public:
    // constructor
    Perm_PS_N( Perm_PS<C> *s, short depth = 1 )
      : Generic_S_N<C>( depth ), _s( s )
    { ; }

    Perm_PS_N( const Perm_PS_N& original )
      : Generic_S_N<C>( original ), _s( original._s )
      { assert( _depth==1 ); }

    virtual void transform( vector<Range>& perm, vector<Range>& remaining ) const = 0;

    Perm_PS<C> * currentSolution( ) const
  		{ return _s; }

    virtual Generic_S_N<C>& operator=( const Generic_S_N<C>& rhs )
    {
      if ( this != &rhs )
      {
        assert( typeid(rhs) == typeid(*this) );
        const Perm_PS_N<C>& drhs = static_cast<const Perm_PS_N<C>&>( rhs );
        Generic_S_N<C>::operator=( rhs ); 
        _s = drhs._s;
      }  
      return *this;
    }

    Perm_PS_N<C>& operator=( const Perm_PS_N<C>& rhs )
    {
      if ( this != &rhs )
        operator=( static_cast<const Generic_S_N<C>&>( rhs ) ); 
      return *this;
    }
};


// ----- class Perm_PS_N_Insertion --------------------------------------------
/*
  this class manages insertion neighbors for Perm_PS solutions
*/

template <class C>
class Perm_PS_N_Insertion
: public Perm_PS_N<C>
{
  public:
    typedef typename C::T T;
    typedef typename C::Range Range;

  protected:
    Range _object;
    Range _posInsertion;  // insert _object before position _posInsertion in _s->_perm
    Range _posRemaining;

    virtual void evaluate();

  public:
    // constructor
    Perm_PS_N_Insertion( Perm_PS<C> *s,
 		                     NeighborhoodPosition position = FirstNeighbor,
                         short depth = 1 ); // so far: only implemented for depth = 1

    Perm_PS_N_Insertion( const Perm_PS_N_Insertion& original )
      : Perm_PS_N<C>( original ), _object( original._object ),
        _posInsertion( original._posInsertion ), _posRemaining( original._posRemaining )
      { assert( _depth==1 ); }

    virtual Generic_S_N<C>* clone() const
      { return new Perm_PS_N_Insertion<C>( *this ); }

    virtual bool isValid( ) const
      { return ( _object != 0 ); }

    virtual void operator++( );

    virtual unsigned long size( ) const
      { return _s->nRemaining()*( _s->n()+1 ); }

    virtual void transform( vector<Range>& perm, vector<Range>& remaining ) const;

    Range object( ) const
      { return _object; }

    Range posInsertion( ) const
      { return _posInsertion; }

    Range posRemaining( ) const
      { return _posRemaining; }

    Range actualDepth( ) const
      { return 1; }

    virtual void print( ostream& output ) const;

    virtual Generic_S_N<C>& operator=( const Generic_S_N<C>& rhs )
    {
      if ( this != &rhs )
      {
        assert( typeid(rhs) == typeid(*this) );
        const Perm_PS_N_Insertion<C>& drhs = static_cast<const Perm_PS_N_Insertion<C>&>( rhs );
        Perm_PS_N<C>::operator=( rhs ); 
        _object = drhs._object;
        _posInsertion = drhs._posInsertion;
        _posRemaining = drhs._posRemaining;
      }  
      return *this;
    }

    Perm_PS_N_Insertion<C>& operator=( const Perm_PS_N_Insertion<C>& rhs )
    {
      if ( this != &rhs )
        operator=( static_cast<const Generic_S_N<C>&>( rhs ) ); 
      return *this;
    }
};


// ----- class Perm_PS_N_Append --------------------------------------------
/*
  this class manages append moves for Perm_PS solutions
*/

template <class C>
class Perm_PS_N_Append
: public Perm_PS_N<C>
{
  public:
    typedef typename C::T T;
    typedef typename C::Range Range;

  protected:
    Range _object;
    Range _posRemaining;
    // append _object (at _posRemaining in _s->_remaining) to _s->_perm

    virtual void evaluate();

  public:
    // constructor
    Perm_PS_N_Append( Perm_PS<C> *s,
 		                  NeighborhoodPosition position = FirstNeighbor,
                      short depth = 1 ); // so far: only implemented for depth = 1

    Perm_PS_N_Append( const Perm_PS_N_Append& original )
      : Perm_PS_N<C>( original ),
        _object( original._object ), _posRemaining( original._posRemaining )
      { assert( _depth==1 ); }

    virtual Generic_S_N<C>* clone() const
      { return new Perm_PS_N_Append<C>( *this ); }

    virtual bool isValid( ) const
      { return ( _object != 0 ); }

    virtual void operator++( );

    virtual unsigned long size( ) const
      { return currentSolution()->nRemaining(); }

    virtual void transform( vector<Range>& perm, vector<Range>& remaining ) const;

    Range object( ) const
      { return _object; }

    Range posRemaining( ) const
      { return _posRemaining; }

    Range actualDepth( ) const
      { return 1; }

    virtual void print( ostream& output ) const;

    virtual Generic_S_N<C>& operator=( const Generic_S_N<C>& rhs )
    {
      if ( this != &rhs )
      {
        assert( typeid(rhs) == typeid(*this) );
        const Perm_PS_N_Append<C>& drhs = static_cast<const Perm_PS_N_Append<C>&>( rhs );
        Perm_PS_N<C>::operator=( rhs ); 
        _object = drhs._object;
        _posRemaining = drhs._posRemaining;
      }  
      return *this;
    }

    Perm_PS_N_Append<C>& operator=( const Perm_PS_N_Append<C>& rhs )
    {
      if ( this != &rhs )
        operator=( static_cast<const Generic_S_N<C>&>( rhs ) ); 
      return *this;
    }
};


// element functions of class CFS_S<C> ---------------------------------


template<class C>
Perm_PS<C>::Perm_PS( Observer<C>* observer,
                     Range nFinal,  // = 0
                     FirstSolution firstSolution, // = Empty
                     vector<Range> startPermutation )  // = vector<Range>()
: Generic_S<C>( observer ),             
  _perm( 1, 0 ),  // only one dummy object included
  _remaining( nFinal+1, 0 )
{
  for( Range i=1; i<=nFinal; ++i )
    _remaining[i] = i;
  
  if ( firstSolution == Empty )
  {
    ;  // nothing to do
  } 
  else if ( firstSolution == OneRandomJob )
  {
    assert( nFinal >= 1 );  // otherwise, this makes no sense
    Range randomJob = rand( nFinal ) + 1;
    _perm.push_back( randomJob );
    _remaining.erase( _remaining.begin() + randomJob );
  } 
  else if ( firstSolution == StartWithGivenSolution )
  {
    _perm = startPermutation;
    
    if ( nFinal >= 1 )
    // erase those objects from remaining that are in _perm
    {
      vector<Range>::iterator objectIt;
      for( Range i=1; i<=_perm.size()-1; ++i )
      {
        objectIt = find( _remaining.begin()+1, _remaining.end(), _perm[i] );
        if ( objectIt != _remaining.end() )
          _remaining.erase( objectIt );
      }
    }
  }
  else
    assert( false );
}


template<class C>
bool Perm_PS<C>::computeEvaluation( const Generic_S_N<C>& move,
                                    T& evaluation, T& delta )
{
  // this is an inefficient default evaluation, which only works for Perm_PS_N moves;
  // this should be redefined in an derived class to efficiently evaluate a neighhbor adaptively

  // this must be redefined when there are special data fields in the derived
  // solution class, which are affected by the evaluate()

  // dynamic downcast

  const Perm_PS_N<C>& specificMove = dynamic_cast<const Perm_PS_N<C>&>( move );

  T oldEvaluation = _evaluation;

  vector<Range> originalPermutation( _perm );
  vector<Range> originalRemaining( _remaining );
  specificMove.transform( _perm, _remaining );
    
  evaluate( );

  delta = _evaluation - oldEvaluation;
  evaluation = -delta;

  // restore current solution (*this)
  _perm = originalPermutation;
  _remaining = originalRemaining;

  _evaluation = oldEvaluation;

  return true;
}


template<class C>
void Perm_PS<C>::doMove( Generic_S_N<C>& move )
{
  // handling of Perm_PS_N moves
  {
    const Perm_PS_N<C> *specificMove = dynamic_cast<const Perm_PS_N<C> *>(&move);
    if ( specificMove )
    {
      if ( _observer->active() )
      {
        _observer->move();
        _observer->getStream() << *specificMove; 
        _observer->flush();
        _observer->moveEvaluation( **specificMove ); 
        if ( specificMove->deltaInformation() )
          _observer->deltaEvaluation( specificMove->delta() ); 
      }

      specificMove->transform( _perm, _remaining );

      if ( specificMove->deltaInformation() )
        _evaluation += specificMove->delta(); 
      else
        evaluate();

      return;
    }
  }

  assert( false );  // move not handled
}


template<class C>
void Perm_PS<C>::print( ostream& output ) const
{
  // format: list the permutation (separated by spaces) 0
  //         evaluation

  for( Range i=1; i<=n(); ++i )
    output << getObjectAtPosition(i) << " ";

  output << "\n" << _evaluation << "\n";
}


// element functions of class Perm_PS_N_Insertion<C> ---------------------------

template<class C>
Perm_PS_N_Insertion<C>::Perm_PS_N_Insertion(
                        Perm_PS<C> *s,
	                      NeighborhoodPosition position, // = FirstNeighbor,
                        short depth ) // =1; so far: only implemented for depth = 1
  : Perm_PS_N<C>( s, depth )
{
  assert ( depth == 1 );

  if ( ( position == InvalidNeighbor ) || ( _s->nRemaining() == 0 ) )
  {
    _object = 0;
    return;
  }

  // else
  if ( position == SomeRandomNeighbor )
  {
    _posInsertion = rand( _s->n() + 1 ) + 1;
    _posRemaining = rand( _s->nRemaining() ) + 1;
  }
  else
  {
    _posInsertion = 1;
    _posRemaining = 1;
  }

  _object = _s->getRemainingObjectAtPosition( _posRemaining );
  evaluate();
}


template<class C>
void Perm_PS_N_Insertion<C>::evaluate()
{
  _deltaInformation =
    _s->computeEvaluation( *this, _evaluation, _delta );
}


template<class C>
void Perm_PS_N_Insertion<C>::operator++( )
{
  if ( _depth >= 2 )
  {
    assert( false );  // not yet done
  }
  else
  {
    if ( _posInsertion <= _s->n() )
      ++_posInsertion;
    else
    {
      if ( _posRemaining < _s->nRemaining() )
      {
        ++_posRemaining;
        _object = _s->getRemainingObjectAtPosition( _posRemaining );
        _posInsertion = 1;
      }
      else
        _object = 0;
    }

    if ( _object != 0 )
      evaluate();
  }
}


template<class C>
void Perm_PS_N_Insertion<C>::transform( vector<Range>& perm, vector<Range>& remaining ) const
{
  perm.insert( perm.begin() + _posInsertion, _object );

  if ( _posRemaining > 0 )
    remaining.erase( remaining.begin() + _posRemaining );
}


template<class C>
void Perm_PS_N_Insertion<C>::print( ostream& output ) const
{
  output << "insert object " << _object
         << " before position " << _posInsertion << "\n";
}


// element functions of class Perm_PS_N_Append<C> ---------------------------

template<class C>
Perm_PS_N_Append<C>::Perm_PS_N_Append(
                     Perm_PS<C> *s,
 		                 NeighborhoodPosition position, // = FirstNeighbor,
                     short depth ) // =1; so far: only implemented for depth = 1
  : Perm_PS_N<C>( s, depth )
{
  assert ( depth == 1 );

  if ( ( position == InvalidNeighbor ) || ( _s->nRemaining() == 0 ) )
  {
    _object = 0;
    return;
  }

  // else
  if ( position == SomeRandomNeighbor )
    _posRemaining = rand( _s->nRemaining() ) + 1;
  else
    _posRemaining = 1;

  _object = _s->getRemainingObjectAtPosition( _posRemaining );
  evaluate();
}


template<class C>
void Perm_PS_N_Append<C>::evaluate()
{
  _deltaInformation =
    _s->computeEvaluation( *this, _evaluation, _delta );
}


template<class C>
void Perm_PS_N_Append<C>::operator++( )
{
  if ( _depth >= 2 )
  {
    assert( false );  // not yet done
  }
  else
  {
    if ( _posRemaining < _s->nRemaining() )
    {
      ++_posRemaining;
      _object = _s->getRemainingObjectAtPosition( _posRemaining );
    }
    else
      _object = 0;
  }

  if ( _object != 0 )
    evaluate();
}


template<class C>
void Perm_PS_N_Append<C>::transform( vector<Range>& perm, vector<Range>& remaining ) const
{
  perm.push_back( _object );

  if ( _posRemaining > 0 )
    remaining.erase( remaining.begin() + _posRemaining );
}


template<class C>
void Perm_PS_N_Append<C>::print( ostream& output ) const
{
  output << "append object " << _object << "\n";
}



// stream output functions -----------------------------------------------------

template <class C>
ostream& operator<<( ostream& output, const Perm_PS<C>& solution )
{
  solution.print( output );
  return output;
}

template <class C>
ostream& operator<<( ostream& output, const Perm_PS_N_Insertion<C>& move )
{
  move.print( output );
  return output;
}

template <class C>
ostream& operator<<( ostream& output, const Perm_PS_N_Append<C>& move )
{
  move.print( output );
  return output;
}


}
//--- end of namespace HotFrame ------------------------------------------------
//------------------------------------------------------------------------------


#endif