// Perm_S: classes to manage solutions, moves and attributes of generic permutation problems

#if !defined(__HotFrame_Perm_S__)
#define __HotFrame_Perm_S__

#include "basic.h"
#include "generic_s.h"


//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
namespace HotFrame
{


// forward declarations
template <class C> class Perm_S;
template <class C> class Perm_S_N_Shift;
template <class C> class Perm_S_N_Swap;
template <class C> class Perm_S_A;
template <class C> class Perm_S_A_Pos;
template <class C> class Perm_S_I;


// ----- class Perm_S -----------------------------------------------------------------
/*
  this class manages solutions represented by permutations (stores, evaluates, provides, ...)
  - needed for all heuristics
*/


template <class C>
class Perm_S 
: public Generic_S<C>
{
  public:
    typedef typename C::T T;
    typedef typename C::Range Range;

    enum FirstSolution { Identity=0, Random };

  protected:
    Range _n;  // client view of object numbering: 1,...,n
    vector<Range> _perm; // _perm[1],...,_perm[_n]

  public:
    Perm_S( Range nObjects, Observer<C>* observer, FirstSolution firstSolution = Identity );

    // included for Negotiator::Mediator, which must handle Perm_s objects
    virtual void evaluate( )
      { _evaluation = 0; }  // dummy
 
    virtual void print( ostream& output ) const;

    virtual void doMove( Generic_S_N<C>& move );

    virtual bool computeEvaluation( const Generic_S_N<C>& move,
                                    T& evaluation, T& delta );

    Range getObjectAtPosition( Range i ) const
    {
      assert( (i>=0) && (i<=_n+1) );
      if (i==0) return 0;
      else if (i==_n+1) return _n+1;
      else return _perm[i];
    }

    Range n( ) const
      { return _n; }

    const vector<Range>& permutation( ) const
      { return _perm; }

    vector<Range>& permutation( )
      { return _perm; }

    // included for Negotiator::Mediator, which must handle Perm_s objects
    virtual Generic_S<C>* clone() const
      { return new Perm_S<C>( *this ); }

    virtual Generic_S<C>& operator=( const Generic_S<C>& rhs )
    {
      if ( this != &rhs )
      {
        assert( typeid(rhs) == typeid(*this) );
        const Perm_S<C>& drhs = static_cast<const Perm_S<C>&>( rhs );
        Generic_S<C>::operator=( rhs ); 
        _n = drhs._n;
        _perm = drhs._perm;
			}  
      return *this;
    }
    
    Perm_S<C>& operator=( const Perm_S<C>& rhs )
    {
      if ( this != &rhs )
        operator=( static_cast<const Generic_S<C>&>( rhs ) ); 
      return *this;
    }
};

//only used for ea:
//template <class C>
//float distance< >( const Perm_S<C>& s1, const Perm_S<C>& s2 )
//{
//  // similarity measure: percentage of common arcs (directed edges) (out of n+1)
//  
//  assert( s1.n() == s2.n() );
//
//  vector<typename C::Range> target( s1.n()+1, 0 );
//
//  for( typename C::Range i=1; i<s1.n(); ++i )
//    target[s1.getObjectAtPosition(i)] = s1.getObjectAtPosition(i+1);
//  target[0] = s1.getObjectAtPosition(1);
//  target[s1.getObjectAtPosition(s1.n())] = s1.n()+1;
//
//  float counter = 0;
//  for( typename C::Range i=1; i<s2.n(); ++i )
//    if ( target[s2.getObjectAtPosition(i)] == s2.getObjectAtPosition(i+1) )
//      counter += 1;
//  if ( target[0] == s2.getObjectAtPosition(1) )
//    counter += 1;
//  if ( target[s2.getObjectAtPosition(s2.n())] == s2.n()+1 )
//    counter += 1;
//
//  return counter/(s1.n()+1);
//}


// ----- class Perm_S_N -------------------------------------------------------
/*
  abstract base class for neighbors of Perm_S solutions
*/

template <class C>
class Perm_S_N
: public Generic_S_N<C>
{
  public:
    typedef typename C::T T;
    typedef typename C::Range Range;

  protected:
    Perm_S<C> *_s;

  public:
    // constructor
    Perm_S_N( Perm_S<C> *s, short depth = 1 )
      : Generic_S_N<C>( depth ), _s( s )
    { ; }

    Perm_S_N( const Perm_S_N& original )
      : Generic_S_N<C>( original ), _s( original._s )
      { ; }

    virtual void transform( vector<Range>& perm ) const = 0;

    Perm_S<C> * currentSolution( ) const
  		{ return _s; }

    virtual Generic_S_N<C>& operator=( const Generic_S_N<C>& rhs )
    {
      if ( this != &rhs )
      {
        assert( typeid(rhs) == typeid(*this) );
        const Perm_S_N<C>& drhs = static_cast<const Perm_S_N<C>&>( rhs );
        Generic_S_N<C>::operator=( rhs ); 
        _s = drhs._s;
      }  
      return *this;
    }

    Perm_S_N<C>& operator=( const Perm_S_N<C>& rhs )
    {
      if ( this != &rhs )
        operator=( static_cast<const Generic_S_N<C>&>( rhs ) ); 
      return *this;
    }
};


// ----- class Perm_S_N_Shift --------------------------------------------------
/*
  this class manages shift neighbors (stores, evaluates, provides, ...)
  - one such neighborhood class is needed for all neighborhood based heuristics
*/

template <class C>
class Perm_S_N_Shift
: public Perm_S_N<C>
{
  public:
    typedef typename C::T T;
    typedef typename C::Range Range;

  protected:
    vector<pair<Range, Range> > _shifts; // starting from _shifts.begin() (i.e. _shifts[0])
    // (x,y) means: shift object currently at position x behind object currently at position y
    // this requires: 1 <= x <= _n, x != y, y != x - 1, 0 <= y <= _n

    virtual void evaluate();

  public:
    static Range _s_maxDistance;

    // constructor
    Perm_S_N_Shift( Perm_S<C> *s,
			              NeighborhoodPosition position = FirstNeighbor,
                    short depth = 1 ); // so far: only implemented for depth = 1

    Perm_S_N_Shift( const Perm_S_N_Shift& original )
      : Perm_S_N<C>( original ), _shifts( original._shifts )
      { assert(_depth==1); }

    virtual Generic_S_N<C>* clone() const
      { return new Perm_S_N_Shift<C>( *this ); }

    virtual bool isValid( ) const
      { return ( _shifts[0].first != 0 ); }

    virtual void operator++( );

    virtual unsigned long size( ) const
      { return _s->n()*( _s->n()+1 ); }

    virtual void print( ostream& output ) const;

    virtual void transform( vector<Range>& perm ) const;

    Range x( Range i = 1 ) const
      { assert(i==1); return _shifts[i-1].first; }

    Range y( Range i = 1 ) const
      { assert(i==1); return _shifts[i-1].second; }

    Range actualDepth( ) const
      { return _shifts.size(); }

    Range maxDistance( ) const
      { return _s_maxDistance; }

    virtual Generic_S_N<C>& operator=( const Generic_S_N<C>& rhs )
    {
      if ( this != &rhs )
      {
        assert( typeid(rhs) == typeid(*this) );
        const Perm_S_N_Shift<C>& drhs = static_cast<const Perm_S_N_Shift<C>&>( rhs );
        Perm_S_N<C>::operator=( rhs ); 
        _shifts = drhs._shifts;
      }  
      return *this;
    }

    Perm_S_N_Shift<C>& operator=( const Perm_S_N_Shift<C>& rhs )
    {
      if ( this != &rhs )
        operator=( static_cast<const Generic_S_N<C>&>( rhs ) ); 
      return *this;
    }
};


template<class C>
typename Perm_S_N_Shift<C>::Range Perm_S_N_Shift<C>::_s_maxDistance = numeric_limits<Range>::max()/2;


// the following template function defines a transformation from Perm_S_N_Shift<C>
// moves to numbers, which is used by the REMTabuCriterion; so one must specialize
// this template function to use the REMTabuCriterion

template <class C>
unsigned long moveNumber( const Perm_S_N_Shift<C>& move )
{
  return ( move._shifts[0].first-1 ) * ( move.currentSolution()->n()+1 )
         + move._shifts[0].second; 
}


// ----- class Perm_S_N_Swap --------------------------------------------------
/*
  this class manages shift neighbors (stores, evaluates, provides, ...)
  - one such neighborhood class is needed for all neighborhood based heuristics
*/

template <class C>
class Perm_S_N_Swap
: public Perm_S_N<C>
{
  public:
    typedef typename C::T T;
    typedef typename C::Range Range;

  protected:
    vector<pair<Range, Range> > _swaps; // starting from _swaps.begin() (i.e. _swaps[0])
    // (x,y) means: swap objects currently at positions x and y
    // this requires: 1 <= x <= _n, 1 <= y <= _n, x < y

    virtual void evaluate();

  public:
    static Range _s_maxDistance;

    // constructor
    Perm_S_N_Swap( Perm_S<C> *s,
 		               NeighborhoodPosition position = FirstNeighbor,
                   short depth = 1 ); // so far: only implemented for depth = 1

    Perm_S_N_Swap( const Perm_S_N_Swap& original )
      : Perm_S_N<C>( original ), _swaps( original._swaps )
      { assert(_depth==1); }

    virtual Generic_S_N<C>* clone() const
      { return new Perm_S_N_Swap<C>( *this ); }

    virtual bool isValid( ) const
      { return ( _swaps[0].first != 0 ); }

    virtual void operator++( );

    virtual unsigned long size( ) const
      { return _s->n()*( _s->n()-1 )/2; }

    virtual void print( ostream& output ) const;

    virtual void transform( vector<Range>& perm ) const;

    Perm_S<C> * currentSolution( ) const
  		{ return _s; }

    Range x( Range i = 1 ) const
      { assert(i==1); return _swaps[i-1].first; }

    Range y( Range i = 1 ) const
      { assert(i==1); return _swaps[i-1].second; }

    Range actualDepth( ) const
      { return _swaps.size(); }

    Range maxDistance( ) const
      { return _s_maxDistance; }

    virtual Generic_S_N<C>& operator=( const Generic_S_N<C>& rhs )
    {
      if ( this != &rhs )
      {
        assert( typeid(rhs) == typeid(*this) );
        const Perm_S_N_Swap<C>& drhs = static_cast<const Perm_S_N_Swap<C>&>( rhs );
        Perm_S_N<C>::operator=( rhs ); 
        _swaps = drhs._swaps;
      }  
      return *this;
    }

    Perm_S_N_Swap<C>& operator=( const Perm_S_N_Swap<C>& rhs )
    {
      if ( this != &rhs )
        operator=( static_cast<const Generic_S_N<C>&>( rhs ) ); 
      return *this;
    }
};


template<class C>
typename Perm_S_N_Swap<C>::Range Perm_S_N_Swap<C>::_s_maxDistance = numeric_limits<Range>::max()/2;

// the following template function defines a transformation from Perm_S_N_Swap<C>
// moves to numbers, which is used by the REMTabuCriterion; so one must specialize
// this template function to use the REMTabuCriterion

template <class C>
unsigned long moveNumber( const Perm_S_N_Swap<C>& move )
{
  return ( move._swaps[0].first-1 ) * ( move.currentSolution()->n()+1 )
         + move._swaps[0].second; 
}


// ----- class Perm_S_A ---------------------------------------------------------------
/*
  this class manages attributes about solutions (e.g., extracts add/drop attributes from moves)
  - needed for all tabu search heuristics that use an attribute based tabu list
*/

template <class C>
class Perm_S_A
{
  public:
    typedef typename C::T  T;
    typedef typename C::Range Range;

  protected:
    Range _x, _y;  // arc (_x,_y)
    // change(tabu2): added the following element
    unsigned long _number; // unique numbering of all attributes

  public:
    Perm_S_A( );  // default constructor has to be provided (constructs dummy object)

    // attributes for a move
    Perm_S_A( const Perm_S_N<C>& move,
              unsigned int attributeIndex = 1 );
    static Range getNumberOfPositiveAttributes( const Perm_S_N<C>& move );
    static Range getNumberOfNegativeAttributes( const Perm_S_N<C>& move )
      { return getNumberOfPositiveAttributes( move ); }

    // attributes for a solution
    Perm_S_A( const Perm_S<C>& s, unsigned int attributeIndex );
    static Range getNumberOfAttributes( const Perm_S<C>& s )
      { return s.n()+1; }

    virtual bool operator==( const Perm_S_N<C>& move ) const;

    virtual void print( ostream& output ) const;

    // change(tabu2): added the following method
    virtual unsigned long number( ) const
      { return _number; }
};


// ----- class Perm_S_A_Pos -------------------------------------------------------------
/*
  this class manages attributes about solutions (e.g., extracts add/drop attributes from moves)
  - needed for all tabu search heuristics that use an attribute based tabu list
*/

template <class C>
class Perm_S_A_Pos
{
  public:
    typedef typename C::T  T;
    typedef typename C::Range Range;

  protected:
    Range _object, _position;

  public:
    Perm_S_A_Pos( const Perm_S_N<C>& move,
                  unsigned int attributeIndex = 1 );
    Perm_S_A_Pos( );  // default constructor has to be provided (constructs dummy object)
    
    static Range getNumberOfPositiveAttributes( const Perm_S_N<C>& move );
    static Range getNumberOfNegativeAttributes( const Perm_S_N<C>& move )
      { return getNumberOfPositiveAttributes( move ); }

    virtual bool operator==( const Perm_S_N<C>& move ) const;

    virtual void print( ostream& output ) const;
};


// ----- class Perm_S_I ---------------------------------------------------------------
/*
  this class manages information about solutions
  - possibly needed for all tabu search heuristics that use a trajectory based memory
  - here: done approximately using hash codes
*/

template <class C>
class Perm_S_I
{
  public:
    typedef typename C::T  T;
    typedef typename C::Range Range;

  protected:
    unsigned long _hashCode;
    static HashCoefficients _s_hashCoefficients;

  public:
    Perm_S_I( const Perm_S<C>& s );

    Perm_S_I( const Perm_S_N<C>& move );

    virtual bool operator==( const Perm_S_I<C>& rhs ) const
      { return ( rhs._hashCode == _hashCode ); }
    virtual bool operator<( const Perm_S_I<C>& rhs ) const
      { return ( rhs._hashCode < _hashCode ); }
    
    Range getHashCode() const
      { return _hashCode; }
    virtual void print( ostream& output ) const;
};


// definition of static class elements -----------------------------------------

template<class C>
HashCoefficients Perm_S_I<C>::_s_hashCoefficients;



// ----- implementation of class Perm_S ----------------------------------------

template<class C>
Perm_S<C>::Perm_S( Range nObjects, Observer<C>* observer, FirstSolution firstSolution ) // = Identity
: Generic_S<C>( observer ),             
  _n( nObjects ),
  _perm( _n+1, 0 )
{
  if ( firstSolution == Identity )
  {
    for( Range i=1; i<=_n; ++i )
      _perm[i] = i;
  } 
  else if ( firstSolution == Random )
  {
    for( Range i=1; i<=_n; ++i )
      _perm[i] = i;
    random_shuffle( _perm.begin()+1, _perm.end() );
  } 
  else
    assert( false );
}


template<class C>
void Perm_S<C>::print( ostream& output ) const
{
  // format: list the permutation (separated by spaces) 0
  //         evaluation

  for( Range i=1; i<=_n; ++i )
    output << getObjectAtPosition(i) << " ";

  output << "\n" << _evaluation << "\n";
}


template<class C>
bool Perm_S<C>::computeEvaluation( const Generic_S_N<C>& move,
                                   T& evaluation, T& delta )
{
  // this is an inefficient default evaluation, which only works for Perm_S_N moves;
  // this should be redefined in an derived class to efficiently evaluate a neighhbor adaptively

  // this must be redefined when there are special data fields in the derived
  // solution class, which are affected by evaluate()

  // dynamic downcast
  const Perm_S_N<C>& specificMove = dynamic_cast<const Perm_S_N<C>&>( move );

  T oldEvaluation = _evaluation;

  vector<Range> originalPermutation( _perm );
  specificMove.transform( _perm );
    
  evaluate( );

  delta = _evaluation - oldEvaluation;
  evaluation = -delta;

  // restore current solution (*this)
  _perm = originalPermutation;

  _evaluation = oldEvaluation;

  return true;
}

template<class C>
void Perm_S<C>::doMove( Generic_S_N<C>& move )
{
  // handling of Perm_S_N moves
  {
    const Perm_S_N<C> *specificMove = dynamic_cast<const Perm_S_N<C> *>(&move);
    if ( specificMove )
    {
      if ( _observer->active() )
      {
        _observer->move();
        _observer->getStream() << *specificMove; 
        _observer->flush();
      }
      _observer->moveEvaluation( **specificMove ); 
      if ( specificMove->deltaInformation() )
        _observer->deltaEvaluation( specificMove->delta() ); 

      specificMove->transform( _perm );

      if ( specificMove->deltaInformation() )
        _evaluation += specificMove->delta(); 
      else
        evaluate();

      if ( _evaluation < _s_upperBound )
      {
        _s_upperBound = _evaluation;
        _observer->newUpperBound( _evaluation );
        _observer->timeWhenNewUpperBoundWasFound( round( (double(clock()))/CLOCKS_PER_SEC, 3 )  );
      }
    
      return;
    }
  }

  assert( false );  // move not handled
}


// ----- implementation of class Perm_S_N_Shift ---------------------------------------

template <class C>
Perm_S_N_Shift<C>
  ::Perm_S_N_Shift( Perm_S<C> *s,
	                  NeighborhoodPosition position, short depth ) // =FirstNeighbor, =1
  : Perm_S_N<C>( s, depth ), _shifts( 1, pair<Range, Range>() )
{ 
  assert( _s_maxDistance > 0 );
  
  if ( ( position == InvalidNeighbor ) || ( _s->n() <= 1 ) )
  {
    _shifts[0].first = 0;
    return;
  }

  // else
  if ( position == SomeRandomNeighbor )
  {
    _shifts[0].first = rand( _s->n() ) + 1; // 1 <= first <= n

    if ( maxDistance() >= _s->n() )
    // no restriction due to maxDistance
    {
      do
        _shifts[0].second = rand( _s->n()+1 );
      while (    ( _shifts[0].second == _shifts[0].first )
              || ( _shifts[0].second == _shifts[0].first-1 ) );

      assert(    ( 1 <= _shifts[0].first ) && ( _shifts[0].first <= _s->n() )
              && ( 0 <= _shifts[0].second ) && ( _shifts[0].second <= _s->n() )
              && ( _shifts[0].first != _shifts[0].second ) && ( _shifts[0].first-1 != _shifts[0].second ) );
    }
    else
    // maxDistance restricts possible shifts
    {
      if (    ( _shifts[0].first > 1 )
           && ( ( rand01() < 0.5 ) || ( _shifts[0].first == _s->n() ) ) )
      // shift to the left
      {
        _shifts[0].second = _shifts[0].first - 2
                            - rand( min( maxDistance(), _shifts[0].first - 1 ) );
      }
      else
      // shift to the right
      {
        _shifts[0].second = _shifts[0].first + 1
                            + rand( min( maxDistance(), _s->n() - _shifts[0].first ) );
      }
    }

    evaluate();
    return;
  }

  _shifts[0].first = 1; 
  _shifts[0].second = 2; 
  evaluate();
}


template<class C>
void Perm_S_N_Shift<C>::evaluate()
{
  _deltaInformation =
    _s->computeEvaluation( *this, _evaluation, _delta );
}


template<class C>
void Perm_S_N_Shift<C>::operator++( )
{
  if ( _depth >= 2 )
  {
    assert( false );  // not yet done
  }
  else
  {
    if (    ( _shifts[0].second < _s->n() )
         && ( _shifts[0].second < _shifts[0].first + maxDistance() ) )
    {
      ++_shifts[0].second;
      if ( ( _shifts[0].second==_shifts[0].first-1 ) || ( _shifts[0].second==_shifts[0].first ) )
        _shifts[0].second = _shifts[0].first+1;
    }
    else
    {
      ++_shifts[0].first;
      _shifts[0].second = max( 0, _shifts[0].first - maxDistance() - 1 );
    }

    if ( _shifts[0].second == _s->n()+1 )
      _shifts[0].first = 0; // marker for 'no more neighbor'
    else
      evaluate();
  }
}


template<class C>
void Perm_S_N_Shift<C>::transform( vector<Range>& perm ) const
{
  Range objectToBeShifted = perm[_shifts[0].first];

  if ( _shifts[0].second <= _shifts[0].first-2 )
  // shift to left
  {
    for( Range i=_shifts[0].first; i>=_shifts[0].second+2; --i )
      perm[i] = perm[i-1];
    perm[_shifts[0].second+1] = objectToBeShifted;
  }
  else
  // shift to right
  {
    for( Range i=_shifts[0].first; i<=_shifts[0].second-1; ++i )
      perm[i] = perm[i+1];
    perm[_shifts[0].second] = objectToBeShifted;
  }                                             
}


template<class C>
void Perm_S_N_Shift<C>::print( ostream& output ) const
{
  for ( unsigned int i=0; i<_shifts.size(); ++i )
    output << " shift object at position " << _shifts[i].first
           << " behind object at position " << _shifts[i].second << "\n";
}


// ----- implementation of class Perm_S_N_Swap ---------------------------------------

template <class C>
Perm_S_N_Swap<C>
  ::Perm_S_N_Swap( Perm_S<C> *s,
	                 NeighborhoodPosition position, short depth ) // =FirstNeighbor, =1
  : Perm_S_N<C>( s, depth ), _swaps( 1, pair<Range, Range>() )
{ 
  assert( _s_maxDistance > 0 );
  
  if ( ( position == InvalidNeighbor ) || ( _s->n() <= 1 ) )
  {
    _swaps[0].first = 0;
    return;
  }

  // else
  if ( position == SomeRandomNeighbor )
  {
    _swaps[0].first = rand( _s->n() ) + 1; // 1 <= first <= n

    if ( maxDistance() >= _s->n() )
    // no restriction due to maxDistance
    {
      do
        _swaps[0].second = rand( _s->n() ) + 1;
      while ( _swaps[0].second == _swaps[0].first );

      if ( _swaps[0].first > _swaps[0].second )
        swap( _swaps[0].first, _swaps[0].second );

      assert(    ( 1 <= _swaps[0].first ) && ( _swaps[0].first <= _s->n() )
              && ( 1 <= _swaps[0].second ) && ( _swaps[0].second <= _s->n() )
              && ( _swaps[0].first < _swaps[0].second ) );
    }
    else
    // maxDistance restricts possible swaps
    {
      if (    ( _swaps[0].first > 1 )
           && ( ( rand01() < 0.5 ) || ( _swaps[0].first == _s->n() ) ) )
      // second < first
      {
        _swaps[0].second = _swaps[0].first - 1
                           - rand( min( maxDistance(), _swaps[0].first - 1 ) );
        swap( _swaps[0].first, _swaps[0].second );
      }
      else
      // second > first
      {
        _swaps[0].second = _swaps[0].first + 1
                           + rand( min( maxDistance(), _s->n() - _swaps[0].first ) );
      }
      assert(    ( 1 <= _swaps[0].first ) && ( _swaps[0].first <= _s->n() )
              && ( 1 <= _swaps[0].second ) && ( _swaps[0].second <= _s->n() )
              && ( _swaps[0].first < _swaps[0].second )
              && ( _swaps[0].second - _swaps[0].first <= maxDistance() ) );
    }

    evaluate();
    return;
  }

  _swaps[0].first = 1; 
  _swaps[0].second = 2; 
  evaluate();
}


template<class C>
void Perm_S_N_Swap<C>::evaluate()
{
  _deltaInformation =
    _s->computeEvaluation( *this, _evaluation, _delta );
}


template<class C>
void Perm_S_N_Swap<C>::operator++( )
{
  if ( _depth >= 2 )
  {
    assert( false );  // not yet done
  }
  else
  {
    if (    ( _swaps[0].second < _s->n() )
         && ( _swaps[0].second < _swaps[0].first + maxDistance() ) )
      ++_swaps[0].second;
    else
    {
      ++_swaps[0].first;
      if ( _swaps[0].first == _s->n() )
      {
        _swaps[0].first = 0; // marker for 'no more neighbor'
        return;
      }
      else
        _swaps[0].second = _swaps[0].first + 1;
    }
  }
  evaluate();
}


template<class C>
void Perm_S_N_Swap<C>::transform( vector<Range>& perm ) const
{
  swap( perm[_swaps[0].first], perm[_swaps[0].second] );
}


template<class C>
void Perm_S_N_Swap<C>::print( ostream& output ) const
{
  for ( unsigned int i=0; i<_swaps.size(); ++i )
    output << " swap object at position " << _swaps[i].first
           << " with object at position " << _swaps[i].second << "\n";
}


// ----- implementation of class Perm_S_A ---------------------------------------

template<class C>
Perm_S_A<C>::Perm_S_A( const Perm_S_N<C>& move,
                       unsigned int attributeIndex ) // =1 
{
  // handling of different kinds of moves

  {
    const Perm_S_N_Shift<C> *shift = dynamic_cast<const Perm_S_N_Shift<C> *>(&move);
    if ( shift )
    {
      assert( shift->actualDepth() == 1 );
  
      if ( attributeIndex == 1 )
      // new arc 1
      {
        _x = shift->currentSolution()->getObjectAtPosition( shift->x()-1 );
        _y = shift->currentSolution()->getObjectAtPosition( shift->x()+1 );
      }    
      else if ( attributeIndex == 2 )
      // new arc 2
      {
        _x = shift->currentSolution()->getObjectAtPosition( shift->y() );
        _y = shift->currentSolution()->getObjectAtPosition( shift->x() );
      }    
      else if ( attributeIndex == 3 )
      // new arc 3
      {
        _x = shift->currentSolution()->getObjectAtPosition( shift->x() );
        _y = shift->currentSolution()->getObjectAtPosition( shift->y()+1 );
      }    
      else if ( attributeIndex == -1 )
      // deleted arc 1
      {
        _x = shift->currentSolution()->getObjectAtPosition( shift->x()-1 );
        _y = shift->currentSolution()->getObjectAtPosition( shift->x() );
      }    
      else if ( attributeIndex == -2 )
      // deleted arc 2
      {
        _x = shift->currentSolution()->getObjectAtPosition( shift->x() );
        _y = shift->currentSolution()->getObjectAtPosition( shift->x()+1 );
      }    
      else if ( attributeIndex == -3 )
      // deleted arc 3
      {
        _x = shift->currentSolution( )->getObjectAtPosition( shift->y() );
        _y = shift->currentSolution( )->getObjectAtPosition( shift->y()+1 );
      }    
      else
        assert( false );  // attributeIndex not handled 

      // change(tabu2): added the following line
      _number = (shift->currentSolution()->n()+1) * _x + _y - 1;

      return;
    }
  }
  
  {
    const Perm_S_N_Swap<C> *swap = dynamic_cast<const Perm_S_N_Swap<C> *>(&move);
    if ( swap )
    {
      assert( swap->actualDepth() == 1 );

      if ( attributeIndex == 1 )
      // new arc 1
      {
        _x = swap->currentSolution()->getObjectAtPosition( swap->x()-1 );
        _y = swap->currentSolution()->getObjectAtPosition( swap->y() );
      }    
      else if ( attributeIndex == 2 )
      // new arc 2
      {
        _x = swap->currentSolution()->getObjectAtPosition( swap->y() );
        _y = swap->currentSolution()->getObjectAtPosition( swap->x()+1 );
      }    
      else if ( attributeIndex == 3 )
      // new arc 3
      {
        _x = swap->currentSolution()->getObjectAtPosition( swap->x() );
        _y = swap->currentSolution()->getObjectAtPosition( swap->y()+1 );
      }    
      else if ( ( swap->y() != swap->x() + 1 ) && ( attributeIndex == 4 ) )
      // new arc 4
      {
        _x = swap->currentSolution()->getObjectAtPosition( swap->y()-1 );
        _y = swap->currentSolution()->getObjectAtPosition( swap->x() );
      }
      else if ( attributeIndex == -1 )
      // deleted arc 1
      {
        _x = swap->currentSolution()->getObjectAtPosition( swap->x()-1 );
        _y = swap->currentSolution()->getObjectAtPosition( swap->x() );
      }    
      else if ( attributeIndex == -2 )
      // deleted arc 2
      {
        _x = swap->currentSolution()->getObjectAtPosition( swap->x() );
        _y = swap->currentSolution()->getObjectAtPosition( swap->x()+1 );
      }    
      else if ( attributeIndex == -3 )
      // deleted arc 3
      {
        _x = swap->currentSolution( )->getObjectAtPosition( swap->y() );
        _y = swap->currentSolution( )->getObjectAtPosition( swap->y()+1 );
      }    
      else if ( ( swap->y() != swap->x() + 1 ) && ( attributeIndex == 4 ) )
      // deleted arc 4
      {
        _x = swap->currentSolution( )->getObjectAtPosition( swap->y()-1 );
        _y = swap->currentSolution( )->getObjectAtPosition( swap->y() );
      }    
      else
        assert( false );  // attributeIndex not handled 

      // change(tabu2): added the following line (see above for shift, might be merged)
      _number = (swap->currentSolution()->n()+1) * _x + _y - 1;

      return;
    }
  }

  assert( false );  // move not handled
}


template<class C>
Perm_S_A<C>::Perm_S_A( const Perm_S<C>& s,
                       unsigned int attributeIndex )
{
  assert( attributeIndex >= 1 );
  assert( attributeIndex <= s.n()+1 );

  _x = s.getObjectAtPosition( attributeIndex - 1 );
  _y = s.getObjectAtPosition( attributeIndex );

  _number = (s.n()+1) * _x + _y - 1;
}


template<class C>
Perm_S_A<C>::Perm_S_A( )
// default constructor has to be provided (constructs dummy object/attribute)
: _x( 0 ), _y( 0 ), _number( numeric_limits<T>::max() )
{
  ;  
}


template<class C>
bool Perm_S_A<C>::operator==( const Perm_S_N<C>& move ) const
{
  // checks iff an attribute of the move equals *this
  // i.e. the move would invert the attribute that is represented by *this

  // we assume a move of depth 1
  
  // handling of different kinds of moves

  {
    const Perm_S_N_Shift<C> *shift = dynamic_cast<const Perm_S_N_Shift<C> *>(&move);
    if ( shift )
    {
      return ( 
               (    ( _x == shift->currentSolution()->getObjectAtPosition( shift->x()-1 ) )
                 && ( _y == shift->currentSolution()->getObjectAtPosition( shift->x() ) )
               ) 
               ||
               (    ( _x == shift->currentSolution()->getObjectAtPosition( shift->x() ) )
                 && ( _y == shift->currentSolution()->getObjectAtPosition( shift->x()+1 ) )
               ) 
               ||
               (    ( _x == shift->currentSolution()->getObjectAtPosition( shift->y() ) )
                 && ( _y == shift->currentSolution()->getObjectAtPosition( shift->y()+1 ) )
               ) 
             );
    }
  }
  
  {
    const Perm_S_N_Swap<C> *swap = dynamic_cast<const Perm_S_N_Swap<C> *>(&move);
    if ( swap )
    {
      return (
              (     ( _x == swap->currentSolution()->getObjectAtPosition( swap->x()-1 ) )
                 && ( _y == swap->currentSolution()->getObjectAtPosition( swap->x() ) )
               ) 
               ||
               (    ( _x == swap->currentSolution()->getObjectAtPosition( swap->x() ) )
                 && ( _y == swap->currentSolution()->getObjectAtPosition( swap->x()+1 ) )
               ) 
               ||
               (    ( _x == swap->currentSolution()->getObjectAtPosition( swap->y()-1 ) )
                 && ( _y == swap->currentSolution()->getObjectAtPosition( swap->y() ) )
               ) 
               ||
               (    ( _x == swap->currentSolution()->getObjectAtPosition( swap->y() ) )
                 && ( _y == swap->currentSolution()->getObjectAtPosition( swap->y()+1 ) )
               ) 
             );
    }
  }  

  assert( false );  // move not handled
  return false;
}


template<class C>
typename Perm_S_A<C>::Range Perm_S_A<C>::getNumberOfPositiveAttributes( const Perm_S_N<C>& move )
{
  // we assume a move of depth 1
  
  // handling of different kinds of moves

  {
    const Perm_S_N_Shift<C> *shift = dynamic_cast<const Perm_S_N_Shift<C> *>(&move);
    if ( shift )
      return 3;
  }

  {
    const Perm_S_N_Swap<C> *swap = dynamic_cast<const Perm_S_N_Swap<C> *>(&move);
    if ( swap )
    {
      if ( swap->y() == swap->x() + 1 )
        return 3;
      else
        return 4;
    }
  }
  
  assert( false );  // move not handled
  return 0;
}


template<class C>
void Perm_S_A<C>::print( ostream& output ) const
{
  output << "arc (" << _x << "," << _y << ")" << "\n";
}


// ----- implementation of class Perm_S_A_Pos -----------------------------------

template<class C>
Perm_S_A_Pos<C>::Perm_S_A_Pos( const Perm_S_N<C>& move,
                               unsigned int attributeIndex ) // =1 
{
  // handling of swap moves

  // shift moves are note handled since this combination of move and attributes
  // would not make sense (?)

  {
    const Perm_S_N_Swap<C> *swap = dynamic_cast<const Perm_S_N_Swap<C> *>(&move);
    if ( swap )
    {
      assert( swap->actualDepth() == 1 );

      if ( attributeIndex == 1 )
      // new assignment 1
      {
        _object = swap->currentSolution()->getObjectAtPosition( swap->x() );
        _position = swap->y();
      }    
      else if ( attributeIndex == 2 )
      // new assignment 2
      {
        _object = swap->currentSolution()->getObjectAtPosition( swap->y() );
        _position = swap->x();
      }    
      else if ( attributeIndex == -1 )
      // deleted assignment 1
      {
        _object = swap->currentSolution()->getObjectAtPosition( swap->x() );
        _position = swap->x();
      }    
      else if ( attributeIndex == -2 )
      // deleted assignment 2
      {
        _object = swap->currentSolution()->getObjectAtPosition( swap->y() );
        _position = swap->y();
      }    
      else
        assert( false );  // attributeIndex not handled 

      return;
    }
  }

  assert( false );  // move not handled
}


template<class C>
Perm_S_A_Pos<C>::Perm_S_A_Pos( )
// default constructor has to be provided (constructs dummy object/attribute)
: _object( 0 ), _position( 0 )
{
  ;  
}


template<class C>
bool Perm_S_A_Pos<C>::operator==( const Perm_S_N<C>& move ) const
{
  // checks iff an attribute of the move equals *this
  // i.e. the move would invert the attribute that is represented by *this

  // we assume a move of depth 1
  
  // handling of swap moves

  // shift moves are note handled since this combination of move and attributes
  // would not make sense (?)

  {
    const Perm_S_N_Swap<C> *swap = dynamic_cast<const Perm_S_N_Swap<C> *>(&move);
    if ( swap )
    {
      return (
               (    ( _object == swap->currentSolution()->getObjectAtPosition( swap->x() ) )
                 && ( _position == swap->x() )
               ) 
               ||
               (    ( _object == swap->currentSolution()->getObjectAtPosition( swap->y() ) )
                 && ( _position == swap->y() )
               ) 
             );
    }
  }  

  assert( false );  // move not handled
  return false;
}


template<class C>
typename Perm_S_A_Pos<C>::Range Perm_S_A_Pos<C>::getNumberOfPositiveAttributes( const Perm_S_N<C>& move )
{
  // we assume a move of depth 1
  
  // shift moves are note handled since this combination of move and attributes
  // would not make sense (?)

  {
    const Perm_S_N_Swap<C> *swap = dynamic_cast<const Perm_S_N_Swap<C> *>(&move);
    if ( swap )
      return 2;
  }
  
  assert( false );  // move not handled
  return 0;
}


template<class C>
void Perm_S_A_Pos<C>::print( ostream& output ) const
{
  output << "object " << _object << " at position " << _position << "\n";
}


// ----- implementation of class Perm_S_I ---------------------------------------

template <class C>
Perm_S_I<C>::Perm_S_I( const Perm_S<C>& s )
{
  if ( _s_hashCoefficients.size() == 0 )
    _s_hashCoefficients.initialize( s.n(), 7*s.n() );

  _hashCode = 0;
  for( Range i=1; i<=s.n(); ++i )
    _hashCode += s.getObjectAtPosition(i) * _s_hashCoefficients.get_coefficient(i);
}


template <class C>
Perm_S_I<C>::Perm_S_I( const Perm_S_N<C>& move )
{
  // handling of different kinds of moves

  {
    const Perm_S_N_Shift<C> *shift = dynamic_cast<const Perm_S_N_Shift<C> *>(&move);
    if ( shift )
    {
      assert( shift->actualDepth() == 1 );

      if ( _s_hashCoefficients.size() == 0 )
        _s_hashCoefficients.initialize( shift->currentSolution()->n(),
                                        7*shift->currentSolution()->n() );

      // TODO: do the following more efficiently

      vector<Range> perm = shift->currentSolution()->permutation();
      shift->transform( perm );
      _hashCode = 0;

      for( Range i=1; i<=shift->currentSolution()->n(); ++i )
        _hashCode += perm[i] * _s_hashCoefficients.get_coefficient(i);
    }
  }

  {
    const Perm_S_N_Swap<C> *swap = dynamic_cast<const Perm_S_N_Swap<C> *>(&move);
    if ( swap )
    {
      assert( swap->actualDepth() == 1 );

      if ( _s_hashCoefficients.size() == 0 )
        _s_hashCoefficients.initialize( swap->currentSolution()->n(),
                                        7*swap->currentSolution()->n() );

      // TODO: do the following more efficiently

      vector<Range> perm = swap->currentSolution()->permutation();
      swap->transform( perm );
      _hashCode = 0;

      for( Range i=1; i<=swap->currentSolution()->n(); ++i )
      {
        if ( ( i != swap->x() ) && ( i != swap->y() ) )
          _hashCode += perm[i] * _s_hashCoefficients.get_coefficient(i);
        else
          if ( i == swap->x() )
            _hashCode += perm[swap->y()] * _s_hashCoefficients.get_coefficient(i);
          else
            _hashCode += perm[swap->x()] * _s_hashCoefficients.get_coefficient(i);
      }
    }
  }
}


template<class C>
void Perm_S_I<C>::print( ostream& output ) const
{
  output << _hashCode << "\n";
}


// stream output functions -----------------------------------------------------
// (all of these just call the respective (virtual) member function)

template <class C>
ostream& operator<<( ostream& output, const Perm_S<C>& solution )
{
  solution.print( output );
  return output;
}

template <class C>
ostream& operator<<( ostream& output, const Perm_S_N_Shift<C>& neighbor )
{
  neighbor.print( output );
  return output;
}

template <class C>
ostream& operator<<( ostream& output, const Perm_S_A<C>& attribute )
{
  attribute.print( output );  
  return output;
}

template <class C>
ostream& operator<<( ostream& output, const Perm_S_A_Pos<C>& attribute )
{
  attribute.print( output );  
  return output;
}

template <class C>
ostream& operator<<( ostream& output, const Perm_S_I<C>& solutionInfo )
{
  solutionInfo.print( output );  
  return output;
}


}
//--- end of namespace HotFrame -----------------------------------------------
//------------------------------------------------------------------------------


#endif
