// SA: generic simulated annealing and variations

#if !defined(__HotFrame_SA__)
#define __HotFrame_SA__

#include "basic.h"
#include "search.h"
#include "interface.h"


//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
namespace HotFrame
{


// ----------------------------------------------------------------------------
// -- AcceptanceCriterion -----------------------------------------------------
// ----------------------------------------------------------------------------

// each acceptance criterion must define a method
//   static bool check( double t, double moveEval, double newSolutionEval )
// the meaning of the parameters is as follows:
// - t: control parameter
// - moveEval: move evaluation, positive means improvement
// - newSolutionEval: evaluation (objective function value) of candidate solution


template <class Configuration>
class ClassicExponentialAcceptanceCriterion
{
  public:
    static bool check( double t, double moveEval, double newSolutionEval = 0 )
      { return ( ( moveEval >= 0 ) || ( exp( moveEval/t ) > rand01() ) ); }
    
    static string getDescription()
      { return "ClassicExponentialAcceptanceCriterion"; }
};


template <class C>
class ClassicThresholdAcceptanceCriterion
{
  public:
    static bool check( double t, double moveEval, double newSolutionEval = 0 )
      { return ( ( moveEval >= 0 ) || ( -moveEval <= t ) ); }
    
    static string getDescription()
      { return "ClassicThresholdAcceptanceCriterion"; }
};


template <class C>
class AbsoluteThresholdAcceptanceCriterion
{
  public:
    static bool check( double t, double moveEval, double newSolutionEval )
      { return ( newSolutionEval <= t ); }
    
    static string getDescription()
      { return "AbsoluteThresholdAcceptanceCriterion"; }
};


// ----------------------------------------------------------------------------
// -- CoolingSchedule ---------------------------------------------------------
// ----------------------------------------------------------------------------

// each cooling schedule must define four methods:
//   static void repetitionIntervalDone( double& t, double tinitial, 
//                                       unsigned long iteration )
//   static void moveAccepted( double& t, double moveEval, double newSolutionEval )
//   static void moveRejected( double& t, double moveEval, double newSolutionEval )
//   static void newBestObjective( double& t, double newSolutionEval )
// the meaning of the parameters is as follows:
// - t: control parameter, to be changed (by reference)
// - iteration: current iteration
// - tinitial: initial value of control parameter
// - moveEval: move evaluation, positive means improvement
// - newSolutionEval: evaluation (objective function value) of candidate solution


template <class C>
class GeometricCooling
{
  public:
    static double _s_alpha;

  public:
    static void repetitionIntervalDone( double& t, double tinitial = 0, 
                                        unsigned long iteration = 0 )
      { t *= _s_alpha; }
    
    static void moveAccepted( double& t, double moveEval, double newSolutionEval )
      { ; }
    static void moveRejected( double& t, double moveEval, double newSolutionEval )
      { ; }
    static void newBestObjective( double& t, double newSolutionEval )
      { ; }

    static string getDescription()
      { return "GeometricCooling"; }
};

template <class C>
double GeometricCooling<C>::_s_alpha 
  = C::alpha;


template <class C>
class HajekCooling
{
  public:
    static void repetitionIntervalDone( double& t, double tinitial, 
                                        unsigned long iteration )
      { t = tinitial / log(1+iteration); }
    
    static void moveAccepted( double& t, double moveEval, double newSolutionEval )
      { ; }
    static void moveRejected( double& t, double moveEval, double newSolutionEval )
      { ; }
    static void newBestObjective( double& t, double newSolutionEval )
      { ; }

    static string getDescription()
      { return "HajekCooling"; }
};


template <class C>
class LundiMeesCooling
{
  public:
    static double _s_alpha;

  public:
    static void repetitionIntervalDone( double& t, double tinitial = 0, 
                                        unsigned long iteration = 0 )
      { t /= ( 1 + _s_alpha*t); }
    
    static void moveAccepted( double& t, double moveEval, double newSolutionEval )
      { ; }
    static void moveRejected( double& t, double moveEval, double newSolutionEval )
      { ; }
    static void newBestObjective( double& t, double newSolutionEval )
      { ; }

    static string getDescription()
      { return "LundiMeesCooling"; }
};

template <class C>
double LundiMeesCooling<C>::_s_alpha 
  = C::alpha;


template <class C>
class DowslandCooling
{
  public:
    static double _s_alpha1;
    static double _s_alpha2;

  public:
    static void repetitionIntervalDone( double& t, double tinitial, 
                                        unsigned long iteration  )
      { ; }
    
    static void moveAccepted( double& t, double moveEval = 0, double newSolutionEval = 0 )
      { t /= ( 1 + _s_alpha1*t); }
    static void moveRejected( double& t, double moveEval = 0, double newSolutionEval = 0 )
      { t /= ( 1 - _s_alpha2*t); }
    
    static void newBestObjective( double& t, double newSolutionEval )
      { ; }

    static string getDescription()
      { return "LundiMeesCooling"; }
};

template <class C>
double DowslandCooling<C>::_s_alpha1 
  = C::alpha1;

template <class C>
double DowslandCooling<C>::_s_alpha2
  = C::alpha2;


template <class C>
class GreatDelugeCooling
{
  public:
    static double _s_alpha1;
    static double _s_alpha2;

  public:
    static void repetitionIntervalDone( double& t, double tinitial, 
                                        unsigned long iteration  )
      { ; }
    
    static void moveAccepted( double& t, double moveEval, double newSolutionEval )
      { t = min( t - _s_alpha1 * t, t - ( t - newSolutionEval ) / _s_alpha2 ); }
    static void moveRejected( double& t, double moveEval, double newSolutionEval )
      { ; }
    
    static void newBestObjective( double& t, double newSolutionEval )
      { ; }

    static string getDescription()
      { return "GreatDelugeCooling"; }
};

template <class C>
double GreatDelugeCooling<C>::_s_alpha1 
  = C::alpha1;

template <class C>
double GreatDelugeCooling<C>::_s_alpha2
  = C::alpha2;


template <class C>
class RecordToRecordTravelCooling
{
  public:
    static double _s_alpha;

  public:
    static void repetitionIntervalDone( double& t, double tinitial, 
                                        unsigned long iteration )
      { ; }
    static void moveAccepted( double& t, double moveEval, double newSolutionEval )
      { ; }
    static void moveRejected( double& t, double moveEval, double newSolutionEval )
      { ; }
    
    static void newBestObjective( double& t, double newSolutionEval )
      { t = newSolutionEval + _s_alpha; }

    static string getDescription()
      { return "RecordToRecordTravelCooling"; }
};

template <class C>
double RecordToRecordTravelCooling<C>::_s_alpha 
  = C::alpha;


// ----------------------------------------------------------------------------
// -- Reheating ---------------------------------------------------------------
// ----------------------------------------------------------------------------

// each reheating schema must define a method
//   static void reheat( double& t, double tinitial, double tbest )
// the meaning of the parameters is as follows:
// - tinitial: initial value of control parameter, to be changed (by reference)
// - tbest: value of control parameter when best solution was found
// - t: control parameter


  template <class C>
  class NoReheating
  {
    public:
      static void reheat( double& tinitial, double tbest = 0, double t = 0 )
        { ; }
    
      static string getDescription()
        { return "NoReheating"; }
  };


template <class C>
class ReheatingToHalfOfInitial
{
  public:
    static void reheat( double& tinitial, double tbest = 0, double t = 0 )
      { tinitial = tinitial / 2; }
    
    static string getDescription()
      { return "ReheatingToHalfOfInitial"; }
};


template <class C>
class ReheatingToBest
{
  public:
    static void reheat( double& tinitial, double tbest, double t = 0 )
      { tinitial = tbest; }
    
    static string getDescription()
      { return "ReheatingToHalfOfInitial"; }
};

template <class C>
class ReheatingToAverageOfBestAndInitial
{
  public:
    static void reheat( double& tinitial, double tbest, double t = 0 )
      { assert( tinitial >= tbest ); 
        tinitial = ( tinitial + tbest ) / 2; }
    
    static string getDescription()
      { return "ReheatingToMiddleBetweenBestAndInitial"; }
};


// ----------------------------------------------------------------------------
// -- GeneralSimulatedAnnealing -----------------------------------------------
// ----------------------------------------------------------------------------

template <class C>
class GeneralSimulatedAnnealing
: public Heuristic<typename C::S>
{
  public:
    typedef typename C::T T;
    typedef typename C::Range Range;
    typedef typename C::CNumeric CNumeric;
    typedef typename C::S S;
    typedef typename C::N N;

    typedef typename C::Cp Cp;

    typedef typename C::AcceptanceCriterion AcceptanceCriterion;
    typedef typename C::CoolingSchedule CoolingSchedule;
    typedef typename C::Reheating Reheating;

  protected:
    Observer<CNumeric> *_observer;
    float _maxTimeInSeconds;
    unsigned int _maxMoves;
    OmegaInterface<S> *_omegaInterface;
    short _depth;
    bool _returnBest;
    double _tinitial;
    unsigned long _maxRepetitions;
    double _deltaRepetitions;
    unsigned int _numberOfReheatings;

  public:
    GeneralSimulatedAnnealing(
        Observer<CNumeric> *observer = 0,
        float maxTimeInSeconds = 0,
        unsigned long maxMoves = 0,
        OmegaInterface<S> *omegaInterface = 0,
        short depth=1,
        bool returnBest=true,
        double tinitial = 100,
        unsigned long maxRepetitions = 1,
        double deltaRepetitions = 1,
        unsigned int numberOfReheatings = 0 );

    virtual void search( S& s, unsigned long maxMoves = 0 );
};

template <class C>
GeneralSimulatedAnnealing<C>
::GeneralSimulatedAnnealing( 
        Observer<CNumeric> *observer, // = 0
        float maxTimeInSeconds, // = 0
        unsigned long maxMoves, // = 0
        OmegaInterface<S> *omegaInterface, // = 0
        short depth, // = 1
        bool returnBest, // = true,
        double tinitial, // = 100
        unsigned long maxRepetitions, // = 1
        double deltaRepetitions, // = 1
        unsigned int numberOfReheatings ) // = 1
: _observer( observer ), _maxTimeInSeconds( maxTimeInSeconds ), _maxMoves( maxMoves ), 
  _omegaInterface( omegaInterface ), _depth( depth ), _returnBest( returnBest ),
  _tinitial( tinitial ), _maxRepetitions( maxRepetitions ), _deltaRepetitions( deltaRepetitions ),
  _numberOfReheatings( numberOfReheatings )
{
  // this may lead to a small memory leak ...
  if ( !observer )
    _observer = new Observer<CNumeric>( );
  if ( !_omegaInterface )
    _omegaInterface = new OmegaInterface<S>( );
}


template <class C>
void GeneralSimulatedAnnealing<C>
::search( S& s, unsigned long maxMoves ) // = 0
{
  // ASSUMPTION:
  //   some cooling schedules and acceptance criteria use delta information,
  //   so we assume that delta information is indeed provided by the neighborhood
  //   when needed (any alternative ...?)
  
  if ( maxMoves != 0 )
    _maxMoves = maxMoves;

  clock_t startTime = clock( );

  _observer->start( "GeneralSimulatedAnnealing with:\n" 
                    + AcceptanceCriterion::getDescription() + "\n"
                    + CoolingSchedule::getDescription() + "\n"
                    + Reheating::getDescription() );

  S bestSolution( s );

  _omegaInterface->newSolution( s );

  double tinitial = _tinitial;  // we may change it
  double t, tbest;

  unsigned long numberOfIterations = 0;

  unsigned long maxRepetitions = _maxRepetitions; // we may change it

  for( unsigned int h=1; h<=_numberOfReheatings+1; ++h )
  {
    if ( h > 1 )
      Reheating::reheat( tinitial, tbest, t );
    else
      tbest = tinitial;
    
    t = tinitial;
        
    unsigned long i=0;  // iteration

    do
    {
      for( unsigned long r=1; r<=maxRepetitions; ++r )
      {
        ++i;
        _observer->iteration( i );
        ++numberOfIterations;

        N nextCandidateNeighbor =
          N( &s, SomeRandomNeighbor, _depth );

        if ( AcceptanceCriterion::check( t, *nextCandidateNeighbor, 
                                         s.f() + nextCandidateNeighbor.delta() ) )
        {
          CoolingSchedule::moveAccepted( t, *nextCandidateNeighbor, 
                                         s.f() + nextCandidateNeighbor.delta() );
          s.doMove( nextCandidateNeighbor );
          _omegaInterface->newSolution( s );
         
          if ( s.f() < bestSolution.f() )
          {  
            bestSolution = s;
            tbest = t;
            CoolingSchedule::newBestObjective( t, s.f() );
          }
        }
        else
          CoolingSchedule::moveRejected( t, *nextCandidateNeighbor, 
                                         s.f() + nextCandidateNeighbor.delta() );
      
        if (    ( _maxTimeInSeconds > 0 )
             && ( ( ( float ( clock() - startTime ) ) / CLOCKS_PER_SEC ) >= _maxTimeInSeconds ) )
          // time limit reached, so stop simulated annealing 
          break;
      }

      CoolingSchedule::repetitionIntervalDone( t, tinitial, i );
      _observer->newTemperature( t );

      if (    ( _maxTimeInSeconds > 0 )
           && ( ( ( float ( clock() - startTime ) ) / CLOCKS_PER_SEC ) >= _maxTimeInSeconds ) )
        // time limit reached, so stop simulated annealing 
        break;

      if ( _omegaInterface->omega( ) )
        break;  // external termination condition fulfilled, so stop the search

      maxRepetitions = static_cast<unsigned long>( maxRepetitions * _deltaRepetitions );
      _observer->newMaxRepetitions( maxRepetitions );

    }
    while( ( maxMoves==0 ) || ( i<=_maxMoves ) );

    if (    ( _maxTimeInSeconds > 0 )
         && ( ( ( float ( clock() - startTime ) ) / CLOCKS_PER_SEC ) >= _maxTimeInSeconds ) )
      // time limit reached, so stop simulated annealing 
      break;

    if ( _omegaInterface->omega( ) )
      break;  // external termination condition fulfilled, so stop the search
  }

  if ( _returnBest )
    s = bestSolution;

  _observer->numberOfIterations( numberOfIterations );
}



// ----------------------------------------------------------------------------
// -- Johnson et al. simulated annealing algorithm ----------------------------
// ----------------------------------------------------------------------------

template <class C>
class SimulatedAnnealingJetal
: public Heuristic<typename C::S>
{
  public:
    typedef typename C::T T;
    typedef typename C::Range Range;
    typedef typename C::CNumeric CNumeric;
    typedef typename C::S S;
    typedef typename C::N N;

    typedef typename C::Cp Cp;

    typedef typename C::AcceptanceCriterion AcceptanceCriterion;
    typedef typename C::CoolingSchedule CoolingSchedule;
    typedef typename C::Reheating Reheating;

  protected:
    Observer<CNumeric> *_observer;
    float _maxTimeInSeconds;
    unsigned long _maxMoves;
    OmegaInterface<S> *_omegaInterface;
    short _depth;
    bool _returnBest;
    float _initialAcceptanceFraction;
    float _frozenAcceptanceFraction;
    float _sizeFactor;
    unsigned int _frozenParameter;
    unsigned int _numberOfReheatings;

  public:
    SimulatedAnnealingJetal(
                   Observer<CNumeric> *observer = 0,
                   float maxTimeInSeconds = 0,
                   unsigned long maxMoves = 0,
                   OmegaInterface<S> *omegaInterface = 0,
                   short depth = 1,
                   bool returnBest = true,
                   float initialAcceptanceFraction = 0.4,
                   float frozenAcceptanceFraction = 0.02, 
                   float sizeFactor = 16,
                   unsigned int frozenParameter = 5,
                   unsigned int numberOfReheatings = 0 );

    virtual void search( S& s, unsigned long maxMoves=0 ); // maxMoves is neglected
};

template <class C>
SimulatedAnnealingJetal<C>
::SimulatedAnnealingJetal(
                 Observer<CNumeric> *observer, // = 0
                 float maxTimeInSeconds, // = 0
                 unsigned long maxMoves, // = 0
                 OmegaInterface<S> *omegaInterface, // = 0
                 short depth, // = 1,
                 bool returnBest, // = true,
                 float initialAcceptanceFraction, // = 0.4,
                 float frozenAcceptanceFraction, // = 0.02, 
                 float sizeFactor, // = 16,
                 unsigned int frozenParameter, // = 5,
                 unsigned int numberOfReheatings ) // = 0,
: _observer( observer ), _maxTimeInSeconds( maxTimeInSeconds ), _maxMoves( maxMoves ),
  _omegaInterface( omegaInterface ), _depth( depth ), _returnBest( returnBest ),
  _initialAcceptanceFraction( initialAcceptanceFraction ),
  _frozenAcceptanceFraction( frozenAcceptanceFraction ), _sizeFactor( sizeFactor ),
  _frozenParameter( frozenParameter ), _numberOfReheatings( numberOfReheatings )
{
  // this may lead to a small memory leak ...
  if ( !observer )
    _observer = new Observer<CNumeric>( );
  if ( !_omegaInterface )
    _omegaInterface = new OmegaInterface<S>( );
}

template <class C>
void SimulatedAnnealingJetal<C>
::search( S& s, unsigned long maxMoves ) // =0, neglected
{
  clock_t startTime = clock( );

  _observer->start( "Simulated Annealing Johnson et al." );

  double tinitial;

  S bestSolution( s );

  _omegaInterface->newSolution( s );

  // determine temperature so that the fraction of accepted moves is about initialAcceptanceFraction
  // (the following procedure does not really provide such a temperature value (which could be
  //  computed) but some kind of approximation (for reasons of run-time efficiency))
  {
    _observer->comment( "start of trial annealing run to determine temperature" );

    S trialSolution( s );

    unsigned long numberOfMovesAccepted = 0;
    vector<T> deterioratingMoveEvaluations;

    N trialNeighbor = N( &trialSolution, SomeRandomNeighbor, _depth );
    unsigned long trialSize = static_cast<unsigned long>( _sizeFactor * trialNeighbor.size() );

    for( unsigned long i=1; i<=trialSize; ++i )
    {
      N nextCandidateNeighbor =
        N( &trialSolution, SomeRandomNeighbor, _depth );

      if ( *nextCandidateNeighbor > 0 )
      {
        trialSolution.doMove( nextCandidateNeighbor );
        ++numberOfMovesAccepted;
      }
      else
        deterioratingMoveEvaluations.push_back( *nextCandidateNeighbor );
    }
    sort( deterioratingMoveEvaluations.begin(), deterioratingMoveEvaluations.end(), greater<T>() );

    long numberOfMovesMissing
      = static_cast<unsigned long>(trialSize*_initialAcceptanceFraction) - numberOfMovesAccepted;
    if ( numberOfMovesMissing < 0 )
      numberOfMovesMissing = 0;

    if ( ( numberOfMovesMissing == 0 ) && ( deterioratingMoveEvaluations.size() == 0 )  )
      deterioratingMoveEvaluations.push_back( 1 );  // should not happen

    tinitial = -deterioratingMoveEvaluations[numberOfMovesMissing] / 0.693;
    _observer->comment( "end of trial annealing run to determine temperature" );
    _observer->newTemperature( tinitial );

    if ( trialSolution.f() < bestSolution.f() )
      bestSolution = trialSolution;
  }

  double t, tbest;
  
  unsigned long numberOfIterations = 0;

  for( unsigned int h=1; h<=_numberOfReheatings+1; ++h )
  {
    if ( h > 1 )
      Reheating::reheat( tinitial, tbest, t );
    else
      tbest = tinitial;

    t = tinitial;

    unsigned int frozenCounter = 0;
    unsigned long i=0;  // iteration
    do
    {
      // set size of repetition interval as sizeFactor times neighbourhood size;
      // note that the neighbourhood size may vary considerably during the search process 
      N trialNeighbor = N( &s, SomeRandomNeighbor, _depth );
      unsigned long maxRepetitions = static_cast<unsigned long>( _sizeFactor * trialNeighbor.size() );

      unsigned long numberOfImprovingMovesDoneInRepetitionInterval = 0;
      unsigned long numberOfNonImprovingMovesDoneInRepetitionInterval = 0;

      for( unsigned long r=1; r<=maxRepetitions; ++r )
      {
        ++i;
        _observer->iteration( i );
        ++numberOfIterations;

        N nextCandidateNeighbor = N( &s, SomeRandomNeighbor, _depth );

        if ( *nextCandidateNeighbor > 0 )
        {
          _observer->improvingSAMove( *nextCandidateNeighbor );

          s.doMove( nextCandidateNeighbor );
          ++numberOfImprovingMovesDoneInRepetitionInterval;
          _omegaInterface->newSolution( s );
          CoolingSchedule::moveAccepted( t, *nextCandidateNeighbor, 
                                         s.f() + nextCandidateNeighbor.delta() );
        }
        else if ( AcceptanceCriterion::check( t, *nextCandidateNeighbor, 
                                              s.f() + nextCandidateNeighbor.delta() ) )
        {
          _observer->deterioratingSAMove( *nextCandidateNeighbor );

          s.doMove( nextCandidateNeighbor );
          ++numberOfNonImprovingMovesDoneInRepetitionInterval;
          _omegaInterface->newSolution( s );
          CoolingSchedule::moveAccepted( t, *nextCandidateNeighbor, 
                                         s.f() + nextCandidateNeighbor.delta() );
        }
        else
          CoolingSchedule::moveRejected( t, *nextCandidateNeighbor, 
                                         s.f() + nextCandidateNeighbor.delta() );

        if ( s.f() < bestSolution.f() )
        {
          bestSolution = s;
          tbest = t;
          frozenCounter = 0;
          CoolingSchedule::newBestObjective( t, s.f() );
        } 
      }

      if ( ( _maxMoves > 0 ) && ( i >= _maxMoves ) )
        // time limit reached, so stop simulated annealing 
        break;

      if (    ( _maxTimeInSeconds > 0 )
           && ( ( ( float ( clock() - startTime ) ) / CLOCKS_PER_SEC ) >= _maxTimeInSeconds ) )
        // time limit reached, so stop simulated annealing 
        break;
 
      if ( _omegaInterface->omega( ) )
        break;  // external termination condition fulfilled, so stop simulated annealing

      CoolingSchedule::repetitionIntervalDone( t, tinitial, i );
      _observer->newTemperature( t );

      if (    ( float(numberOfImprovingMovesDoneInRepetitionInterval
                      +numberOfNonImprovingMovesDoneInRepetitionInterval)
                / float(maxRepetitions) <= _frozenAcceptanceFraction ) 
           || ( numberOfImprovingMovesDoneInRepetitionInterval == 0 ) )
        ++frozenCounter;
        // the second criterion takes into account that all performed moves may be
        // moves on a big plane with evaluation 0
    }
    while( frozenCounter < 5 );

    if (    ( _maxTimeInSeconds > 0 )
         && ( ( ( float ( clock() - startTime ) ) / CLOCKS_PER_SEC ) >= _maxTimeInSeconds ) )
      // time limit reached, so stop simulated annealing 
      break;

    if ( _omegaInterface->omega( ) )
      break;  // external termination condition fulfilled, so stop the search
  }

  if ( _returnBest )
    s = bestSolution;

  _observer->numberOfIterations( numberOfIterations );
}



}
//--- end of namespace HotFrame -----------------------------------------------
//------------------------------------------------------------------------------


#endif
