// SEARCH: basic generic local search algorithms


#if !defined(__HotFrame_SEARCH__)
#define __HotFrame_SEARCH__


#include "basic.h"
#include "interface.h"


//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
namespace HotFrame
{


// -----------------------------------------------------------------------------
// -- (Do-Nothing-) Heuristic --------------------------------------------------
// -----------------------------------------------------------------------------

template <class S>
class Heuristic
{
  public:
    virtual ~Heuristic( ) 
      { ; }
    virtual void search( S& s, unsigned long maxMoves=0 )
      { ; }
};


// -----------------------------------------------------------------------------
// -- NeighborSelection --------------------------------------------------------
// -----------------------------------------------------------------------------

// each neighbor selection scheme must define a method
//   static N select( S& solution, short depth = 1  )


template <class C>
class BestPositiveNeighbor
{
  public:
    typedef typename C::S S;
    typedef typename C::N N;

  public:
    static string getDescription()
      { return "select best positive neighbor"; }

    static N select( S& solution, short depth = 1 )
    {
      N neighbor = N( &solution, FirstNeighbor, depth );

      if ( ! ( neighbor.isValid() ) )
        return neighbor;

      //uncomment the following lines for documenting neighborhood traversal
      //if ( solution.observer()->active() )
      //{
      //  solution.observer()->currentNeighbor();
      //  solution.observer()->getStream() << neighbor;
      //  solution.observer()->getStream() << *neighbor << "\n";
      //  solution.observer()->flush();
      //}

      N best = neighbor;

      ++neighbor;

      while ( neighbor.isValid() )
      { 
        //uncomment the following lines for documenting neighborhood traversal
        //if ( solution.observer()->active() )
        //{
        //  solution.observer()->currentNeighbor();
        //  solution.observer()->getStream() << neighbor;
        //  solution.observer()->getStream() << *neighbor << "\n";
        //  solution.observer()->flush();
        //}

        if ( *best < *neighbor )
            best = neighbor;
          ++neighbor;
      }

      // if we have not found any neighbor with positive evaluation return invalid neighbor
      if ( *best <= 0 )
        best = neighbor;

      return best;
  }
};


template <class C>
class BestNeighbor
{
  public:
    typedef typename C::S S;
    typedef typename C::N N;

  public:
    static string getDescription()
      { return "select best neighbor"; }

    static N select( S& solution, short depth = 1  )
    {
      N neighbor = N( &solution, FirstNeighbor, depth );

      if ( ! ( neighbor.isValid() ) )
        return neighbor;
      
      if ( solution.observer()->active() )
      {
        solution.observer()->currentNeighbor();
        solution.observer()->getStream() << neighbor;
        solution.observer()->getStream() << *neighbor << "\n";
        solution.observer()->flush();
      }

      N best = neighbor;

      ++neighbor;

      while ( neighbor.isValid() )
      { 
        if ( solution.observer()->active() )
        {
          solution.observer()->currentNeighbor();
          solution.observer()->getStream() << neighbor;
          solution.observer()->getStream() << *neighbor << "\n";
          solution.observer()->flush();
        }

        if ( *best < *neighbor )
          best = neighbor;
        ++neighbor;
      }

      return best;
  }
};


template <class C>
class FirstPositiveNeighbor
{
  public:
    typedef typename C::S S;
    typedef typename C::N N;

  public:
    static string getDescription()
      { return "select first positive neighbor"; }

    static N select( S& solution, short depth = 1  )
    {
      N neighbor = N( &solution, FirstNeighbor, depth );

      if ( ! ( neighbor.isValid() ) )
        return neighbor;
      
      if ( solution.observer()->active() )
      {
        solution.observer()->currentNeighbor();
        solution.observer()->getStream() << neighbor;
        solution.observer()->getStream() << *neighbor << "\n";
        solution.observer()->flush();
      }

      while ( ( neighbor.isValid() ) && ( *neighbor<=0 ) )
      {
        ++neighbor;

        if ( ( solution.observer()->active() ) && ( neighbor.isValid() ) )
        {
          solution.observer()->currentNeighbor();
          solution.observer()->getStream() << neighbor;
          solution.observer()->getStream() << *neighbor << "\n";
          solution.observer()->flush();
        }
      }
    
      return neighbor;
    }
};


template <class C>
class RandomNeighbor
{
  public:
    typedef typename C::S S;
    typedef typename C::N N;

  public:
    static string getDescription()
      { return "select random neighbor"; }

    static N select( S& solution, short depth = 1  )
    {
      return N( &solution, SomeRandomNeighbor, depth );
    }
};


// -----------------------------------------------------------------------------
// -- IteratedLocalSearch ------------------------------------------------------
// -----------------------------------------------------------------------------

template <class C>
class IteratedLocalSearch
: public Heuristic<typename C::S>
{
  public:
    typedef typename C::T T;
    typedef typename C::Range Range;
    typedef typename C::CNumeric CNumeric;
    typedef typename C::S S;
    typedef typename C::N N;
    typedef typename C::NeighborSelection NeighborSelection;

  protected:
    Observer<CNumeric> *_observer;
    float _maxTimeInSeconds;
    unsigned long _maxMoves;
    unsigned long _repetitions;
    OmegaInterface<S> *_omegaInterface;
    Heuristic<S> *_diversification;
    short _depth;
    bool _returnBest;

  public:

    IteratedLocalSearch( Observer<CNumeric> *observer = 0,
                         float maxTimeInSeconds=0,
                         unsigned long maxMoves=0,
                         unsigned long repetitions=1,
                         OmegaInterface<S> *omegaInterface=0,
                         Heuristic<S> *diversification=0,
                         short depth=1,
                         bool returnBest=true );
    
    virtual void search( S& s, unsigned long maxMoves=0  );
};



template <class C>
IteratedLocalSearch<C>
::IteratedLocalSearch( Observer<CNumeric> *observer,
                       float maxTimeInSeconds, unsigned long maxMoves, unsigned long repetitions,
                       OmegaInterface<S> *omegaInterface,
                       Heuristic<S> *diversification, short depth, bool returnBest )
: _observer( observer ), _maxTimeInSeconds( maxTimeInSeconds ), _maxMoves( maxMoves ),
  _repetitions( repetitions ), _omegaInterface( omegaInterface ),
  _diversification( diversification ), _depth( depth ), _returnBest( returnBest )
{
  // this may lead to a small memory leak ...

  if ( !observer )
    _observer = new Observer<CNumeric>( );
  if ( !diversification )
    _diversification = new Heuristic<S>( );
  if ( !_omegaInterface )
    _omegaInterface = new OmegaInterface<S>( );
}

template <class C>
void IteratedLocalSearch<C>
::search( S& s, unsigned long maxMoves ) // =0
{
  if ( maxMoves != 0 )
    _maxMoves = maxMoves;
  
  clock_t startTime = clock();

  unsigned long numberOfIterations = 0;
 
  _observer->start( "iterated local search, " + NeighborSelection().getDescription() );

  S bestSolution( s );

  _omegaInterface->newSolution( s );

  for( unsigned long r=1; ( ( r<=_repetitions ) || ( _repetitions==0 ) ); ++r )
  { 
    bool omegaTermination = false;
   
    for( unsigned long i=1; ( ( i<=_maxMoves ) || ( _maxMoves==0 ) ); ++i )
    {
      if ( s.f() == s.lowerBound() )
      {
        _observer->searchSuspended( "optimum reached (lower bound)" );
        break;
      }

      _observer->iteration( i );
      ++numberOfIterations;

      N neighbor = NeighborSelection::select( s, _depth );

      if ( ! ( neighbor.isValid() ) )
      {
        _observer->searchSuspended( "current solution has no neighbor that is valid in this context" );
        break;
      }

      s.doMove( neighbor );

      if ( _observer->active() )
      {
        _observer->currentSolution();
        _observer->getStream() << s;
        _observer->flush();
      }

      if ( ( _returnBest ) && ( s.f() < bestSolution.f() ) )
        bestSolution = s;

      _observer->newSolutionWithEvaluation( s.f() );
  
      _omegaInterface->newSolution( s );

      if (    ( _maxTimeInSeconds > 0 )
           && ( ( ( float ( clock() - startTime ) ) / CLOCKS_PER_SEC ) >= _maxTimeInSeconds ) )
        // time limit reached, so stop searching
        break;

      if ( _omegaInterface->omega( ) )
      {    
        omegaTermination = true;  // external termination condition fulfilled, so stop searching
        break;
      }
    }

    if ( s.f() == s.lowerBound() )
    {
      _observer->searchSuspended( "optimum reached (lower bound)" );
      break;
    }

    if (    ( _maxTimeInSeconds > 0 )
         && ( ( ( float ( clock() - startTime ) ) / CLOCKS_PER_SEC ) >= _maxTimeInSeconds ) )
      // time limit reached, so stop searching
      break;
 
    if ( omegaTermination )
      break;  // external termination condition fulfilled, so stop searching

    if ( r != _repetitions )  // i.e., there is a next run to diversify for
    {
      _diversification->search( s );

      if ( s.f() < bestSolution.f() )
        bestSolution = s;
    }
  }

  if ( _returnBest )
    s = bestSolution;

  _observer->numberOfIterations( numberOfIterations );
}


// -----------------------------------------------------------------------------
// -- IteratedLocalSearchID (IncreasingDepth) ------------------------------------
// -----------------------------------------------------------------------------

// IteratedLocalSearchID might have been integrated with IteratedLocalSearch;
// however, maybe we want to change the procedures later in unforeseen ways;
// so we keep it separated so far


template <class C>
class IteratedLocalSearchID
: public Heuristic<typename C::S>
{
  public:
    typedef typename C::T T;
    typedef typename C::Range Range;
    typedef typename C::CNumeric CNumeric;
    typedef typename C::S S;
    typedef typename C::N N;
    typedef typename C::NeighborSelection NeighborSelection;

  protected:
    Observer<CNumeric> *_observer;
    short _maxDepth;  // maximal neighborhood depth
    bool _reset;       // reset neighborhood depth to 0 after every move ?
    float _maxTimeInSeconds;
    unsigned long _maxMoves;
    unsigned long _repetitions;
    OmegaInterface<S> *_omegaInterface;
    Heuristic<S> *_diversification;
    bool _returnBest;

  public:

    IteratedLocalSearchID( Observer<CNumeric> *observer = 0,
                           short maxDepth=2,
                           bool reset=false,
                           float maxTimeInSeconds=0,
                           unsigned long maxMoves=0,
                           unsigned long repetitions=1,
                           OmegaInterface<S> *omegaInterface=0,
                           Heuristic<S> *diversification=0,
                           bool returnBest=true );

    void search( S& s, unsigned long maxMoves=0  );
};


template <class C>
IteratedLocalSearchID<C>
::IteratedLocalSearchID( Observer<CNumeric> *observer, short maxDepth, bool reset,
                       float maxTimeInSeconds, unsigned long maxMoves,
                       unsigned long repetitions, OmegaInterface<S> *omegaInterface,
                       Heuristic<S> *diversification, bool returnBest )
: _observer( observer ), _maxDepth( maxDepth ), _reset( reset ), 
  _maxTimeInSeconds( maxTimeInSeconds ), _maxMoves( maxMoves ),
  _repetitions( repetitions ), _omegaInterface( omegaInterface ),
  _diversification( diversification ), _returnBest( _returnBest )
{
  // this may lead to a small memory leak ...

  if ( !observer )
    _observer = new Observer<CNumeric>( );
  if ( !diversification )
    _diversification = new Heuristic<S>( );
  if ( !_omegaInterface )
    _omegaInterface = new OmegaInterface<S>( );
}


template <class C>
void IteratedLocalSearchID<C>
::search( S& s, unsigned long maxMoves ) //=1
{
  if ( maxMoves != 0 )
    _maxMoves = maxMoves;

  clock_t startTime = clock( );
 
  unsigned long numberOfIterations = 0;

  _observer->start( "iterated local search, increasing depth, " + NeighborSelection().getDescription() );

  S bestSolution( s );

  _omegaInterface->newSolution( s );

  for( unsigned long r=1; r<=_repetitions; ++r )
  { 
    short depth = 1;

    for( unsigned long i=1; ( ( i<=_maxMoves ) || ( _maxMoves==0 ) ); ++i )
    {
      _observer->iteration( i );
      ++numberOfIterations;

      N neighbor = NeighborSelection::select( s, depth );

      if ( ! ( neighbor.isValid() ) )
      {
        if ( depth >= _maxDepth )
        {
          _observer->searchSuspended( "current solution has no neighbor that is valid in this context" );
          break;
        }
        else  
        // switch search to next neighborhood depth
        {
          ++depth;
          _observer->newNeighborhoodDepth( depth );
          continue;
        }
      }

      s.doMove( neighbor );
      
      if ( _observer->active() )
      {
        _observer->currentSolution();
        _observer->getStream() << s;
        _observer->flush();
      }

      if ( _reset )
        depth = 1;

      if ( ( _returnBest ) && ( s.f() < bestSolution.f() ) )
        bestSolution = s;

      _observer->newSolutionWithEvaluation( s.f() );
  
      _omegaInterface->newSolution( s );
 
      if (    ( _maxTimeInSeconds > 0 )
           && ( ( ( float ( clock() - startTime ) ) / CLOCKS_PER_SEC ) >= _maxTimeInSeconds ) )
        // time limit reached, so stop search
        break;

      if ( _omegaInterface->omega( ) )
        break;  // external termination condition fulfilled, so stop tabu search
    }

    if (    ( _maxTimeInSeconds > 0 )
         && ( ( ( float ( clock() - startTime ) ) / CLOCKS_PER_SEC ) >= _maxTimeInSeconds ) )
      // time limit reached, so stop search
      break;
 
    if ( _omegaInterface->omega( ) )
      break;  // external termination condition fulfilled, so stop tabu search

    if ( r != _repetitions )  // i.e., there is a next run to diversify for
    {
      _diversification->search( s );

      if ( s.f() < bestSolution.f() )
        bestSolution = s;
    }

  }

  if ( _returnBest )
    s = bestSolution;

  _observer->numberOfIterations( numberOfIterations );
}



}
//--- end of namespace HotFrame ------------------------------------------------
//------------------------------------------------------------------------------


#endif
