// tabu: generic tabu search algorithms

#if !defined(__HotFrame_TABU__)
#define __HotFrame_TABU__

#include "basic.h"
#include "search.h"
#include "interface.h"


//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
namespace HotFrame
{

template <class C> class TabuCriterion;


// ----------------------------------------------------------------------------
// -- AspirationCriterion ---------------------------------------------------
// ----------------------------------------------------------------------------

// each tabu neighbor selection scheme must define a method
//     static bool check( N& neighbor, TabuCriterion<C>& tabu )


template <class C>
class NoAspirationCriterion
{
  public:
    typedef typename C::N N;
    typedef typename C::S S;

  public:
    static string description()
      { return "NoAspirationCriterion"; }

    static bool check( N& neighbor, TabuCriterion<C>& tabu )
      { return false; }
};


template <class C>
class NewBestSolutionAspirationCriterion
{
  public:
    typedef typename C::T T;
    typedef typename C::N N;
    typedef typename C::S S;

  public:
    static string description()
      { return "NewBestSolutionAspirationCriterion"; }

    static bool check( N& neighbor, TabuCriterion<C>& tabu )
    {
      // of course, we can only apply this criterion if the neighbor provides delta information
      if ( !( neighbor.deltaInformation() ) )
        return false;

      // due to strange rounding errors, sometimes, the aspiration criterion
      // may be fulfilled though the values compared should be equal;
      // the following hack might, hopefully, remedy such a situation
      if ( ( neighbor.currentSolution()->f() + neighbor.delta()
               + 500*numeric_limits<T>::epsilon()
             < neighbor.currentSolution()->upperBound() )
           && ( *neighbor >= 0 ) )
      {
        neighbor.currentSolution()->observer()->aspirationCriterionSatisfied();
        return true;
      }
      else
        return false;
    }
};



// ----------------------------------------------------------------------------
// -- TabuNeighborSelection ---------------------------------------------------
// ----------------------------------------------------------------------------

// each tabu neighbor selection scheme must define a method
//     static N select( S& solution, TabuCriterion<Cp>& tabuCriterion, short depth = 1  )


template <class C>
class BestAdmissibleNeighbor
{
  public:
    typedef typename C::S S;
    typedef typename C::N N;
    typedef typename C::AspirationCriterion AspirationCriterion;
    typedef typename C::Cp Cp;

  public:
    static string description()
      { return "BestAdmissibleNeighbor"; }

    static N select( S& solution, TabuCriterion<Cp>& tabuCriterion, short depth = 1 )
    {
      N move = N( &solution, FirstNeighbor, depth );

      if ( ! ( move.isValid() ) )
        return move;  // i.e., neighborhood is empty

      if ( solution.observer()->active() )
      {
        solution.observer()->currentNeighbor();
        solution.observer()->getStream() << move;
        solution.observer()->getStream() << *move << "\n";
        solution.observer()->flush();
      }

      N bestAdmissible = move;

      bool admissibleMoveFound
           = (    ( ! ( tabuCriterion.tabu( bestAdmissible ) ) )
               || ( AspirationCriterion::check( bestAdmissible, tabuCriterion ) ) );

      ++move;
      
      while( move.isValid() )
      {
        if ( solution.observer()->active() )
        {
          solution.observer()->currentNeighbor();
          solution.observer()->getStream() << move;
          solution.observer()->getStream() << *move << "\n";
          solution.observer()->flush();
        }

        if ( (!admissibleMoveFound) || ( *bestAdmissible < *move ) )
        // this would be a new best move, so check whether this move is admissible
        {
          if (    ( ! ( tabuCriterion.tabu( move ) ) )
               || ( AspirationCriterion::check( move, tabuCriterion ) ) )
          {
            bestAdmissible = move;
            admissibleMoveFound = true;
          }

        }

        ++move;
      }

      if ( ! admissibleMoveFound ) 
        // i.e., we have not found any admissible neighbor, so return the invalid neighbor
        bestAdmissible = move;

      return bestAdmissible;
    }
};


template <class C>
class BestNeighborConsideringPenalties
{
  public:
    typedef typename C::T T;
    typedef typename C::S S;
    typedef typename C::N N;
    typedef typename C::AspirationCriterion AspirationCriterion;
    typedef typename C::Cp Cp;

  public:
    static double _s_penaltyFactor;

  public:
    static string description()
      { return "BestNeighborConsideringPenalties"; }

    static N select( S& solution, TabuCriterion<Cp>& tabuCriterion, short depth = 1 )
    {
      N move = N( &solution, FirstNeighbor, depth );
      if ( ! ( move.isValid() ) )
        return move;

      if ( solution.observer()->active() )
      {
        solution.observer()->currentNeighbor();
        solution.observer()->getStream() << move;
        solution.observer()->getStream() << *move << "\n";
        solution.observer()->flush();
      }

      N best = move;
      
      T bestEvaluationConsideringPenalties
        = *best - ( AspirationCriterion::check( move, tabuCriterion ) ?
                    0 : _s_penaltyFactor*tabuCriterion.tabuDegree( best ) );

      ++move;
      
      while ( move.isValid() )
      { 
        if ( solution.observer()->active() )
        {
          solution.observer()->currentNeighbor();
          solution.observer()->getStream() << move;
          solution.observer()->getStream() << *move << "\n";
          solution.observer()->flush();
        }

        if ( bestEvaluationConsideringPenalties
          < *move - ( AspirationCriterion::check( move, tabuCriterion ) ?
                      0 : _s_penaltyFactor*tabuCriterion.tabuDegree( move ) ) )
        {
          best = move;
          bestEvaluationConsideringPenalties
            = *move - ( AspirationCriterion::check( move, tabuCriterion ) ?
                        0 : _s_penaltyFactor*tabuCriterion.tabuDegree( move ) );
        }
        ++move;
      }

      return best;
    }
};

template <class C>
double BestNeighborConsideringPenalties<C>::_s_penaltyFactor 
  = C::penaltyFactor;



// ----------------------------------------------------------------------------
// -- TabuSearch --------------------------------------------------------------
// ----------------------------------------------------------------------------

template <class C>
class TabuSearch
: public Heuristic<typename C::S>
{
  public:
    typedef typename C::T T;
    typedef typename C::Range Range;
    typedef typename C::CNumeric CNumeric;
    typedef typename C::S S;
    typedef typename C::N N;
    typedef typename C::S_A S_A;
    typedef typename C::S_I S_I;

    typedef typename C::Cp Cp;

    typedef typename C::TabuNeighborSelection TabuNeighborSelection;
    typedef typename C::AspirationCriterion AspirationCriterion;

  protected:
    Observer<CNumeric> *_observer;
    float _maxTimeInSeconds;
    unsigned long _maxMoves;
    OmegaInterface<S> *_omegaInterface;
    TabuCriterion<Cp> *_tabuCriterion;
    Heuristic<typename C::S> *_diversification;
    short _depth;

  public:
    TabuSearch( Observer<CNumeric> *observer = 0,
                float maxTimeInSeconds = 0,
                unsigned long maxMoves = 0,
                OmegaInterface<S> *_omegaInterface = 0,
                TabuCriterion<Cp> *tabuCriterion = 0,
                Heuristic<S> *diversification = 0,
                short depth = 1 );

    virtual void search( S& s, unsigned long maxMoves=0  );
};


template <class C>
TabuSearch<C>
::TabuSearch( Observer<CNumeric> *observer, // =0
              float maxTimeInSeconds, // =0
              unsigned long maxMoves, // =0
              OmegaInterface<S> *omegaInterface, // =0
              TabuCriterion<Cp> *tabuCriterion,  // =0
              Heuristic<S> *diversification,  // =0
              short depth ) // =1
: _observer( observer ), _maxTimeInSeconds( maxTimeInSeconds ), _maxMoves( maxMoves ), 
  _omegaInterface( omegaInterface ), _tabuCriterion( tabuCriterion ),
  _diversification( diversification ), _depth( depth )
{
  // this may lead to a small memory leak ...

  if ( !observer )
    _observer = new Observer<CNumeric>( );
  if ( !_omegaInterface )
    _omegaInterface = new OmegaInterface<S>( );
  if ( !tabuCriterion )
    _tabuCriterion = new TabuCriterion<Cp>( );
  if ( !diversification )
    _diversification = new Heuristic<S>( );
}


template <class C>
void TabuSearch<C>
::search( S& s, unsigned long maxMoves )
{
  if ( maxMoves != 0 )
    _maxMoves = maxMoves;
  
  clock_t startTime = clock( );

  S startingSolution( s );

  _observer->start( "tabu search" );
  _observer->comment( "neighbor selection rule: " + TabuNeighborSelection::description() );
  _observer->comment( "tabu criterion: " + _tabuCriterion->description() );
  _observer->comment( "aspiration criterion: " + AspirationCriterion::description() );

  _tabuCriterion->addToHistory( s, 0 );

  S bestSolution( s );

  _omegaInterface->newSolution( s );

  unsigned long i;
  for( i=1; ( ( i<=_maxMoves ) || ( _maxMoves==0 ) ); ++i )
  {
    _observer->iteration( i );

    if ( s.f() == s.lowerBound() )
    {
      _observer->searchSuspended( "optimum reached (lower bound)" );
      break;
    }

    N bestAdmissibleNeighbor (
      TabuNeighborSelection::select( s, *_tabuCriterion, _depth ) );

    if ( ! ( bestAdmissibleNeighbor.isValid() ) )
    // no admissible neighbor, so get a random neighbor
    {
      bestAdmissibleNeighbor =
        N( &s, SomeRandomNeighbor, _depth );

      if ( ! ( bestAdmissibleNeighbor.isValid() ) )
      // no neighbor at all
      {
        _observer->searchSuspended( "current solution has no neighbor" );
        break;
      }

      _observer->noAdmissibleNeighbor( "random neighbor selected" );
    }

    unsigned long escapeTrigger;

    escapeTrigger = _tabuCriterion->addToHistory( bestAdmissibleNeighbor, i );

    s.doMove( bestAdmissibleNeighbor );

    if ( escapeTrigger == 0 )
      escapeTrigger = _tabuCriterion->addToHistory( s, i );

    if ( s.f() < bestSolution.f() )
      bestSolution = s;

    _observer->newSolutionWithEvaluation( s.f() );

    if ( _observer->active() )
    {
      _observer->currentSolution( );
      _observer->getStream() << s;
      _observer->flush();
    }

    _omegaInterface->newSolution( s );

    if ( escapeTrigger > 0 )
    {
      _observer->escape( "tabu memory triggered escape" ) ;
      _tabuCriterion->escape( );
      if ( escapeTrigger == 1 )
        _diversification->search( s );
      else
        _diversification->search( s, escapeTrigger > 50 ? 50 : escapeTrigger );
    }

    _tabuCriterion->print( );

    if (    ( _maxTimeInSeconds > 0 )
         && ( ( ( float ( clock() - startTime ) ) / CLOCKS_PER_SEC ) >= _maxTimeInSeconds ) )
         break;  // time limit reached, so stop tabu search

    if ( _omegaInterface->omega( ) )
      break;  // external termination condition fulfilled, so stop tabu search
  }

  s = bestSolution;

  _observer->numberOfIterations( i );
}


// ----------------------------------------------------------------------------
// -- Tabu Criteria -----------------------------------------------------------
// ----------------------------------------------------------------------------

// each tabu criterion must define methods as shown for class TabuCriterion,
// which represents the tabu criterion base class (nothing is tabu)


template <class C>
class TabuCriterion
{
  public:
    typedef typename C::T T;
    typedef typename C::Range Range;
    typedef typename C::CNumeric CNumeric;
    typedef typename C::S S;
    typedef typename C::N N;
    typedef typename C::S_A S_A;
    typedef typename C::S_I S_I;

  public:
    virtual ~TabuCriterion( )
      { ; }

    virtual unsigned long addToHistory( const N& move, unsigned long iteration=0 )
      { return 0; }  // no escape

    virtual unsigned long addToHistory( const S& solution, unsigned long iteration=0 )
      { return 0; }  // no escape

    virtual bool tabu( const N& move ) const  
      { return false; }  // i.e. nothing is tabu

    virtual double tabuDegree( const N& move ) const  
      { return tabu( move ); }  // 0 <= tabuDegree <= 1 (=strictly tabu=default)

    virtual void escape( )
      { ; }

    virtual void print( ) const
      { ; }

    virtual string description()
      { return "NoTabuCriterion"; }
};

// ----------------------------------------------------------------------------
// -- Static Tabu Criterion ---------------------------------------------------
// ----------------------------------------------------------------------------

template <class C>
class StaticTabuCriterion
: public TabuCriterion<C>
{
  public:
    typedef typename C::T T;
    typedef typename C::Range Range;
    typedef typename C::CNumeric CNumeric;
    typedef typename C::S S;
    typedef typename C::N N;
    typedef typename C::S_A S_A;
    typedef typename C::S_I S_I;

  private:
    vector<S_A> _tabuList;
    Range _currentListPosition;  // stores position of last attribute included
    Range _tabuThreshold; // how many attributes of a move have to be in _tabuList
                          // to classify the move as tabu

    Observer<CNumeric> *_observer;

  public:
    StaticTabuCriterion( Observer<CNumeric> *observer = 0,
                         Range tabuListLength=7, Range tabuThreshold=1);

    unsigned long addToHistory( const N& move, unsigned long iteration=0 );

    unsigned long addToHistory( const S& solution, unsigned long iteration=0 );

    bool tabu( const N& move ) const;

    double tabuDegree( const N& move ) const;

    void escape( );

    void print( ) const;

    Range tabuListLength( ) const
      { return static_cast<Range>( _tabuList.size() ); }

    string description()
      { return "StaticTabuCriterion"; }
};


template <class C>
StaticTabuCriterion<C>::StaticTabuCriterion
( Observer<CNumeric> *observer,
  Range tabuListLength, Range tabuThreshold ) // =7,=1
: _tabuList( tabuListLength ), _currentListPosition( 0 ), _tabuThreshold( tabuThreshold ),
  _observer( observer )
{
  if ( !observer )
    _observer = new Observer<CNumeric>( );

  _observer->tabuCriterionInitialized( "static" );
}


template <class C>
unsigned long StaticTabuCriterion<C>::addToHistory( const N& move, unsigned long iteration )
{
  for( Range i=1; i<=S_A::getNumberOfPositiveAttributes( move ); ++i )
  {
    ++_currentListPosition;
    if ( _currentListPosition == tabuListLength( ) )
      _currentListPosition = 0;

    _tabuList[_currentListPosition] = S_A( move, i );

    if ( _observer->active() )
    {
      _observer->attributeAddedToHistory( );
      _observer->getStream() << _tabuList[_currentListPosition];
      _observer->flush();
    }
  }

  return 0;
}


template <class C>
unsigned long  StaticTabuCriterion<C>::addToHistory( const S& solution,
                                                     unsigned long iteration )
{
  // nothing to do
  return 0;
}


template <class C>
bool StaticTabuCriterion<C>::tabu( const N& move ) const
{
  Range numberOfAttributesFound = 0;

  Range tempListPosition = _currentListPosition;
  do
  {
    if ( _tabuList[tempListPosition] == move )
      ++numberOfAttributesFound;

    // TODO: think about attributes that are more than one time in the tabu list
    //       should this case be treated in a special way?

    ++tempListPosition;
    if ( tempListPosition == tabuListLength( ) )
      tempListPosition = 0;
  }
  while( tempListPosition != _currentListPosition );

  // in special cases move.getNumberOfNegativeAttributes() might be smaller than _tabuThreshold;
  //   so it should be reasonable to define the following condition
  if ( numberOfAttributesFound >= min( int(_tabuThreshold),
                                       int(S_A::getNumberOfNegativeAttributes( move )) ) )
    return true;
  else
    return false;
}


template <class C>
double StaticTabuCriterion<C>::tabuDegree( const N& move ) const
{
  Range numberOfAttributesFound = 0;

  Range tempListPosition = _currentListPosition;
  do
  {
    if ( _tabuList[tempListPosition] == move )
      ++numberOfAttributesFound;

    // TODO: think about attributes that are more than one time in the tabu list
    //       should this case be treated in a special way?

    ++tempListPosition;
    if ( tempListPosition == tabuListLength( ) )
      tempListPosition = 0;
  }
  while( tempListPosition != _currentListPosition );

  // in special cases move.getNumberOfNegativeAttributes() might be smaller than _tabuThreshold;
  //   so it should be reasonable to define the following condition
  if ( numberOfAttributesFound >= min( int(_tabuThreshold),
                                       int (S_A::getNumberOfNegativeAttributes( move ) ) ) )
    return 1;
  else
    return numberOfAttributesFound / _tabuThreshold;
}


template <class C>
void StaticTabuCriterion<C>::escape( )
{
  _currentListPosition = 0;

  for( unsigned int i=0; i<_tabuList.size(); ++i )
    _tabuList[i] = S_A( );  // dummy move 
}


template <class C>
void StaticTabuCriterion<C>::print( ) const
{
  if ( _observer->active() )
  {
    _observer->currentTabuList( );

    Range tempListPosition = _currentListPosition;
    do
    {
      _observer->getStream() << _tabuList[tempListPosition];
      if ( tempListPosition == 0 )
        tempListPosition = tabuListLength( )-1;
      else
        --tempListPosition;
    }
    while ( tempListPosition != _currentListPosition );

    _observer->flush();
  }
}


// ----------------------------------------------------------------------------
// -- REM Tabu Criterion ------------------------------------------------------
// ----------------------------------------------------------------------------

// works only for single attribute moves with a moveNumber() function available

template <class C>
class REMTabuCriterion
: public TabuCriterion<C>
{
  public:
    typedef typename C::T T;
    typedef typename C::Range Range;
    typedef typename C::CNumeric CNumeric;
    typedef typename C::S S;
    typedef typename C::N N;
    typedef typename C::S_A S_A;
    typedef typename C::S_I S_I;

  private:
    vector<unsigned long> _runningList;
    Range _nextRunningListPosition;  // stores position of next move to be included

    unsigned long _tabuDuration;
    unsigned long _rcsLengthParameter;

    Range _maximumMoveNumber;    // assumption: possible moves are
                                 //             numbered 1.._maximumMoveNumber
                                 //             (some of these numbers may not be used)

    vector<unsigned long> _tabuStatus;  // iteration number (=_nextRunningListPosition)
                                        // when move was last set tabu (status to be kept
                                        // active _tabuDuration iterations)

    vector<unsigned long> _residualListPred;
    vector<unsigned long> _residualListSucc;

    Observer<CNumeric> *_observer;

  public:
    REMTabuCriterion( Observer<CNumeric> *observer=0,
                      unsigned long tabuDuration=1,
                      unsigned long rcsLengthParameter=1,
                      unsigned long maximumMoveNumber
                        = numeric_limits<unsigned long>::max() );

    unsigned long addToHistory( const N& move, unsigned long iteration=0 );

    unsigned long addToHistory( const S& solution, unsigned long iteration=0 );

    bool tabu( const N& move ) const;  // tabu or not tabu ...

    void escape( );

    void print( ) const;

    void printRunningList( ) const;

    string description()
      { return "REMTabuCriterion"; }
};


template <class C>
REMTabuCriterion<C>::REMTabuCriterion
( Observer<CNumeric> *observer, unsigned long tabuDuration,
  unsigned long rcsLengthParameter, unsigned long maximumMoveNumber )
: _runningList( ),
  _nextRunningListPosition( 0 ),
  _tabuDuration( tabuDuration ),
  _rcsLengthParameter( rcsLengthParameter ),
  _maximumMoveNumber( maximumMoveNumber ),
  _tabuStatus( maximumMoveNumber+1, 0 ),
  _residualListPred( maximumMoveNumber+1 ),
  _residualListSucc( maximumMoveNumber+1 ),
  _observer( observer )
{
  if ( !observer )
    _observer = new Observer<CNumeric>( );

  _observer->tabuCriterionInitialized("REM" );
}


template <class C>
unsigned long REMTabuCriterion<C>::addToHistory( const N& move, unsigned long iteration )
{
  //  unsigned long numberOfNewMove = moveNumber<C::CNumeric>( move );
  unsigned long numberOfNewMove = moveNumber( move );

  _runningList.push_back( numberOfNewMove );

  if ( _observer->active() )
  {
    _observer->attributeAddedToHistory( );
    _observer->getStream() << move;
    _observer->flush();
  }

  ++_nextRunningListPosition;

  // initialization of residual list and tabu status
  unsigned long i;
  for( i=1; i<=_maximumMoveNumber; ++i )
    _residualListSucc[i] = i;  // means not in list

  unsigned long currentLengthOfResidualList = 0;
  unsigned long firstEntryOfResidualList = 0;  // 0 indicates no entry

  // succesively build residual list and set moves tabu that constitute a
  // residual list of length 1

  i = _nextRunningListPosition;
  do
  {
    --i;

    unsigned long moveNumber = _runningList[i];

    if ( _residualListSucc[moveNumber] != moveNumber )
    // moveNumber is already contained in the residual list,
    // that means we have two inverse moves,
    // so we have to delete this move from the residual list
    {
      unsigned long pred = _residualListPred[moveNumber];
      unsigned long succ = _residualListSucc[moveNumber];
      _residualListSucc[pred] = succ;
      _residualListPred[succ] = pred;
      _residualListSucc[moveNumber] = moveNumber;
      if ( moveNumber == firstEntryOfResidualList )
        firstEntryOfResidualList = succ;
      --currentLengthOfResidualList;
    }
    else
    // moveNumber is not contained in the residual list,
    // so we have to add this move to the residual list
    {
      _residualListSucc[moveNumber] = firstEntryOfResidualList;
      _residualListPred[firstEntryOfResidualList] = moveNumber;
      _residualListPred[moveNumber] = 0;
      firstEntryOfResidualList = moveNumber;
      ++currentLengthOfResidualList;
    }

    if ( currentLengthOfResidualList <= _rcsLengthParameter )
    // set all members of residual list tabu
    {
      unsigned long j = firstEntryOfResidualList;
      do
      {
      	_tabuStatus[j] = _nextRunningListPosition;
        // actually _nextRunningListPosition-1 would be better
        // but 0 is needed to mark not yet tabu moves
        // (see also the corresponding condition below in tabu())

        j = _residualListSucc[j];
      }
      while ( j != 0 );
    }
  }
  while( i != 0 );

  return 0;
}


template <class C>
unsigned long REMTabuCriterion<C>::addToHistory( const S& solution, unsigned long iteration )
{
  // nothing to do

  return 0;
}


template <class C>
inline
bool REMTabuCriterion<C>::tabu( const N& move ) const
{
  // check iff move.getMoveNumber was set tabu not longer than _tabuDuration steps ago
  // assumption: _nextRunningListPosition corresponds to iteration number

  //  return (    ( _tabuStatus[moveNumber<C::CNumeric>( move )] != 0 )
  //           && ( _nextRunningListPosition - _tabuStatus[moveNumber<C::CNumeric>( move )] < _tabuDuration ) );

  return (    ( _tabuStatus[moveNumber( move )] != 0 )
           && ( _nextRunningListPosition - _tabuStatus[moveNumber( move )] < _tabuDuration ) );
}


template <class C>
void REMTabuCriterion<C>::escape( )
{
  _runningList.erase( _runningList.begin(), _runningList.end() );
  _nextRunningListPosition = 0;

  for( unsigned long i=1; i<=_maximumMoveNumber; ++i )
    _tabuStatus[i] = 0;
}


template <class C>
void REMTabuCriterion<C>::print( ) const
{
  if ( _observer->active() )
  {
    _observer->currentTabuList( );

    for( unsigned long i=1; i<=_maximumMoveNumber; ++i )
      _observer->getStream() << i << ":" << _tabuStatus[i] << " ";
    _observer->getStream() << "\n";

    _observer->flush();
  }
}


template <class C>
void REMTabuCriterion<C>::printRunningList( ) const
{
  if ( _observer->active() )
  {
    _observer->currentRunningList( );
  
    typename vector<Range>::iterator it = _runningList.begin();
    while( it != _runningList.end() )
      _observer->getStream() << *it << " ";
    _observer->getStream() << "\n";

    _observer->flush();
  }
}

// ----------------------------------------------------------------------------
// -- REMM Tabu Criterion -----------------------------------------------------
// ----------------------------------------------------------------------------

// multi attribute moves, incomplete and experimental ...

template <class C>
class REMMTabuCriterion
: public TabuCriterion<C>
{
  public:
    typedef typename C::T T;
    typedef typename C::Range Range;
    typedef typename C::CNumeric CNumeric;
    typedef typename C::S S;
    typedef typename C::N N;
    typedef typename C::S_A S_A;
    typedef typename C::S_I S_I;

  private:
    vector<Range> _runningList;
    Range _nextRunningListPosition;  // stores position of next move to be included

    Range _rcsLengthParameter;
    Range _tabuDuration;

    Range _maximumNumberOfAttributes;  // assumption: attributes are numbered 1.._maximumNumberOfAttributes
                                       // (some of these numbers may not be used)

    vector<Range> _tabuStatus;  // iteration number (=_nextRunningListPosition) when move was
                                // last set tabu (status to be kept active _tabuDuration iterations)

    vector<Range> _residualListPred;
    vector<Range> _residualListSucc;

    Observer<CNumeric> *_observer;

  public:
    REMMTabuCriterion( Observer<CNumeric> *observer,
                       Range maximumNumberOfAttributes, Range rcsLengthParameter=1, Range tabuDuration=1 );

    unsigned long addToHistory( const N& move, unsigned long iteration );

    unsigned long addToHistory( const S& solution, unsigned long iteration=0 );

    bool tabu( const N& move ) const;  // tabu or not tabu ...

    void escape( );

    void print( ) const;

    void printRunningList( ) const;

    void escription()
      { return "REMMTabuCriterion"; }
};


template <class C>
REMMTabuCriterion<C>::REMMTabuCriterion
( Observer<CNumeric> *observer,
  Range maximumNumberOfAttributes, Range rcsLengthParameter, Range tabuDuration )
: _runningList( ),
  _nextRunningListPosition( 0 ),
  _rcsLengthParameter( rcsLengthParameter ),
  _tabuDuration( tabuDuration ),
  _maximumNumberOfAttributes( maximumNumberOfAttributes ),
  _tabuStatus( maximumNumberOfMoves+1, 0 ),
  _residualListPred( maximumNumberOfMoves+1 ),
  _residualListSucc( maximumNumberOfMoves+1 ),
  _observer( observer )
{
  _observer->tabuCriterionInitialized( "REM multi-attribute" );
}


template <class C>
unsigned long REMMTabuCriterion<C>::addToHistory( const N& move, unsigned long iteration )
{
  Range numberOfNewMove = move.getMoveNumber( );

  _runningList.push_back( numberOfNewMove );

  ++_nextRunningListPosition;

  // initialization of residual list and tabu status
  Range i;
  for( i=1; i<=_maximumNumberOfMoves; ++i )
    _residualListSucc[i] = i;  // means not in list

  Range currentLengthOfResidualList = 0;
  Range firstEntryOfResidualList = 0;  // 0 indicates no entry

  // succesively build residual list and set moves tabu that constitute a
  // residual list of length 1

  i = _nextRunningListPosition;
  do
  {
    --i;

    Range moveNumber = _runningList[i];

    if ( _residualListSucc[moveNumber] != moveNumber )
    // moveNumber is already contained in the residual list,
    // that means we have two inverse moves,
    // so we have to delete this move from the residual list
    {
      Range pred = _residualListPred[moveNumber];
      Range succ = _residualListSucc[moveNumber];
      _residualListSucc[pred] = succ;
      _residualListPred[succ] = pred;
      _residualListSucc[moveNumber] = moveNumber;
      if ( moveNumber == firstEntryOfResidualList )
        firstEntryOfResidualList = succ;
      --currentLengthOfResidualList;
    }
    else
    // moveNumber is not contained in the residual list,
    // so we have to add this move to the residual list
    {
      _residualListSucc[moveNumber] = firstEntryOfResidualList;
      _residualListPred[firstEntryOfResidualList] = moveNumber;
      _residualListPred[moveNumber] = 0;
      firstEntryOfResidualList = moveNumber;
      ++currentLengthOfResidualList;
    }

    if ( currentLengthOfResidualList <= _rcsLengthParameter )
    // set all members of residual list tabu
    {
      Range j = firstEntryOfResidualList;
      do
      {
      	_tabuStatus[j] = _nextRunningListPosition;
        // actually _nextRunningListPosition-1 would be better
        // but 0 is needed to mark not yet tabu moves
        // (see also the corresponding condition below in tabu())

        j = _residualListSucc[j];
      }
      while ( j != 0 );
    }
  }
  while( i != 0 );

  return 0;
}


template <class C>
unsigned long REMMTabuCriterion<C>::addToHistory( const S& solution, unsigned long iteration )
{
  // nothing to do

  return 0;
}


template <class C>
inline
bool REMMTabuCriterion<C>::tabu( const N& move ) const
{
  // check iff move.getMoveNumber was set tabu not longer than _tabuDuration steps ago
  // assumption: _nextRunningListPosition corresponds to iteration number

  return (    ( _tabuStatus[move.getMoveNumber()] != 0 )
           && ( _nextRunningListPosition - _tabuStatus[move.getMoveNumber()] < _tabuDuration ) );

}


template <class C>
void REMMTabuCriterion<C>::escape( )
{
}


template <class C>
void REMMTabuCriterion<C>::print( ) const
{
  if ( _observer->active() )
  {
    _observer->currentTabuList( );

    for( Range i=1; i<=_maximumNumberOfMoves; ++i )
      _observer->getStream() << i << ":" << _tabuStatus[i] << " ";
    _observer->getStream() << "\n";

    _observer->flush();
  }
}


template <class C>
void REMMTabuCriterion<C>::printRunningList( ) const
{
  if ( _observer->active() )
  {
    _observer->currentRunningList( );
  
    typename vector<Range>::iterator it = _runningList.begin();
    while( it != _runningList.end() )
      _observer->getStream() << *it << " ";
    _observer->getStream() << "\n";

    _observer->flush();
  }
}


// ----------------------------------------------------------------------------
// --  Traversed Solution Class -----------------------------------------------
// ----------------------------------------------------------------------------

 
template <class S_I>
class TraversedSolution
{
  private:
    S_I _solutionInfo;
    mutable unsigned long _lastTraversed;
    mutable unsigned long _numberOfTraverses;

  public:
    TraversedSolution( const S_I& solutionInfo, unsigned long iteration = 0 )
      : _solutionInfo( solutionInfo ), _lastTraversed( iteration ), _numberOfTraverses( 1 )
      { ; }

    bool operator<( const TraversedSolution& rhs ) const
      { return ( _solutionInfo < rhs.getSolutionInfo() ); }

    bool operator==( const TraversedSolution& rhs ) const
      { return ( _solutionInfo == rhs.getSolutionInfo() ); }

    S_I getSolutionInfo( ) const
      { return _solutionInfo; }

    unsigned long getLastTraversed( ) const
      { return _lastTraversed; }
    void setLastTraversed( unsigned long lastTraversed ) const
      { _lastTraversed = lastTraversed; }

    unsigned long getNumberOfTraverses( ) const
      { return _numberOfTraverses; }
    void incrementNumberOfTraversesByOne( ) const
      { ++_numberOfTraverses; }
};


// ----------------------------------------------------------------------------
// -- StrictTabuCriterionByTrajectory -----------------------------------------
// ----------------------------------------------------------------------------

template <class C>
class StrictTabuCriterionByTrajectory
: public TabuCriterion<C>
{
  public:
    typedef typename C::T T;
    typedef typename C::Range Range;
    typedef typename C::CNumeric CNumeric;
    typedef typename C::S S;
    typedef typename C::N N;
    typedef typename C::S_A S_A;
    typedef typename C::S_I S_I;

  protected:
   // TODO: think about a map instead of a set 
    set<TraversedSolution<S_I>, less<TraversedSolution<S_I> > > _trajectory;

    Observer<CNumeric> *_observer;

  public:
    StrictTabuCriterionByTrajectory( Observer<CNumeric> *observer = 0 );

    unsigned long addToHistory( const N& move, unsigned long iteration );

    unsigned long addToHistory( const S& solution, unsigned long iteration );

    bool tabu( const N& move ) const;  // tabu or not tabu ...

    void escape( );  // diversification, so reset tabu memory

    void print( ) const;

    string description()
      { return "SimpleTrajectoryTabuCriterion"; }
};


template <class C>
StrictTabuCriterionByTrajectory<C>
  ::StrictTabuCriterionByTrajectory( Observer<CNumeric> *observer )
: _observer( observer )
{
  if ( !observer )
    _observer = new Observer<CNumeric>( );

  _observer->tabuCriterionInitialized( "trajectory" );
}


template <class C>
unsigned long StrictTabuCriterionByTrajectory<C>::addToHistory( const N& move,
                                                                unsigned long iteration )
{
  // nothing to do

  return 0;
}


template <class C>
unsigned long StrictTabuCriterionByTrajectory<C>
              ::addToHistory( const S& solution, unsigned long iteration )
{
  if ( _observer->active() )
    _observer->solutionInfoAddedToHistory( );

  // the following line has been replaced by two lines due to some gcc295 problems
  //   TraversedSolution<S_I> traversedSolution( S_I( solution ), iteration );
  S_I solutionInfo( solution );
  TraversedSolution<S_I> traversedSolution( solutionInfo, iteration );

  typename set<TraversedSolution<S_I>, less<TraversedSolution<S_I> > >::iterator
  it = _trajectory.find( traversedSolution );

  if ( it != _trajectory.end() )
  {
    (*it).setLastTraversed( iteration );
    (*it).incrementNumberOfTraversesByOne( );

    if ( _observer->active() )
      _observer->getStream() << *it;
  }
  else
  {
    _trajectory.insert( traversedSolution );

    if ( _observer->active() )
      _observer->getStream() << traversedSolution;
  }

  if ( _observer->active() )
    _observer->flush();

  return 0;
}


template <class C>
bool StrictTabuCriterionByTrajectory<C>::tabu( const N& move ) const
{
  return ( _trajectory.end() != _trajectory.find( TraversedSolution<S_I>( S_I( move ) ) ) );
}


template <class C>
void StrictTabuCriterionByTrajectory<C>::escape( )
{
  _trajectory.erase( _trajectory.begin(), _trajectory.end() );
}


template <class C>
void StrictTabuCriterionByTrajectory<C>::print( ) const
{
  if ( _observer->active() )
  {
    _observer->currentTabuList( );

    for( typename set<TraversedSolution<S_I>, less<TraversedSolution<S_I> > >::const_iterator
         it = _trajectory.begin(); it != _trajectory.end(); ++it )
      _observer->getStream() << *it;

    _observer->flush();
  }
}


// ----------------------------------------------------------------------------
// -- Reactive Tabu Criterion  ------------------------------------------------
// ----------------------------------------------------------------------------

template <class C>
class ReactiveTabuCriterion
: public TabuCriterion<C>
{
  public:
    typedef typename C::T T;
    typedef typename C::Range Range;
    typedef typename C::CNumeric CNumeric;
    typedef typename C::S S;
    typedef typename C::N N;
    typedef typename C::S_A S_A;
    typedef typename C::S_I S_I;

  protected:
    vector<S_A> _tabuList;
    Range _currentListPosition;  // stores position of last attribute included
    Range _currentTabuListLength;
    Range _tabuThreshold;

    set<TraversedSolution<S_I>, less<TraversedSolution<S_I> > > _trajectory;
    set<TraversedSolution<S_I>, less<TraversedSolution<S_I> > > _oftenRepeatedSolutions;

    unsigned long _lastReaction;
    unsigned long _movingAverage;

    float _adaptationParameter1;
    short _adaptationParameter2;

    unsigned int _chaosParameter;
    unsigned int _repetitionParameter;  

    float _smoothingParameter;

    void extendTabuList( );
    void shortenTabuList( );

    Observer<CNumeric> *_observer;

  public:
    ReactiveTabuCriterion( Observer<CNumeric> *observer=0,
                           Range tabuThreshold=1,
                           float adaptationParameter1=1.2,
                           short adaptationParameter2=2,
                           Range maxTabuListLength=50,
                           unsigned int chaosParameter=3,
                           unsigned int repetitionParameter=3,
                           float smoothingParameter = 0.5 );

    unsigned long addToHistory( const N& move, unsigned long iteration );

    unsigned long addToHistory( const S& solution, unsigned long iteration );

    bool tabu( const N& move ) const;  // tabu or not tabu ...

    void escape( );  // diversification, so reset tabu memory

    void print( ) const;

    string description()
      { return "ReactiveTabuCriterion"; }
};


template <class C>
ReactiveTabuCriterion<C>
  ::ReactiveTabuCriterion( Observer<CNumeric> *observer,
                           Range tabuThreshold, // =1
                           float adaptationParameter1, short adaptationParameter2, // = 1.2, =2
                           Range maxTabuListLength, unsigned int chaosParameter, // =50, =3
                           unsigned int repetitionParameter, float smoothingParameter ) // =3,=0.5
  : _tabuList( maxTabuListLength ), _currentListPosition( 0 ), _currentTabuListLength( 1 ),
    _tabuThreshold( tabuThreshold ), _lastReaction( 0 ), _movingAverage( 0 ),
    _adaptationParameter1( adaptationParameter1 ), _adaptationParameter2( adaptationParameter2 ), 
    _chaosParameter( chaosParameter ), _repetitionParameter( repetitionParameter ),
    _smoothingParameter( smoothingParameter ), _observer( observer )
{
  if ( !observer )
    _observer = new Observer<CNumeric>( );

  _tabuList[_currentListPosition] = S_A( );  // dummy move

  _observer->tabuCriterionInitialized( "reactive" );
}


template <class C>
unsigned long ReactiveTabuCriterion<C>::addToHistory( const N& move, unsigned long iteration )
{
  for( Range i=1; i<=S_A::getNumberOfPositiveAttributes( move ); ++i )
  {
    ++_currentListPosition;
    if ( _currentListPosition == _currentTabuListLength )
      _currentListPosition = 0;

    _tabuList[_currentListPosition] = S_A( move, i );

    if ( _observer->active() )
    {
      _observer->attributeAddedToHistory( );
      _observer->getStream() << _tabuList[_currentListPosition];
      _observer->flush();
    }
  }

  return 0;
}


template <class C>
unsigned long ReactiveTabuCriterion<C>
              ::addToHistory( const S& solution, unsigned long iteration )
{
  if ( _observer->active() )
    _observer->solutionInfoAddedToHistory( );

  // the following line has been replaced by two lines due to some gcc295 problems
  //  TraversedSolution<S_I> traversedSolution( S_I( solution ), iteration );
  S_I solutionInfo( solution );
  TraversedSolution<S_I> traversedSolution( solutionInfo, iteration );


  typename set<TraversedSolution<S_I>, less<TraversedSolution<S_I> > >::iterator
    it = _trajectory.find( traversedSolution );

  if ( it != _trajectory.end() )
  // solution already visited, so update this, extend tabu list size, and update parameters
  {
    if ( _movingAverage == 0 )
      _movingAverage = static_cast<unsigned long>( ( iteration - (*it).getLastTraversed() ) / 2 );
    else
      _movingAverage = static_cast<unsigned long>( 
                         _smoothingParameter*(iteration-(*it).getLastTraversed())
                         + (1-_smoothingParameter)*_movingAverage );

    _observer->newMovingAverage( _movingAverage );

    (*it).setLastTraversed( iteration );
    (*it).incrementNumberOfTraversesByOne( );

    if ( _observer->active() )
    {
      _observer->getStream() << *it;
      _observer->flush();
    }
       
    extendTabuList( );
    _lastReaction = iteration;

    if ( (*it).getNumberOfTraverses( ) == _chaosParameter )
    {
      _oftenRepeatedSolutions.insert( traversedSolution );
      if ( _oftenRepeatedSolutions.size() >= _repetitionParameter )
        return ( _movingAverage > 0 ? _movingAverage : 7 );  // ESCAPE!
    }
  }
  else
  // solution not yet visited, so add it into _trajectory
  // and possibly shorten tabu list size (if some iterations without repetitions done)
  {
    _trajectory.insert( traversedSolution );

    if ( _observer->active() )
    {
      _observer->getStream() << traversedSolution;
      _observer->flush();
    }

    if (    ( iteration - _lastReaction > _movingAverage )
         && ( _lastReaction != 0 ) )
    {
      shortenTabuList( );
      _lastReaction = iteration;
    }
  }

  return 0;
}


template <class C>
bool ReactiveTabuCriterion<C>::tabu( const N& move ) const
{
  Range numberOfAttributesFound = 0;

  Range tempListPosition = _currentListPosition;
  do
  {
    if ( _tabuList[tempListPosition] == move )
      ++numberOfAttributesFound;
    // TODO: think about attributes that are more than one time in the tabu list
    //       should this case be treated in a special way?

    ++tempListPosition;
    if ( tempListPosition == _currentTabuListLength )
      tempListPosition = 0;
  }
  while( tempListPosition != _currentListPosition );

  if ( numberOfAttributesFound >= min( int(_tabuThreshold),
                                       int (S_A::getNumberOfNegativeAttributes( move ) ) ) )
    return true;
  else
    return false;
}


template <class C>
void ReactiveTabuCriterion<C>::extendTabuList( )
{
  Range oldTabuListLength = _currentTabuListLength;
  _currentTabuListLength = min( max ( Range(_adaptationParameter1*_currentTabuListLength),
                                      Range(_currentTabuListLength+_adaptationParameter2) ),
                                Range( _tabuList.size() ) );

  if ( oldTabuListLength < _currentTabuListLength )
  // shift some old moves to the back, set left positions as empty moves
  {
    Range diff = _currentTabuListLength - oldTabuListLength;

    Range i;
    
    for( i=_currentTabuListLength-1; ( ( i > _currentListPosition+diff ) && ( i>= _currentTabuListLength-diff) ); --i )
      _tabuList[i] = _tabuList[i-diff];

    for( i=oldTabuListLength; i<=oldTabuListLength+diff-1; ++i )
      _tabuList[i] = S_A( );

    _observer->tabuListLengthExtended( _currentTabuListLength );
  }
}


template <class C>
void ReactiveTabuCriterion<C>::shortenTabuList( )
{
  Range oldTabuListLength = _currentTabuListLength;
  _currentTabuListLength = max( min ( Range(_currentTabuListLength/_adaptationParameter1),
                                      Range(_currentTabuListLength-_adaptationParameter2) ),
                                Range( 1 ) );

  if ( oldTabuListLength > _currentTabuListLength )
  {
    if ( _currentListPosition >= _currentTabuListLength )
    {
      Range shift = _currentListPosition - _currentTabuListLength + 1;
      
      if ( shift < _currentTabuListLength )
      // all overlapping attributes can be shifted to the front
      {
        for( Range i=0; i<=shift-1; ++i )
          _tabuList[i] = _tabuList[_currentTabuListLength + i];
        _currentListPosition = shift-1;
      }
      else
      // only the most recent overlapping attributes can be shifted to the front
      {
        for( Range i=0; i<=_currentTabuListLength-1; ++i )
          _tabuList[i] = _tabuList[_currentListPosition-_currentTabuListLength+1 + i];
        _currentListPosition = _currentTabuListLength-1;
      }
    }      
    // else part is skipped by intention; i.e. we do not shift attributes from
    // the old end to the new end (which would be required to keep the recency
    // structure of the attributes in the tabu list) but we keep the list as it is
    // (in order to provoke some more variety)

    _observer->tabuListLengthShortened( _currentTabuListLength );
  }
}


template <class C>
void ReactiveTabuCriterion<C>::escape( )
{
  _currentTabuListLength = 1;
  _currentListPosition = 0;
  _tabuList[_currentListPosition] = S_A( );  // dummy move
  _movingAverage = 0;
  _lastReaction = 0;  // 0 means no last reaction

  _trajectory.erase( _trajectory.begin(), _trajectory.end() );
  _oftenRepeatedSolutions.erase( _oftenRepeatedSolutions.begin(), _oftenRepeatedSolutions.end() );
}


template <class C>
void ReactiveTabuCriterion<C>::print( ) const
{
  if ( _observer->active() )
  {
    // print tabu list
    _observer->currentTabuList( );

    Range tempListPosition = _currentListPosition;
    for( Range i=1; i<=_currentTabuListLength; ++i )
    {
      _observer->getStream() << _tabuList[tempListPosition];
      if ( tempListPosition == 0 )
        tempListPosition = _currentTabuListLength-1;
      else
        --tempListPosition;
    }

    _observer->flush();

    // print trajectory
    _observer->currentTrajectory( );

    for( typename set<TraversedSolution<S_I>, less<TraversedSolution<S_I> > >::const_iterator
           it = _trajectory.begin(); it != _trajectory.end(); ++it )
      _observer->getStream() << *it;

    _observer->flush();
  }
}


template <class S_I>
ostream& operator<<( ostream& output, const TraversedSolution<S_I>& traversedSolution )
{
  output << traversedSolution.getSolutionInfo()
         << " [l=" << traversedSolution.getLastTraversed()
         << ",#=" << traversedSolution.getNumberOfTraverses()
         << "] " << "\n";

  return output;
}


}
//--- end of namespace HotFrame -----------------------------------------------
//------------------------------------------------------------------------------


#endif
