// MTSP_P: class to read, generate, write and provide multiple traveling salesman like problems (mtsp)

#if !defined(__MTSP_P__)
#define __MTSP_P__

#include "basic.h"
#include "matrix.h"


//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
namespace HotFrame
{


template <class C>
class MTSP_P
{
  public:
    typedef typename C::T T;
    typedef typename C::Range Range;

  protected:

    Range _n; // number of objects (numbering of objects: i = 1, ...,n)
    Range _m; // number of resources (numbering of resources: j = 1, ...,m)

    Matrix<T, Range> _cost;

    T _maxCost;

    string _name;  // of problem instance, may be empty

  public:

    MTSP_P( istream& input );

    MTSP_P( Range n, Range m, T minCost, T maxCost );  // construct uniform random cost matrix

    MTSP_P( const MTSP_P<C>& original )
      : _n( original._n ), _m( original._m ),
        _cost( original._cost ), _name( original._name )
      { };

    void print( ostream& output ) const;

    Range numberOfObjects( ) const
      { return _n; }

    Range numberOfResources( ) const
      { return _m; }

    T getCost( Range i, Range j ) const
      {
        assert( (i>=0) && (i<=_n+1) );
        assert( (j>=0) && (j<=_n+1) );
        if ( (i==0) || (i==_n+1) || ( j==0) || (j==_n+1) )
          return 0;
        else
          return _cost( i, j );
      }

    T maxCost( ) const
      { return _maxCost; }

    MTSP_P<C>& operator=( const MTSP_P<C>& rhs )
    {
      if ( this == &rhs )
        return *this;
      _n = rhs._n;
      _m = rhs._m;
      _cost = rhs._cost;
      _maxCost = rhs._maxCost;
      _name = rhs._name;
      return *this;
    }
};


template<class C>
MTSP_P<C>::MTSP_P( istream& input )
: _cost( 0 ), _maxCost( 0 )
{
  // format: _n _m <CR>
  //         c[11] c[12] ... c[1n] <CR>
  //         ...
  //         c[n1] c[n2] ... c[nn] <CR>

  input >> _n;  // number of objects

  input >> _m;  // number of resources

  _cost.assign( _n );

  for( Range i=1; i<=_n; ++i )
    for( Range j=1; j<=_n; ++j )
    {
      T data;
      input >> data;
      _cost( i, j ) = data;

      if ( ( data > _maxCost ) && ( i != j ) )
        _maxCost = data;
    }
}


template<class C>
MTSP_P<C>::MTSP_P( Range n, Range m, T minCost, T maxCost )
: _n( n ), _m( m ), _cost( n ), _maxCost( 0 )
{
  // pseudo-random generation of a problem instance with n objects
  // integer cost values times are sampled from a rectangular distribution from minCost to maxCost
  // (e.g.: 1,99 means: sample randomly from 1,2,3,...,99)

  for( i=1; i<=_n; ++i )
    for( j=1; j<=_n; ++j )
    {
      _cost( i, j ) = minTime + rand( maxCost - minCost + 1 );
      if ( _cost( i, j ) > _maxCost )
        _maxCost = _cost( i, j );
    }
}


template<class C>
void MTSP_P<C>::print( ostream& output ) const
{
  // format: _n _m <CR>
  //         c[11] c[12] ... c[1n] <CR>
  //         ...
  //         c[n1] c[n2] ... c[nn] <CR>

  output << _n << " " << _m << "\n";

  for( Range i=1; i<=_n; ++i )
  {
    for( Range j=1; j<=_n; ++j )
      output << getCost( i, j ) << " ";
    output << "\n";
  }
}


template <class C>
ostream& operator<<( ostream& output, const MTSP_P<C>& problem )
{
  problem.print( output );
  return output;
}


}
//--- end of namespace HotFrame -----------------------------------------------
//------------------------------------------------------------------------------


#endif
