// MTSP_S: class to manage solutions of multiple traveling salesman like problems (mtsp)


#if !defined(__HotFrame_MTSP_S__)
#define __HotFrame_MTSP_S__


#include "basic.h"
#include "mtsp_p.h"
#include "as_s.h"


//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
namespace HotFrame
{


template <class C> class MTSP_S;  // forward


// declaration of class MTSP_S -------------------------------------------------

template <class C>
class MTSP_S
: public AS_S<C>
{
  protected:
    MTSP_P<C>& _problem;
    T _maximumLoad;
    vector<T> _currentLoad;

  public:
    enum FirstSolution { NoAssignment=0 };

    MTSP_S( MTSP_P<C>& problem,
            Observer<C>* observer,
            FirstSolution firstSolution = NoAssignment );

    virtual void evaluate( );

    virtual bool computeEvaluation( const Generic_S_N<C>& neighbor, T& evaluation, T& delta );

    // some technical stuff:
    
    virtual Generic_S<C>* clone() const
      { return new MTSP_S<C>( *this ); }

    virtual Generic_S<C>& operator=( const Generic_S<C>& rhs )
    {
      if ( this != &rhs )
      {
        assert( typeid(rhs) == typeid(*this) );
        const MTSP_S<C>& drhs = static_cast<const MTSP_S<C>&>( rhs );
        AS_S<C>::operator=( rhs ); 
        _problem = drhs._problem;
        _maximumLoad = drhs._maximumLoad;
        _currentLoad = drhs._currentLoad;
      }  
      return *this;
    }

    MTSP_S<C>& operator=( const MTSP_S<C>& rhs )
    {
      if ( this != &rhs )
        operator=( static_cast<const Generic_S<C>&>( rhs ) ); 
      return *this;
    }
};


// element functions of class MTSP_S<C> ---------------------------------

template<class C>
MTSP_S<C>::MTSP_S( MTSP_P<C>& problem,
                   Observer<C>* observer,
                   FirstSolution firstSolution )
: AS_S<C>( problem.numberOfResources(), observer ),
  _problem( problem ),
  _currentLoad( problem.numberOfObjects() + 1, 0 )
{
  if ( firstSolution == NoAssignment )
  {
    for( Range i=1; i<=problem.numberOfObjects(); ++i )
      addObject( i );
  } 
  else
    assert( false );

  evaluate( );

  _s_lowerBound = 0;  // ?
  _s_upperBound = _evaluation;
}


template<class C>
void MTSP_S<C>::evaluate( )
{
  _maximumLoad = 0;

  for( Range i=1; i<=numberOfResources(); ++i )
  {
    _currentLoad[i] = 0;
    for( Range j=1; j<=numberOfObjectsAssignedToResource( i )-1; ++j )
      _currentLoad[i] += _problem.getCost( objectOfResourceAtPosition( i, j ), 
                                           objectOfResourceAtPosition( i, j+1 ) );
    if ( _currentLoad[i] > _maximumLoad )
      _maximumLoad = _currentLoad[i];
  }

  _evaluation = _maximumLoad;

  if ( numberOfObjectsAssignedToResource( 0 ) > 0 )
  // penalty costs
  {
    _currentLoad[0] = 0;
    for( Range j=1; j<=numberOfObjectsAssignedToResource( 0 )-1; ++j )
      _currentLoad[0] += _problem.getCost( objectOfResourceAtPosition( 0, j ), 
                                           objectOfResourceAtPosition( 0, j+1 ) );
    _evaluation += numberOfObjects() * (1+_currentLoad[0]) * _problem.maxCost();
  }
}

// directly compute some evaluation of a considered move
// (without this method, the evaluate() would be used to compute the objective
//  function value of the neighbor solution and the difference to the 
//  objective function value of the current solution gives the move evaluation)
template<class C>
bool MTSP_S<C>::computeEvaluation( const Generic_S_N<C>& neighbor, T& evaluation, T& delta )
{
  evaluation = 0;

  // evaluation of a shift neighbor by upcasting
  {
    const AS_S_N_Shift<C> *move = dynamic_cast<const AS_S_N_Shift<C> *>(&neighbor);
    if ( move != NULL )
    {
      T newLoadOfResourceFrom = _currentLoad[move->resourceFrom()];
      T newLoadOfResourceTo = _currentLoad[move->resourceTo()];
      
      // deleted edges

      if ( move->positionFrom() >= 2 )
        newLoadOfResourceFrom -= _problem.getCost(
           objectOfResourceAtPosition( move->resourceFrom(), move->positionFrom()-1 ),
           objectOfResourceAtPosition( move->resourceFrom(), move->positionFrom()   ) );           

      if ( move->positionFrom() <= numberOfObjectsAssignedToResource( move->resourceFrom() ) - 1 )
        newLoadOfResourceFrom -= _problem.getCost(
           objectOfResourceAtPosition( move->resourceFrom(), move->positionFrom()   ),
           objectOfResourceAtPosition( move->resourceFrom(), move->positionFrom()+1 ) );           
        
      if (    ( move->positionTo() >= 2 )
           && ( move->positionTo() <= numberOfObjectsAssignedToResource( move->resourceTo() ) ) )
        newLoadOfResourceTo -= _problem.getCost(
           objectOfResourceAtPosition( move->resourceTo(), move->positionTo()-1   ),
           objectOfResourceAtPosition( move->resourceTo(), move->positionTo()   ) );           
        
      // inserted edges

      if (    ( move->positionFrom() >= 2 )
           && ( move->positionFrom() <= numberOfObjectsAssignedToResource( move->resourceFrom() ) - 1 ) )
        newLoadOfResourceFrom += _problem.getCost(
           objectOfResourceAtPosition( move->resourceFrom(), move->positionFrom()-1 ),
           objectOfResourceAtPosition( move->resourceFrom(), move->positionFrom()+1 ) );           
        
      if ( move->positionTo() >= 2 )
        newLoadOfResourceTo += _problem.getCost(
           objectOfResourceAtPosition( move->resourceTo(), move->positionTo()-1 ),
           objectOfResourceAtPosition( move->resourceFrom(), move->positionFrom() ) );           
        
      if ( move->positionTo() <= numberOfObjectsAssignedToResource( move->resourceTo() ) )
        newLoadOfResourceTo += _problem.getCost(
           objectOfResourceAtPosition( move->resourceFrom(), move->positionFrom() ),
           objectOfResourceAtPosition( move->resourceTo(), move->positionTo() ) );           
        
      if ( move->resourceFrom() == move->resourceTo() )
        newLoadOfResourceFrom = newLoadOfResourceTo
          = ( newLoadOfResourceFrom + newLoadOfResourceTo ) / 2;
        
      // determine new maximum load
      
      T adaptedMaximumLoad = 0;
      for( Range i=1; i<=numberOfResources(); ++i )
      {
        if ( i == move->resourceFrom() )
        {
          if ( newLoadOfResourceFrom > adaptedMaximumLoad )
            adaptedMaximumLoad = newLoadOfResourceFrom;
        }
        else if ( i == move->resourceTo() )
        {
          if ( newLoadOfResourceTo > adaptedMaximumLoad )
            adaptedMaximumLoad = newLoadOfResourceTo;
        }
        else
          if ( _currentLoad[i] > adaptedMaximumLoad )
            adaptedMaximumLoad = _currentLoad[i];
      }
      
      if ( adaptedMaximumLoad != _maximumLoad )
        evaluation = _maximumLoad - adaptedMaximumLoad;
      else
      {
        // consider the one or two resources with modified loads:
        // favor some kind of load reduction aiming for a possibility to
        // eventually reduce the maximum load by means of successive moves

        T loadReduction;
        if ( move->resourceFrom() == move->resourceTo() )
          evaluation = _currentLoad[move->resourceFrom()] - newLoadOfResourceFrom;
          // we should scale this value as below (but this lead to worse results ...?)
        else
        {
          loadReduction = min ( _currentLoad[move->resourceFrom()], _currentLoad[move->resourceTo()] )
                          - min ( newLoadOfResourceFrom, newLoadOfResourceTo );

          // the use of loadReduction as move evaluation requires some kind of
          // scaling so that we do not supersede a move that reduces the maximum load

          // ToDo: do this better (here we assume integral costs and do some hack ...)

          if ( loadReduction > 0 )
            evaluation = 1.001 - 1/loadReduction;
          else if ( loadReduction < 0 )
            evaluation = -1.001 + 1/loadReduction;
          else
            evaluation = 0;
        }
      }

      if ( move->resourceFrom() == 0 )
        evaluation += 3 * _problem.maxCost();
        //        evaluation += 3 * ( 1 + _currentLoad[0] - newLoadOfResourceFrom ) * _problem.maxCost();
      else if ( move->resourceTo() == 0 )
        evaluation += 3 * ( -1 + _currentLoad[0] - newLoadOfResourceTo ) * _problem.maxCost();
    }
    else
      assert( false );
  }

  return false;  // i.e., delta not computed
}


// stream output functions -----------------------------------------------------

template <class C>
ostream& operator<<( ostream& output, const MTSP_S<C>& solution )
{
  solution.print( output );
  return output;
}



}
//--- end of namespace HotFrame -----------------------------------------------
//------------------------------------------------------------------------------


#endif
