// main file for experimenting with with multiple TSPs (mtsp)
// (this is just on arbitrary ad hoc program used for some experiments,
//  it may serve as an example on how to access some of the functionality
//  of HotFrame)

// some general includes

#include "basic.h"
#include "observer.h"

// include metaheuristic components

#include "search.h"
#include "search_configs.h"

#include "sa.h"
#include "sa_configs.h"

#include "tabu.h"
#include "tabu_configs.h"

// include problem-specific components

#include "mtsp_s.h"


// all HotFrame components are defined in a corresponding namespace

using namespace HotFrame;


// uncomment the following line when switching from interactive debugging to
//   command line calls (this definition is only used in this main file)

// #define COMMANDLINE

/*
   command line parameters (or see below):
   1: file name of problem instance
   2: file name of debug file
   3: heuristic method (0,1,..., see below)
   4: tabu search criterion (0,1,..., see below)
   5: max time in seconds (0 means no restriction)
   6: max number of iterations (0 means no restriction)
   7: neighborhood depth (not always meaningful)
*/


// ----- Configurations --------------------------------------------------------

struct CNumeric
{
  typedef double T;   // primary data type (e.g., costs are of type float)
  typedef int Range;  // primary range type (e.g., int as range for numbering of elements)
};

template<class C>
struct CpTemplateMTSP
{
  typedef typename C::T T;
  typedef typename C::Range Range;
  typedef C CNumeric;

  typedef MTSP_P<CNumeric>       P;    // problem type
  typedef MTSP_S<CNumeric>       S;    // solution space
  typedef AS_S_N_Shift<CNumeric> N;    // neighborhood
  typedef AS_S_A1<CNumeric>      S_A;  // solution / move attributes
  typedef AS_S_I<CNumeric>       S_I;  // solution information
};
typedef CpTemplateMTSP<CNumeric> Cp;



// ----- main ------------------------------------------------------------------


int main( int argc, char* argv[] )
{
  Cp::S::_s_positionIsRelevant = true;


  // problem constructor argument (e.g. file name of file with problem instance)

#ifdef COMMANDLINE
  ifstream cproblem( argv[1] );
#else
//    ifstream cproblem( "ft53.mtsp" ); 
    ifstream cproblem( "ex1.mtsp" ); 
#endif

  Cp::P problem( cproblem );


  // file name of debug file

#ifdef COMMANDLINE
  ofstream cdebug( argv[2] );
#else
  ofstream cdebug( "debug.out" );
#endif

  // observer may write to cdebug or cout or ...

  Observer<CNumeric> * observer
//   = new Observer<CNumeric>( );
   = new StreamingObserver<CNumeric>( cout );
//   = new JustTheBestEvaluationToStream<CNumeric>( cdebug );
//   = new JustTheBestEvaluationAndTheCPUTimeToStream<CNumeric>( cout );

//  observer->deactivate();

  observer->currentProblem();
  observer->getStream() << problem;
  observer->flush();

  // create initial solution
  Cp::S solution( problem, observer, Cp::S::NoAssignment ); 

  observer->currentSolution();
  observer->getStream() << solution;
  observer->flush();


  // improving method (not all methods may be available here)
  //                     0       1   2   3     4   5    6     7     8       9    10  11  12  13    14  15     16
  enum HeuristicMethod { noim=0, sd, fd, sdma, rw, isd, sdid, fdid, sdmaid, saj, sa, ta, gd, rtrt, ts, tspen, ea };
#ifdef COMMANDLINE
  HeuristicMethod iMethod = HeuristicMethod( atoi(argv[3]) );
#else
  HeuristicMethod iMethod = ts;
#endif


  // escape method (use some reasonable option out of HeuristicMethod)
  // explicit escape moves are used by iterated local search and tabu search
  
  HeuristicMethod escapeStrategy = rw;


  // tabu search criterion (only used for tabu search methods)
  //                 0       1    2           3     
  enum TSCriterion { stat=0, rem, trajectory, reactive };
#ifdef COMMANDLINE
  TSCriterion tsCriterion = TSCriterion( atoi(argv[4]) );
#else
  TSCriterion tsCriterion = reactive;
#endif


  // set some method parameters (may or may not be used below)

#ifdef COMMANDLINE
  float maxTimeInSeconds = atof(argv[5]);
#else
  float maxTimeInSeconds = 0;
#endif


#ifdef COMMANDLINE
  unsigned long maxIt = atol(argv[6]);
#else
  unsigned long maxIt = 10;
#endif


// only use depth = 1 for the mtsp so far
#ifdef COMMANDLINE
  int depth = atoi(argv[7]);
#else
  int depth = 1;
#endif


  // construct some objects

  Heuristic<Cp::S> * heuristic
    = new Heuristic<Cp::S>;

  OmegaInterface<Cp::S> *omegaInterface
    = new OmegaInterface<Cp::S>();

  Heuristic<Cp::S> *doNothingHeuristic
    = new Heuristic<Cp::S>( );


  Heuristic<Cp::S> *escape
    = new Heuristic<Cp::S>( );

	switch( escapeStrategy ) 
	{
	  case rw:       // 10 random moves
  		escape = new IteratedLocalSearch<CRandomWalk<Cp> >
                   ( observer, 0, 10, 1, omegaInterface, doNothingHeuristic, 1, false );
		  break;
	  case saj:
		  escape = new SimulatedAnnealingJetal< CSimulatedAnnealingJetal<Cp> >
					         ( observer, 0, 0, 0, 1, false );
		  break;
	  default:  // no escape
  		escape = doNothingHeuristic;
	}


// IteratedLocalSearch
//   see search.h for documentation
//   argument 1: observer
//   argument 2: maxTimeInSeconds (default=0, i.e. no restriction)
//   argument 3: maximum number of moves (default=0, i.e. no restriction)
//   argument 4: repetitions (default=1)
//   argument 5: omega interface (default=no/false)
//   argument 6: diversification (default=no diversification)
//   argument 7: neighborhood depth (default=1)
//   argument 8: return best solutions found (instead of final solution visited) (default=true)

  if ( iMethod == sd )
    heuristic = new IteratedLocalSearch<CSteepestDescent<Cp> >
    ( observer );

  if ( iMethod == fd )
    heuristic = new IteratedLocalSearch<CFirstDescent<Cp> >
    ( observer, 0  );

  if ( iMethod == rw )
    heuristic = new IteratedLocalSearch<CRandomWalk<Cp> >
    ( observer, 0, maxIt );

  if ( iMethod == sdma )
    heuristic = new IteratedLocalSearch<CSteepestDescentMildestAscent<Cp> >
    ( observer, 0, maxIt, 1, omegaInterface, doNothingHeuristic, 1, true );

  if ( iMethod == isd )
    heuristic = new IteratedLocalSearch<CSteepestDescent<Cp> >
    ( observer, 0, 0, 5, omegaInterface, escape );


// Variations of SIMULATED ANNEALING
//   see sa.h for documentation
//   argument 1: observer
//   argument 2: maxTimeInSeconds (default=0, i.e. no restriction)
//   argument 3: maxMoves (default=0, i.e. no restriction)
//   argument 4: omega interface (default=no/false)
//   argument 5: neighborhood depth (default=1)
//   argument 6: return best solutions found (instead of final solution visited) (default=true)
//   argument 7: initialTemperature (default=100)
//   argument 8: repetitionsInterval (default=1)
//   argument 9: deltaRepetitions (default=1)
//   argument 10: numberOfReheatings (default=0)

  if ( iMethod == sa )
    heuristic = new GeneralSimulatedAnnealing< CClassicSimulatedAnnealing<Cp> >
    ( observer, 5 );

  if ( iMethod == ta )
    heuristic = new GeneralSimulatedAnnealing< CThresholdAccepting<Cp> >
    ( observer, 5 );

  if ( iMethod == gd )
    heuristic = new GeneralSimulatedAnnealing< CGreatDeluge<Cp> >
    ( observer, 5 );

  if ( iMethod == rtrt )
    heuristic = new GeneralSimulatedAnnealing< CRecordToRecordTravel<Cp> >
    ( observer, 5 );


// SIMULATED ANNEALING according to Johnson et al. (1989)
//   see sa.h for documentation
//   argument 1: observer
//   argument 2: maxTimeInSeconds (default=0, i.e. no restriction)
//   argument 3: maxMoves (default=0, i.e. no restriction)
//   argument 4: omega interface (default=no/false)
//   argument 5: neighborhood depth (default=1)
//   argument 6: return best solutions found (instead of final solution visited) (default=true)
//   argument 7: initialAcceptanceFraction (default=0.4)
//   argument 8: frozenAcceptanceFraction (default=0.02)
//   argument 9: sizeFactor (default=16)
//   argument 10: frozenParameter (default=5)
//   argument 11: numberOfReheatings (default=0)

  if ( iMethod == saj )
    heuristic = new SimulatedAnnealingJetal< CSimulatedAnnealingJetal<Cp> >
    ( observer, 0, 0, 0, 1, true, 0.4f, 0.02f, 16, 5, 4 );


// TABU SEARCH
//   see tabu.h for documentation
//   argument 1: observer
//   argument 2: maxTimeInSeconds (default=0, i.e. no restriction)
//   argument 3: maximum number of moves (0 means no restriction)
//   argument 4: omega interface (default=no/false)
//   argument 5: tabu criterion (default=no tabu criterion, meaningless?)
//   argument 6: escape heuristic (default=no diversification)
//   argument 7: neighborhood depth (default=1)

  if ( ( iMethod == ts ) || ( iMethod == tspen ) )
  {
    // select/define tabu criterion

    TabuCriterion<Cp> *tabuCriterion;
    switch( tsCriterion ) 
    {
      case stat:  // arguments: observer, tabu list length (default=7), tabu threshold (default=1)
        tabuCriterion = new StaticTabuCriterion<Cp>( observer, 3, 1 );
        break;
      case trajectory:  // arguments: observer
        tabuCriterion = new StrictTabuCriterionByTrajectory<Cp>( observer );
        break;
      case reactive:  // arguments: observer, tabu threshold (default=1),
                      // adaptationParameter1 (default=1.2), adaptationParameter2 (default=2),
                      // maximum tabu list length (default=50), chaosParameter (default=3),
                      // repetitionParameter (default=3), smoothingParameter = 0.5
        tabuCriterion = new ReactiveTabuCriterion<Cp>( observer, 2 );
        break;
      default:  // nothing is tabu
        tabuCriterion = new TabuCriterion<Cp>( );
    }

    // select/define actual tabu search method

    if ( iMethod == ts )
      heuristic = new TabuSearch< CTabuSearchByTabuization<Cp> >
        ( observer, maxTimeInSeconds, maxIt, omegaInterface, tabuCriterion, escape, 1 );
    else  // iMethod == tspen
      heuristic = new TabuSearch< CTabuSearchByPenalties<Cp> >
        ( observer, maxTimeInSeconds, maxIt, omegaInterface, tabuCriterion, escape, 1 );
  }

  
  //  if ( iMethod == ea )
  //    heuristic = new EvolutionaryAlgorithm<ThisEAConfiguration>
  //    ( observer, problem, 10, 0, 10 );


  // execute heuristic

  heuristic->search( solution );

  observer->activate();
  observer->bestEvaluation( solution.upperBound( ) );
  observer->currentSolution();
  observer->getStream() << solution;
  observer->flush();
  observer->end( " " );
  observer->cpuTimeOfProcess( round( (double(clock()))/CLOCKS_PER_SEC, 3 )  );

  
  delete heuristic;
  delete doNothingHeuristic;
  delete escape;
  delete omegaInterface;

  
  return 0;
}
